
"""
The addressbook holds a list of permission slips for recipients to which we
are allowed to send mail, and incoming permissions for senders from whom we
will accept mail.

Each OwnerEntry has a separate AddressBook.
"""

from petmail import util, permission, crypto, request
from petmail.interfaces import IAddressBook, IIDRecord, \
     IOwnerEntry, IOwnerPermission, IPermissionSlip
from petmail import interfaces
from petmail.message import OutboundMessage
from petmail.idrecord import IDRecord
from petmail import plugin
from petmail.transport import mail

from twisted.python.components import implements
from twisted.python import log
from twisted.internet import defer
from twisted.application import service

import os



# standard AddressBookEntry plugins

class UpdateTheirIDRecord(plugin.AddressBookEntryPlugin):
    """If their copy of my ID Record is out of date, include a copy. If we
    don't have specific permission, then we probably aren't in their address
    book, so include a copy regardless of how recently we sent the last one."""

    lastSentOwnerEntry = 0 # the last time we sent them our IDRecord

    def outboundMessage(self, msg):
        lognode = msg.log("entry-UpdateTheirIDRecord")
        if msg.anonymous:
            lognode.log("anonymous, not sending our IDRecord")
            return
        if not self.entry().ourPermission:
            # we probably aren't in their addressbook, so send it.
            lognode.log("no ourPermission, so sending it")
            return self.addIDRecord(msg, lognode)
        lastSent = self.lastSentOwnerEntry
        lastModified = self.owner().lastModified
        lognode.log("lastSent %s, lastModified %s" % (lastSent,lastModified))
        if lastSent < lastModified:
            return self.addIDRecord(msg, lognode)

    def addIDRecord(self, msg, lognode):
        d = msg.senderEntry.sign()
        d.addCallback(self._gotSigned, msg, lognode)
        return d

    def _gotSigned(self, signed, msg, lognode):
        msg.addNode("senderIDRecord", signed)
        msg.worthSending = True
        self.lastSentOwnerEntry = util.now()
        # TODO: update lastSent when ack is received
        lognode.log("included copy of our IDRecord")

class AddReferences(plugin.AddressBookEntryPlugin):
    """Take any references to other Identities and create the appropriate
    message nodes so the recipient can add them to their addressbook.
    """

    def outboundMessage(self, msg):
        if not msg.references:
            return True # nothing to add
        nodes = []
        for entry in msg.references:
            refnode = { 'name': entry.getName(),
                        'id': entry.idRecord.ID,
                        'idrecord' : entry.idRecord.signed,
                        'offerPermission' : 0,
                        }
            #lognode2 = msg.log("adding %s" % entry.getName(), parent=lognode)

            # if we vouch for the person we're sending this to,
            #  AND the referencee allows us to grant third-party permission
            # then we tell the recipient that we can help them get permission
            recip = msg.recipientEntry
            if recip and not msg.anonymous:
                msg.log("vouchFor=%s, ourPermission=%s" % \
                        (recip.vouchFor, bool(entry.ourPermission)),
                        )#parent=lognode2)
                if entry.ourPermission:
                    msg.log("ourPermission.canGrantPermission=%s" % \
                            entry.ourPermission.canGrantPermission,
                            )#parent=lognode2)
                if (recip.vouchFor
                    and entry.ourPermission
                    and entry.ourPermission.canGrantPermission):
                    msg.log("offerPermission=1")#, parent=lognode2)
                    refnode['offerPermission'] = 1
            nodes.append(refnode)
        msg.addNode('references', nodes)

    
class ReturnPermission(plugin.AddressBookEntryPlugin):
    # TODO: move this stamp from AddressBookEntry to here
    #lastSentTheirPermission = 0

    def outboundMessage(self, msg):
        # When sending an rfc822 message, it is polite to give the recipient
        # permission to respond.
        d = self.maybeGrantPermission(msg)
        d.addCallback(self._updatePermission)

    def maybeGrantPermission(self, msg):
        owner = msg.senderEntry
        recip = msg.recipientEntry
        # grant permission if we are really talking to them, expect to talk
        # to them again in the future, and aren't specifically prohibited
        # from granting it

        if (msg.hasNode("rfc822")
            and not msg.anonymous
            and owner
            and recip
            and not recip.theirPermission # already have it
            and not recip.doNotGrantReturnPermission
            and owner.grantReturnPermission):
            msg.log("granting return permission")
            owner.grantPermissionTo(recip, msg)
            return defer.succeed(msg)
        return defer.succeed(msg)

    def _updatePermission(self, msg):
        # If the recipient has Permission (the 'theirPermission' field of
        # their address book entry), and the timestamps indicate that their
        # copy is out of date, I will insert a PermissionSlip into the
        # outbound message.
        recip = msg.recipientEntry
        if (not msg.anonymous
            and recip
            and recip.needToSendTheirPermission()):
            msg.log("updating their permission")
            # TODO: maybe let encodePermissionSlip() return a Deferred
            msg.addNode("yourPermission", recip.encodePermissionSlip())
            msg.worthSending = True
            recip.sentTheirPermission(util.now())
            # TODO: update p.lastSent when ack is received. Retransmits.


class UpdateMyIDRecord(plugin.AddressBookEntryPlugin):
    """Update my copy of their ID Record. This is the catcher for the update
    inserted by the UpdateTheirIDRecord outbound processor."""
    #provides = ["updatedIDRecord"]

    def inboundMessage(self, msg):
        signed = msg.getNode("senderIDRecord")
        if not signed:
            return
        # TODO: it would be nice to avoid the decrypt here for cases where
        # we don't need to update anything. Maybe the record's timestamp
        # should be sent outside the signed blob, if it matches our current
        # copy we can skip this step. Doubly useful because updateIDRecord
        # also does an addPubkey.
        msg.log("updating my ID Record")
        d = self.entry().updateIDRecord(signed, msg)
        return d

class UpdateOurPermission(plugin.AddressBookEntryPlugin):
    """If the inbound message has a PermissionSlip, verify the signature and
    update the OurPermission field of their address book entry. If
    requested, acknowledge receipt of the PermissionSlip. This is the
    catcher for the UpdateTheirPermission outbound processor.
    """
    #provides = ["updatedOurPermission"]

    def inboundMessage(self, msg):
        if not msg.hasNode("yourPermission"):
            return
        slip = msg.getNode("yourPermission") # TODO: part of rfc822 node?
        # slip might be None, indicating we have no specific permission
        if slip:
            msg.log("updating my PermissionSlip")
        else:
            msg.log("ourPermission was revoked")
        d = self.entry().updateOurPermission(slip)
        d.addCallback(self._didUpdate, msg)
        # TODO: send ack, if requested
        return d

    def _didUpdate(self, result, msg):
        msg.senderEntry.modified()

class HandleVoucherRequest(plugin.AddressBookEntryPlugin):
    def inboundMessage(self, msg):

        req = msg.getNode("check-voucher")
        if not req:
            return True
        requestorID = req['requestorID']
        # requestorID claims that we vouch for them
        doVouch = False

        # they should be in our addressbook
        book = self.agent().getAddressBook(msg.recipientEntry.ID)
        requestor = book.lookupID(requestorID)
        if requestor:
            doVouch = requestor.vouchFor

        reply = OutboundMessage(self.agent())
        reply.addNode('check-voucher-reply', { 'requestorID': requestorID,
                                               'doVouch': doVouch })
        self.agent().sendReply(reply, msg)

### end plugins


class AddressBook(service.MultiService):
    # TODO: needs test coverage
    __implements__ = IAddressBook, service.MultiService.__implements__
    debug = 0
    owner = None

    def __init__(self):
        service.MultiService.__init__(self)
        self.entries = []
        self.watchers = []

    def setServiceParent(self, parent):
        assert implements(parent, IOwnerEntry)
        service.MultiService.setServiceParent(self, parent)
        self.owner = parent # TODO: legacy

    def disownServiceParent(self):
        service.MultiService.disownServiceParent(self)
        del self.owner

    def __getstate__(self):
        d = service.MultiService.__getstate__(self)
        d['watchers'] = []
        return d

    def getOwner(self):
        return self.parent

    def subscribe(self, callback):
        """Register a callback to be run when the addressbook changes (such
        as when an entry is added, perhaps due to receipt of a message from
        a new sender). The callback is run with one argument, this
        AddressBook instance. The callback is not persisted."""
        self.watchers.append(callback)
    def unsubscribe(self, callback):
        self.watchers.remove(callback)
    def notifyWatchers(self):
        for w in self.watchers:
            w(self)

    def notify(self, action, entry, lognode=None):
        # fire entry-changed plugins. Also notify any subscribers.
        d = defer.succeed(None)
        owner = self.getOwner()
        if owner:
            d.addCallback(lambda res:
                          owner.entryChanged(action, entry, lognode))
            agent = owner.getAgent()
            if agent:
                d.addCallback(lambda res:
                              agent.entryChanged(action, entry, lognode))
        d.addCallback(lambda res: self.notifyWatchers())
        return d

    def countEntries(self):
        return len(self.entries)

    def lookupID(self, ID):
        if not isinstance(ID, type("")):
            print "AddressBook.lookupID got %s (type %s), not string" % \
                  (ID, type(ID))
            assert False
        for e in self.entries:
            if e.idRecord and e.idRecord.ID == ID:
                return e
        return None
    def lookupPetName(self, name):
        # TODO: deprecate this
        return self.lookupName(name)
    def lookupName(self, name):
        for e in self.entries:
            if e.petname == name:
                return e
        return None
    def lookupAddress(self, address):
        for e in self.entries:
            if address in e.idRecord.addresses:
                return e

    def addEntry(self, entry, message=None):
        # message is the message which caused this entry to be added
        # TODO: stash (parts of) the message
        assert implements(entry, interfaces.IAddressBookEntry)
        entry.setServiceParent(self)
        self.entries.append(entry)
        return self.notify("add", entry)

    def addAddress(self, addr):
        d = self.getOwner().getAgent().retrieveAddress(addr)
        d.addCallback(self.addIDRecord, reason="AddressBook.addAddress")
        return d

    def addFromSigned(self, signed, reason="AddressBook.addFromSigned"):
        # TODO: deprecate this method. BUT: it is used by
        # ReceivedMessage.addSender()
        return self.addIDRecord(signed, reason)

    def addArmoredIDRecord(self, armored,
                           reason="AddressBook.addArmoredIDRecord"):
        signed = util.dearmor(armored, "IDRECORD")
        return self.addIDRecord(signed, reason)

    def addIDRecord(self, signed, reason="AddressBook.addIDRecord"):
        """Add the record to the address book, or update the existing entry.
        Return a Deferred which will fire (with the new entry) when the entry
        has been fully added.
        """
        e = AddressBookEntry()
        d = e.createFromIDRecord(signed, self.owner.parent.crypto, reason)
        d.addCallback(self._addFromSigned1, reason)
        d.addCallback(lambda res: e)
        return d

    def _addFromSigned1(self, entry, reason):
        # TODO: should addEntry accept an InboundMessage or simple text?
        return self.addEntry(entry, reason)

    def deleteCorrespondent(self, entry):
        return self.deleteEntry(entry)

    def deleteEntry(self, entry):
        """Returns a Deferred which fires when the entry has been deleted.
        """
        if entry.getTheirPermission:
            entry.revokeTheirPermission("rfc822")
            # would normally do entry.modified() here, but we need to push
            # out the update immediately instead of waiting for the holdoff
            # time
            d = entry.updateRecipient()
            d.addErrback(lambda err: None) # ignore failures
        else:
            d = defer.succeed(None)
        d.addCallback(self._deleteEntry2, entry)
        return d

    def _deleteEntry2(self, results, entry):
        entry.disownServiceParent()
        self.entries.remove(entry)
        return self.notify("delete", entry)

class AddressBookEntry(plugin.PluginHolderMixin,
                       service.MultiService):
    __implements__ = (interfaces.IAddressBookEntry,
                      interfaces.ICorrespondent,
                      service.MultiService.__implements__)
    updateHoldoffTime = 30
    doUpdate = True
    addressbook = None # TODO: legacy
    surbSupply = None
    # plugins
    acceptablePluginInterface = interfaces.IEntryPlugin
    standardPlugins = [UpdateTheirIDRecord, AddReferences, ReturnPermission,
                       UpdateMyIDRecord, UpdateOurPermission,
                       HandleVoucherRequest]

    def __init__(self):
        service.MultiService.__init__(self)
        # self.parent is the AddressBook
        # self.parent.parent is the OwnerEntry
        # self.parent.parent.parent is the Agent
        self.petname = None
        self.idRecord = None
        self.ourPermission = None   # what we use to send mail to them
        # .ourPermission is always a PermissionSlip
        self.theirPermission = None # what we require when they send to us
        # .theirPermission is always an OwnerPermission
        self.vouchFor = True
        self.doNotGrantReturnPermission = False
        self.lastSentTheirPermission = 0 # last time we sent .theirPermission
        self.watchers = [] # these care about the entry being changed

        # extensions
        for p in self.standardPlugins:
            self.addPlugin(p())

    def __getstate__(self):
        d = self.__dict__.copy()
        d['watchers'] = []
        return d

    def setServiceParent(self, parent):
        assert implements(parent, IAddressBook)
        service.MultiService.setServiceParent(self, parent)
        self.addressbook = parent # TODO: legacy

    def disownServiceParent(self):
        service.MultiService.disownServiceParent(self)
        del self.addressbook

    def getProperty(self, pname):
        if pname == "petname":
            return self.petname
        if pname == "vouchFor":
            return self.vouchFor
        if pname == "grantReturnPermission":
            return not self.doNotGrantReturnPermission
        if pname == "id":
            return self.idRecord.ID
        if pname == "name":
            return self.idRecord.name
        if pname == "addresses":
            return self.idRecord.addresses
        if pname == "transports":
            return map(lambda t: ":".join(map(str, t)),
                       self.idRecord.transports)
        raise IndexError("%s has no property named %s" % (self, pname))

    def setProperty(self, pname, value):
        assert pname in ("petname", "vouchFor", "grantReturnPermission")
        d = defer.succeed(None)
        if pname == "petname":
            assert type(value) is str or value is None
            self.petname = value
            # tell the addressbook that the petname has changed, since one
            # of its watchers may use the petname as a display index
            if self.addressbook:
                d.addCallback(lambda res:
                              self.addressbook.notify("changed", self))
        if pname == "vouchFor":
            assert type(value) is bool
            self.vouchFor = value
        if pname == "grantReturnPermission":
            assert type(value) is bool
            self.doNotGrantReturnPermission = not value
        d.addCallback(self.modified)
        return d

    def getOurPermission(self, ptype="rfc822"):
        # this obtains our specific permission if available, otherwise it
        # falls back to the generic permission published by the
        # correspondent
        if ptype != "rfc822":
            raise IndexError("unknown permission type %s" % (ptype,))
        if self.ourPermission is not None:
            return self.ourPermission
        return self.idRecord.getPermission(ptype)

    def getOurPermissionType(self, ptype):
        if ptype != "rfc822":
            raise IndexError("unknown permission type %s" % (ptype,))
        if self.ourPermission is not None:
            return "specific"
        return "generic"

    def needToSendTheirPermission(self, debug=False):
        p = self.theirPermission
        if not p:
            if self.lastSentTheirPermission:
                # revoked
                if debug: print "permission was revoked, must update"
                return True
            if debug: print "no permission to update"
        else:
            if debug:
                print "last sent %d, lastModified %d" % \
                      (self.lastSentTheirPermission, p.lastModified)
            if self.lastSentTheirPermission < p.lastModified:
                if debug: print "permission changed recently, must update"
                return True
            if debug: print "permission is up-to-date"
        return False
    def sentTheirPermission(self, when):
        if self.theirPermission:
            self.lastSentTheirPermission = when
        else:
            self.lastSentTheirPermission = 0

    def getOwner(self):
        """Return the OwnerEntry of which this entry is a child."""
        if self.addressbook:
            return self.addressbook.owner
        return None

    def createFromIDRecord(self, signed, cagent, cause):
        """Create our .idRecord with the signed IDRecord they just sent us.
        Return a Deferred that will fire (with this entry as an argument)
        when the record is complete."""
        # TODO: record 'cause' as the cause of the update
        assert not self.idRecord
        self.idRecord = IDRecord()
        d = self.idRecord.addFromSigned(signed, cagent)
        d.addCallback(lambda x, entry=self: entry)
        return d

    def updateIDRecord(self, signed, cause):
        """Update this entry with the signed IDRecord they just sent us.
        Return a Deferred that will fire (with this entry as an argument)
        when the update has completed."""
        # TODO: record 'cause' as the cause of the update
        d = self.idRecord.updateFromSigned(signed)
        d.addCallback(self.modified)
        d.addCallback(lambda res: self)
        return d

    def subscribe(self, callback):
        self.watchers.append(callback)

    def unsubscribe(self, callback):
        self.watchers.remove(callback)

    def modified(self, results=None):
        for w in self.watchers:
            w(self)
        return self.notify("change")

    def notify(self, action, lognode=None):
        d = self.runPlugins("entryChanged", action, self, lognode)
        owner = self.getOwner()
        if owner:
            d.addCallback(lambda res:
                          owner.entryChanged(action, self, lognode))
            agent = owner.getAgent()
            if agent:
                d.addCallback(lambda res:
                              agent.entryChanged(action, self, lognode))
        return d

    def updateOurPermission(self, encoded):
        """Accept an encoded permission slip (usually from a
        'yourPermission' message node) and update the .ourPermission
        attribute with its contents.

        You are responsible for calling .modified() after this.
        """

        if self.ourPermission:
            self.ourPermission.unsubscribe(self.modified)
            self.ourPermission = None

        if encoded:
            if not self.ourPermission:
                self.ourPermission = permission.PermissionSlip()
                self.ourPermission.subscribe(self.modified)
            self.ourPermission.setFromEncoded(encoded)
        return self.modified()

    def _setOurPermission(self, perm):
        # for testing
        assert implements(perm, IPermissionSlip)
        if self.ourPermission:
            self.ourPermission.unsubscribe(self.modified)
        self.ourPermission = perm
        self.ourPermission.subscribe(self.modified)

    def getTheirPermission(self, ptype="rfc822"):
        if ptype != "rfc822":
            raise IndexError("unknown permission type %s" % (ptype,))
        return self.theirPermission

    def setTheirPermission(self, ptype, perm):
        assert ptype == "rfc822"
        assert implements(perm, IOwnerPermission)
        self.theirPermission = perm
        self.scheduleUpdate()

    def revokeTheirPermission(self, ptype):
        assert ptype == "rfc822"
        if self.theirPermission == None:
            return
        self.theirPermission = None
        self.scheduleUpdate()
    # TODO: changing the permission should send an update too. The
    # Permission object should call .parent.scheduleUpdate()

    def scheduleUpdate(self):
        self.modified()
        # tell the agent that we need to update this recipient with their
        # new settings
        if not self.addressbook:
            return
        if not self.doUpdate:
            return
        # TODO: don't schedule multiple updates
        owner = self.addressbook.owner
        agent = owner.parent
        why = "Must update recipient %s with their new settings" % \
              self.getName()
        d = agent.scheduleEvent(self.updateHoldoffTime, why)
        d.addCallback(self.updateRecipient)

    def updateRecipient(self, dummy=None):
        """Returns a Deferred which fires when the update message has been
        sent."""

        if not self.addressbook:
            return # this entry was deleted
        owner = self.addressbook.owner
        if not owner.parent:
            return # the owner was deleted
        # TODO: don't do the update if it isn't necessary (like, say, a
        # message was sent in the interim which took care of the matter).
        # This used to be handled with a .worthSending flag in the
        # OutboundMessage, but that was gross. Having the behavior sit solely
        # in the UpdateTheirIDRecord plugin is gross too: the decision-making
        # should probably take place here.
        agent = owner.parent
        m = OutboundMessage(agent)
        m.setSenderEntry(owner)
        m.setRecipientEntry(self)
        return agent.sendMessage(m)


    def getName(self):
        if self.petname != None:
            return self.petname
        if self.idRecord and self.idRecord.name != None:
            return '"%s"' % self.idRecord.name
        if self.idRecord:
            return self.idRecord.ID
        return "????"

    def encodePermissionSlip(self):
        """Return a string with an encoded copy of .theirPermission, which
        can be given to the sender. Returns None if they do not have
        specific permission.
        """

        if self.theirPermission:
            return self.theirPermission.encode()
        return None

    def getTransport(self, message, wantSecure):
        # if we have a SURBSupply for this recipient, we must use it
        if self.surbSupply:
            return mail.MailWithSURBs(self.surbSupply, message)
        return None # fall back to the IDRecord
