
"""
The Petmail agent runs on the local system, holds the private key, and acts
as an SMTP/POP/IMAP proxy for a local legacy MUA.
"""

import os.path

from twisted.application import service, internet
from twisted.internet import defer, reactor
from twisted.python.components import implements
from twisted.python import log, failure
from twisted.persisted import styles
from twisted.cred import portal

from petmail import addressbook, owner, message, permission, mailbox
from petmail import interfaces, processors, crypto, util, request, names
from petmail import plugin

from petmail.transport import dummy, mail, pb, imap, pop
try:
   from petmail.transport import jabber
except ImportError:
   # happens the first time the interpreter tries to import it
   jabber = None
if jabber and not jabber.supported:
   # happens second and later times
   jabber = None

from petmail.mm_util import MixminionContext
from petmail.idrecord import IDRecord
import petmail.log


# standard Agent plugins

class HandleReply(plugin.AgentPlugin):
    """See if the message is in response to an outstanding request."""

    def inboundMessage(self, msg):
        replyid = msg.getNode("replyID")
        if replyid is not None:
            msg.log("gotReply")
            self.agent().gotReply(replyid, msg)

class HandlePing(plugin.AgentPlugin):
    """See if the message is a ping request."""

    def inboundMessage(self, msg):
        pingID = msg.getNode("ping")
        if pingID:
            agent = self.agent()
            msg.log("sending ping response")
            pong = message.OutboundMessage(agent)
            pong.addNode('pong', pingID)
            agent.sendReply(pong, msg)



## end plugins


class Agent(plugin.PluginHolderMixin,
            service.MultiService,
            styles.Versioned):
    """This is the main object which defines the Petmail agent. There will
    be exactly one of these in each agent process (it can accomodate
    multiple senders). This is the object which is serialized to record the
    persistent state of the service, including the contents of the
    addressbook and all the stored Permissions.

    """

    __implements__ = (interfaces.IAgent,
                      portal.IRealm,
                      service.MultiService.__implements__,
                      #styles.Versioned.__implements__,
                      ) 
    persistenceVersion = 1

    debugProcessing = False
    validateSource = True
    smarthost = "localhost"
    doNotScheduleEvents = False
    acceptablePluginInterface = interfaces.IAgentPlugin
    standardPlugins = [HandleReply, HandlePing]

    def __init__(self, basedir):
        service.MultiService.__init__(self)
        self.basedir = basedir
        self.senders = {}
        self.addressbooks = {}
        self.inboxes = {}
        self.portal = mailbox.makePortal(self)
        self.pop3_port, self.imap_port, self.smtp_port = None, None, None
        self.inboundProcessors = []
        self.outboundProcessors = []
        self.inboundMessageHandlers = []
        self.addressLookupAgents = []
        self.addressPublishAgents = []
        self.smtpScheme = "Mail:localhost:petmail-user@localhost"
        self.nntpScheme = None
        self.mixminionPath = None
        self.mixminionVersion = None

        self.nextRequestID = 1
        self.outstandingRequests = {}
        self.scheduledEvents = [] # this is a list of (when, Request) tuples
        self.scheduledEventTimer = None

        # these lists track messages that are being processed. Each one must
        # support a .why() method which returns a string to indicate what the
        # message is waiting for. They should also have a timeout mechanism.
        self.pendingInboundMessages = []
        self.pendingOutboundMessages = []

        self.inlets = {} # keys are inlets, values are lists of owners

        self.eventLog = petmail.log.Log("agent")
        if 0:
            ds = debug.DebugService()
            ds.setServiceParent(self)
        self.defaultReturnPermission = permission.OwnerPermission()
        for p in self.standardPlugins:
            self.addPlugin(p())
        self.log("Agent initialized")

    def upgradeToVersion1(self): # post-0.1.2
        self.mixminionPath = None
        self.mixminionVersion = None
        self.inboxes = {}
        for o in self.addressbooks.keys():
            self.inboxes[o] = mailbox.Mailbox(o)
        self.portal = mailbox.makePortal(self)
        self.pop3_port, self.imap_port, self.smtp_port = None, None, None

    def log(self, *args, **kwargs):
        return self.eventLog.log(*args, **kwargs)

    def init(self):
        self.populateProcessors()
        self.crypto = crypto.EncryptionAgent(os.path.join(self.basedir,
                                                          "crypto"))
        if self.mixminionPath:
            self.mmc = MixminionContext(os.path.join(self.basedir, "mmc"))
            self.mmc.set_mm_path(self.mixminionPath)
            self.mmc.setServiceParent(self)

    def __getstate__(self):
        timer = self.scheduledEventTimer
        state = service.MultiService.__getstate__(self)
        state = styles.Versioned.__getstate__(self, state)
        state['scheduledEventTimer'] = None
        assert self.scheduledEventTimer == timer
        return state

    def startService(self):
        self.log("Agent.startService")
        self.populateProcessors()
        service.MultiService.startService(self)
        reactor.callLater(6, self.checkScheduledEvents)

    def stopService(self):
        self.log("Agent.stopService")
        if self.scheduledEventTimer:
            self.scheduledEventTimer.cancel()
            self.scheduledEventTimer = None
        return service.MultiService.stopService(self)

    def newIdentity(self, progress=None):
        """Returns a Deferred which will fire with the new OwnerEntry."""
        me = owner.OwnerEntry()
        d = self.crypto.generateKey("name", progress)
        d.addCallback(me.setKey)
        if progress:
            d.addCallback(lambda x: progress.done())
        d.addCallback(lambda x: me)
        return d

    def createIdentity(self, progress=None, perm="generic"):
        # TODO: deprecate createIdentity, use newIdentity instead
        d = self.newIdentity(progress)

        if perm == "generic":
            perm = self.newPermission("rfc822")
            # generic permission should never let just anybody grant it to
            # others
            perm.canGrantPermission = False
        def setperm(me):
            me.setGenericPermission("rfc822", perm)
            return me
        d.addCallback(setperm)
        return d

    # OwnerEntry functions

    def addIdentity(self, i):
        # TODO: deprecate addOwner, use addIdentity instead
        d = self.addOwner(i)
        d.addCallback(lambda x: i.publish())
        return d

    def addOwner(self, owner):
        # TODO: deprecate
        self.senders[owner.ID] = owner
        book = addressbook.AddressBook()
        book.setServiceParent(owner)
        self.addressbooks[owner] = book
        self.inboxes[owner] = mailbox.Mailbox(owner)
        owner.setServiceParent(self)
        d = self.ownerChanged("add", owner)
        return d

    def removeOwner(self, owner):
        owner.disownServiceParent()
        del self.senders[owner.ID]
        del self.addressbooks[owner]
        del self.inboxes[owner]
        d = self.ownerChanged("delete", owner)
        return d

    def getAllIdentities(self):
        ownerlist = self.addressbooks.keys()
        ownerlist.sort(lambda x,y: cmp(x.getName(), y.getName()))
        return ownerlist

    def getIdentity(self, id):
        return self.senders[id]

    def getOwnerByName(self, name):
        for sender in self.senders.values():
            if name == sender.getName():
                return sender
        raise KeyError("no owner named '%s'" % name)

    def getInbox(self, owner):
        return self.inboxes[owner]


    def setSMTP(self, scheme):
        self.smtpScheme = scheme

    def setNNTP(self, scheme):
        self.nntpScheme = scheme

    def setAddressLookup(self, scheme):
        while self.addressLookupAgents:
            self.addressLookupAgents.pop()
        while self.addressPublishAgents:
            self.addressPublishAgents.pop()
        if not len(scheme):
            return
        for s in scheme.split(","):
            la = names.getLookupAgentFromString(s)
            self.registerAddressLookupAgent(la)
    def getAddressLookup(self):
        return ",".join([la.describe() for la in self.addressLookupAgents])

    def registerAddressLookupAgent(self, agent):
        self.addressLookupAgents.append(agent)
        self.addressPublishAgents.append(agent)
    def unregisterAddressLookupAgent(self, agent):
        self.addressLookupAgents.remove(agent)
        self.addressPublishAgents.remove(agent)


    def getAddressBook(self, senderID):
        return self.addressbooks[self.getSender(senderID)]
    def getSender(self, senderID):
        return self.senders[senderID]

    def retrieveAddress(self, address):
        # TODO: move this into AddressBook?
        """This returns a deferred which fires with a signed IDRecord, or
        errbacks with an Unsendable if no record could be retrieved."""

        # if 'address' is really an armored IDRecord, just de-armor it
        if util.is_armored(address, "IDRECORD"):
            signed = util.dearmor(address, "IDRECORD")
            return defer.succeed(signed)

        methods = self.addressLookupAgents[:]
        d = defer.succeed(None)
        # as wacky as it seems, this double-callback solves the problem of
        # allowing the very first method.lookup() to raise an immediate
        # exception
        d.addBoth(self._retrieveAddress1, address, methods)
        d.addBoth(self._retrieveAddress1, address, methods)
        d.addCallback(self._retrieveAddressDone)
        return d

    def _retrieveAddress1(self, results, address, methods):
        if not results or isinstance(results, failure.Failure):
            if not methods:
                raise interfaces.Unsendable, \
                      "none of my lookup agents could handle address '%s'" % \
                      address
            method = methods.pop(0)
            d = method.lookup(self, address)
            d.addBoth(self._retrieveAddress1, address, methods)
            return d
        else:
            return results

    def _retrieveAddressDone(self, signed):
        if not type(signed) == str:
            self.log("address lookup returned a non-string", signed)
            assert type(signed) == str
        return signed

    def populateProcessors(self):
        self.inboundProcessors = [pc() for pc in processors.allInbound]
        self.outboundProcessors = [pc() for pc in processors.allOutbound]

    def getInboundProcessors(self):
        return self.inboundProcessors[:]

    def getOutboundProcessors(self):
        return self.outboundProcessors[:]

    def getTicketAgents(self):
        # override this to provide ticket agents
        return []

    def addressBookChanged(self, addrbook):
        # override this to update something when the address book changes
        pass

    def newPermission(self, ptype):
        if ptype != "rfc822":
            raise IndexError("unknown permission type %s" % (ptype,))
        p = permission.OwnerPermission()
        return p

    def newInlet(self, itype, **args):
        inletTypes = {"dummy": dummy.DummyInlet,
                      "pb": pb.PBInlet,
                      "maildir": mail.MaildirInlet,
                      "mailsocket": mail.MailSocket,
                      "imap": imap.IMAPPoller,
                      "pop3": pop.POPPoller,
                      }
        if jabber:
            inletTypes["jabber"] = jabber.JabberInlet
        return inletTypes[itype](**args)

    def setMixminionPath(self, path, version=None):
        # setMixminionPath(None) means it is not available
        self.mixminionPath = path
        if path:
            self.mixminionVersion = version
    def getMixminionPath(self):
        return self.mixminionPath, self.mixminionVersion

    def canDecryptType3(self):
        if self.mixminionPath:
            return True
        return False

    def decryptType3(self, mmencrypted):
        """I decrypt an encrypted Type III anonymous message, which was sent
        through one of my SURBs."""
        return self.mixminionContext.decrypt(mmencrypted)

    # inlets

    def addInlet(self, owner, inlet):
        if self.inlets.has_key(inlet):
            self.inlets[inlet].append(owner)
        else:
            self.inlets[inlet] = [owner]
            inlet.setServiceParent(self)

    def removeInlet(self, owner, inlet):
        # returns a Deferred that indicates when the service has been
        # stopped: i.e., when the TCP port has stopped listening
        self.inlets[inlet].remove(owner)
        if not self.inlets[inlet]:
            del self.inlets[inlet]
            return inlet.disownServiceParent()

    # sending

    def newMessage(self, mtype="rfc822"):
        m = message.OutboundMessage(self)
        return m

    def sendMessageToNameOrAddress(self, msg, name):
        # this can accept a pet name
        book = self.getAddressBook(msg.senderEntry.ID)
        e = book.lookupPetName(name)
        if e:
            msg.setRecipientEntry(r)
            return self.sendMessage(msg)
        return self.sendMessageToAddress(msg, name)
        

    def sendMessageToAddress(self, msg, address):
        assert msg.senderEntry
        d = self.retrieveAddress(address)
        d.addCallback(self._sendMessageToAddress2, msg)
        return d
    def _sendMessageToAddress2(self, signed, msg):
        if not signed:
            raise Unsendable("unable to look up address")
        if msg.getNode('rfc822'):
            # add the recipient to our address book. This will add their
            # public key as a side-effect.
            #msg.log("adding entry to addressbook", parent=lognode)
            book = self.getAddressBook(msg.senderEntry.ID)
            msg.log("installing outbound AddressBook Entry")
            d = book.addFromSigned(signed, reason="sending rfc822 msg")
            d.addCallback(msg.setRecipientEntry)
            d.addCallback(lambda dummy: self.sendMessage(msg))
            return d
        # just add their key so we can send the message. This implies that
        # we don't particularly expect to send them another message. TODO:
        # control messages to frequent destinations (remailer users talking
        # to Address/Permission Servers, etc) should reference an IDRecord
        # and not an address, otherwise they'll have to do multiple long
        # lookup cycles for the same target.
        #msg.log("adding their IDRecord to our cryptoagent", parent=lognode)
        r = IDRecord()
        d = r.addFromSigned(signed, self.crypto)
        d.addCallback(msg.setRecipientRecord)
        d.addCallback(lambda dummy: self.sendMessage(msg))
        return d


    def sendMessage(self, msg):
        assert msg.recipientEntry or msg.recipientRecord
        self.pendingOutboundMessages.append(msg)
        d = msg.startProcessing()
        d.addCallback(self.deliverMessage)
        d.addBoth(self._sendMessage2, msg)
        return d

    def deliverMessage(self, msg):
        return msg.deliver()

    def _sendMessage2(self, res, msg):
        if isinstance(res, failure.Failure):
            msg.log("sendMessageFailed", res)
            self.log("sendMessageFailed")
        else:
            msg.log("delivery complete")
        msg.stopLogging()
        self.pendingOutboundMessages.remove(msg) # TODO: only upon ack
        return res

    # things that may happen during send

    def grantReturnPermission(self, entry, msg):
        # add Specific Permission to this address book entry
        p = self.defaultReturnPermission.copy()
        entry.setTheirPermission(p)
        entry.modified()


    # receiving

    def receiveMessage(self, encrypted, source=None, debug=False):
        """Feed an inbound message to the decryption/verification stage. If
        accepted, the message will be put on the inbox and
        notifyNewMessage() will be called.

        This will be called by a transport layer, triggered by a mailbox
        poll or SMTP receive from the outside world.

        Returns a Deferred which will fire when the message has been fully
        processed. Remember that it might set up timers to make something
        happen later.
        
        @type encrypted: C{string}
        @param encrypted: the encrypted message blob

        @type source: C{tuple}
        @param source: (inlet, surbidentity, surbid)
        """

        self.log("Agent.receiveMessage")
        rm = message.ReceivedMessage(self, encrypted)
        if debug: rm.debug = True
        d = rm.decrypt()
        d.addCallback(self._receiveMessage2, rm, source)
        d.addErrback(self._receiveMessageFailed, rm)
        return d

    def _receiveMessage2(self, recipientEntry, rm, source):
        if self.validateSource:
            # TODO: assert that the source information is ok (it was
            # received on the right inlet, with the right surbid, etc)
            inlet, surbidentity, surbid = source
            if not recipientEntry in self.inlets[inlet]:
                self.log("WEIRD: " +
                         "message for %s received on wrong inlet %s" % \
                         (recipientEntry, inlet))
                # TODO: yell, this indicates a shared-pseudonym guessing
                # attack, or someone working with an out-of-date IDRecord
        self.pendingInboundMessages.append(rm)
        self.acceptMessage(rm)
        d = rm.startProcessing()
        d.addCallback(self._receiveMessageDone)
        return d

    def _receiveMessageDone(self, msg):
        """This is called when an inbound message has been fully processed.
        Everything interesting has already been done by the Processors, so
        all that remains is to log receipt (and possibly send back an
        acknowledgement)."""
        self.pendingInboundMessages.remove(msg)

    def _receiveMessageFailed(self, why, msg):
        # an error occurred during processing, should log
        print "receiveMessageFailed", why
        self.pendingInboundMessages.remove(msg)
        return why

    def acceptMessage(self, rm):
        """A message has arrived. This is called before any plugins or
        processors are invoked. The message is delivered to all handlers.
        Note that this merely means the message has been received, and the
        sender and recipient have been checked (.senderEntry is either valid
        or None). No permission checks have been performed, those are left up
        to a plugin."""
        for h in self.inboundMessageHandlers:
            h(rm)

    def rejectRFC822(self, rm, why):
        # still provided for testing purposes
        pass


    # request/response handling

    def getRequestID(self):
        # TODO: use nonce here to prevent people from messing with them
        reqid = self.nextRequestID
        self.nextRequestID += 1
        return reqid

    def addRequest(self, req):
        assert implements(req, interfaces.IRequest)
        assert not self.outstandingRequests.has_key(req.reqid)
        self.outstandingRequests[req.reqid] = req

    def sendRequest(self, msg):
        """Like sendMessage, but marks the message to indicate that we want
        to receive a reply. The Deferred we return is fired when the
        response comes back."""

        assert implements(msg, interfaces.IOutboundMessage)
        reqid = self.getRequestID()
        msg.addNode("requestID", reqid)
        req = request.Request("request")
        req.reqid = reqid
        self.addRequest(req)
        self.sendMessage(msg)
        return req.deferred

    # request is caught by an InboundProcessor of some sort, which will
    # call sendReply

    def sendReply(self, msg, inReplyTo):
        reqid = inReplyTo.getNode("requestID")
        msg.addNode("replyID", reqid)
        msg.setSenderEntry(inReplyTo.recipientEntry)
        # must get the sender's IDRecord to respond to them
        # TODO: let the message handle more of this, as it may have a
        # response channel that is independent of the IDRecord
        d = inReplyTo.getIDRecord()
        d.addCallback(self._sendReply2, msg)
        return d

    def _sendReply2(self, senderIDRecord, msg):
        msg.setRecipientRecord(senderIDRecord)
        return self.sendMessage(msg)

    # reply is caught in processors.HandleReply and sent to gotReply

    def gotReply(self, reqid, message):
        req = self.outstandingRequests.get(reqid)
        if req:
            req.fire(message)
            del self.outstandingRequests[reqid]


    # scheduled events (restart-survivable DelayedCalls)

    def scheduleEvent(self, seconds, description):
        """I schedule an event to happen some time in the future. I return a
        Deferred which will fire (with None) when the time is reached."""

        if self.doNotScheduleEvents:
            #return defer.fail(RuntimeError("this Agent does not schedule events"))
            return defer.Deferred() # will never fire
        req = request.Request(description)
        e = (util.now() + seconds, req)
        self.scheduledEvents.append(e)
        self.checkScheduledEvents()
        return req.deferred

    def checkScheduledEvents(self):
        e = self.log("Agent.checkScheduledEvents [%d]" % \
                     len(self.scheduledEvents))
        # remember that this is reentrant
        if self.scheduledEventTimer:
            if self.scheduledEventTimer.active():
                self.scheduledEventTimer.cancel()
            self.scheduledEventTimer = None
        now = util.now()
        ready = []
        timeout = None
        for ev in self.scheduledEvents[:]:
            when, req = ev
            if when <= now:
                self.log(" event ready to fire:", req.why(), parent=e)
                self.scheduledEvents.remove(ev)
                ready.append(req)
            else:
                self.log(" still waiting %d:" % (when-now), req.why(),
                         parent=e)
                if timeout == None:
                    timeout = when
                else:
                    timeout = min(timeout, when)
        if timeout != None:
            self.log("next event is in %d seconds" % (timeout-now), parent=e)
            relative = (timeout+1) - now # wake up a trifle late
            timer = reactor.callLater(relative, self.checkScheduledEvents)
            self.scheduledEventTimer = timer
        for req in ready:
            req.fire(None) # reentrancy can happen here
        self.log("Agent.checkScheduledEvents done", parent=e)


    # other utilities

    def pingRecipient(self, recipient):
        """Ping a recipient's agent. Returns a Deferred that fires (with the
        response message) when the response comes back."""

        self.log("pinging %s" % recipient.getName())
        owner = recipient.addressbook.owner
        m = message.OutboundMessage(self)
        m.setSenderEntry(owner)
        m.setRecipientEntry(recipient)
        pingid = 123
        m.addNode('ping', pingid)
        d = self.sendRequest(m)
        d.addCallback(self._gotPong, pingid)
        return d

    def _gotPong(self, message, pingid):
        assert message.getNode("pong") == pingid
        self.log("ping response")
        return message

    # POP/IMAP proxy support
    def requestAvatar(self, avatarID, mind, *interfaces):
        if mailbox.IMailbox in interfaces:
            o = self.getOwnerByName(avatarID)
            mbox = self.getInbox(o)
            return mailbox.IMailbox, mbox, lambda: None
        assert 0

    def setProxy(self, pop3_port, imap_port, smtp_port):
        if self.pop3_port != pop3_port:
            if self.pop3_port:
                self.pop3_server.disownServiceParent()
                self.pop3_server = None
                self.pop3_port = None
            if pop3_port:
                f = mailbox.POP3Factory(self.portal)
                server = internet.TCPServer(pop3_port, f)
                server.setServiceParent(self)
                self.pop3_server = server
                self.pop3_port = pop3_port
        if imap_port:
            raise Unimplemented("IMAP proxy not yet implemented")
        if self.smtp_port != smtp_port:
            if self.smtp_port:
                self.smtp_server.disownServiceParent()
                self.smtp_server = None
                self.smtp_port = None
            if smtp_port:
                deliverer = mailbox.Deliverer(self)
                f = mailbox.SMTPFactory(deliverer)
                server = internet.TCPServer(smtp_port, f)
                server.setServiceParent(self)
                self.smtp_server = server
                self.smtp_port = smtp_port

    # plugins
    def ownerChanged(self, action, owner, lognode=None):
        return self.runPlugins("ownerChanged", action, owner, lognode)
    def entryChanged(self, action, entry, lognode=None):
        return self.runPlugins("entryChanged", action, entry, lognode)

    # startup utility methods
    def run(self):
        self.startService()
        reactor.run()
        self.stopService()

    def stop(self):
        reactor.stop()


def run():
    a = Agent()
    a.startService()
    reactor.run()
    a.stopService()

if __name__ == '__main__':
    run()
