
import os, os.path, re
from base64 import encodestring, decodestring
from StringIO import StringIO

from twisted.internet import protocol, defer, reactor
from twisted.python import log

from petmail import util
from petmail.interfaces import UndecryptableError, BadSignatureError

useBase64 = True

# for this fake crypto, the keys and keyIDs made in triples:
#  "pubkey1", "privkey1", "keyid1"
# this lets us convert from keyids to keys without actually having a database
# of key values

def assertKeyID(s):
    assert(s.startswith("keyid"))

class PublicKey(object):
    def __init__(self, agent, keyid):
        if not agent.fake:
            assert len(keyid) == 40
        self.agent = agent
        self.keyid = keyid
    def getID(self):
        return self.keyid
    def getKey(self):
        return self.agent.getKey(self.keyid)
    def encryptAndSign(self, signer, plaintext):
        assert isinstance(signer, PrivateKey)
        assert signer.agent == self.agent
        return self.agent.encryptAndSign(self, signer, plaintext)
    def encrypt(self, plaintext):
        return self.agent.encrypt(self, plaintext)
    def __repr__(self):
        return "<PublicKey %s>" % self.getID()

class PrivateKey(PublicKey):
    def sign(self, plaintext):
        return self.agent.sign(self, plaintext)
    def pubkey(self):
        return PublicKey(self.agent, self.keyid)
    def __repr__(self):
        return "<PrivateKey %s>" % self.getID()

# verify must be done directly with the agent, as we don't know who it is
# encrypted to

class FakeEncryptionAgent(object):
    """I handle encryption needs for a PETmail agent. I wrap a GnuPG home
    directory and (and keyring) and provides methods to add keys, encrypt,
    decrypt, sign, and verify messages.

    My API is in terms of keyids.
    """

    fake = True

    def __init__(self, basedir=None):
        self.pubkeys = {}
        self.privkeys = {}

    def createKeyring(self):
        """Create my basedir and populate it with the standard GnuPG files.
        This must be done once before the encryption agent is useable.
        """
        pass
    
    def generateKey(self, name, progress=None):
        key = util.generateMessageID()
        keyid = "keyid" + key
        self.privkeys[keyid] = "privkey" + key
        self.pubkeys[keyid] = "pubkey" + key
        if progress:
            progress.set_progress(1.0)
        return defer.succeed(PrivateKey(self, keyid))

#    def getID(self, pubkey):
#        assert(pubkey.startswith("pubkey"))
#        return "keyid" + pubkey[6:]

    def addPubkey(self, pubkey):
        """I take a string with an armored public key (such as would come
        out of getKey(), below), extract the key, and add it to my keyring.
        After this, the key can be used as a target for encryption.
        """
        # do a gpg --import
        assert pubkey.startswith("pubkey")
        keyid = "keyid" + pubkey[6:]
        self.pubkeys[keyid] = pubkey
        return defer.succeed(PublicKey(self, keyid))

    def getKey(self, keyid):
        """I return a string with an armored version of the given public
        key, suitable for inclusion in a message. This string has enough
        information to allow anyone who receives it to encrypt messages to
        this key's owner.
        """
        return defer.succeed(self.pubkeys[keyid])

    def encryptAndSign(self, recipient, me, plaintext):
        """Encrypt and sign the message.

        @type recipientkey: C{string}
        @param recipientkey: the keyid to which the message should be encrypted
        @type ourkey: C{string}
        @param ourkey: the keyid that should be used to sign the message
        @type plaintext: C{string}
        @param plaintext: the (binary) message to be encrypted

        @rtype: C{string}
        @returns: a (binary) encrypted message
        """

        assert isinstance(me, PrivateKey)
        assert isinstance(recipient, PublicKey)
    
        if isinstance(plaintext, StringIO):
            plaintext = plaintext.read()
    
        s = "[%s,%s,%s]" % (recipient.getID(), me.getID(), plaintext)
        if useBase64:
            s = encodestring(s)
        return defer.succeed(s)

    def encrypt(self, recipient, plaintext):
        """Encrypt the message.

        @type recipientkey: C{string}
        @param recipientkey: the keyid to which the message should be encrypted
        @type plaintext: C{string}
        @param plaintext: the (binary) message to be encrypted

        @rtype: C{string}
        @returns: a (binary) encrypted message
        """

        assert isinstance(recipient, PublicKey)
    
        if isinstance(plaintext, StringIO):
            plaintext = plaintext.read()
    
        s = "[%s,,%s]" % (recipient.getID(), plaintext)
        if useBase64:
            s = encodestring(s)
        return defer.succeed(s)


    def decrypt(self, crypttext):
        if isinstance(crypttext, StringIO):
            crypttext = crypttext.read()
        if useBase64:
            crypttext = decodestring(crypttext)
        assert(crypttext[0] == '[')
        assert(crypttext[-1] == ']')
        comma1 = crypttext.index(',')
        recipID = crypttext[1:comma1]
        if recipID != "":
            if not self.privkeys.has_key(recipID):
                why = "message is not encrypted to us: recip=%s\n" % recipID
                why += "privkeys: [%s]" % ",".join(self.privkeys.keys())
                raise UndecryptableError(why)
            recip = PrivateKey(self, recipID)
        else:
            recip = None
        comma2 = crypttext.index(',', comma1+1)
        sigID = crypttext[comma1+1:comma2]
        if not self.pubkeys.has_key(sigID):
            sigID = None
        plaintext = crypttext[comma2+1:-1]
        return defer.succeed((recip, sigID, plaintext))

    def sign(self, me, plaintext):
        """Sign the message.

        @type ourkey: C{string}
        @param ourkey: the keyid that should be used to sign the message
        @type plaintext: C{string}
        @param plaintext: the (binary) message to be encrypted

        @rtype: C{string}
        @returns: a (binary) signed message
        """

        assert isinstance(me, PrivateKey)
        if isinstance(plaintext, StringIO):
            plaintext = plaintext.read()
        s = "[,%s,%s]" % (me.getID(), plaintext)
        if useBase64:
            s = encodestring(s)
        return defer.succeed(s)

    def store(self, plaintext):
        s = "[,,%s]" % plaintext
        if useBase64:
            s = encodestring(s)
        return defer.succeed(s)

    def dumpSigned(self, crypttext):
        recip, signer, plaintext = decrypt(crypttext)
        obj = util.decode(plaintext)

        print "signed by sigID:", signer.getID()
        print "obj:", obj
        if type(obj) == type({}) and obj.get("permission"):
            print "permission:", util.decode(obj['permission'])


keygenInput = """Key-Type: DSA
Key-Length: 1024
# We are only testing syntax this time, so dry-run
Subkey-Type: ELG-E
Subkey-Length: 1024
Name-Real: %(name)s
Name-Comment: %(comment)s
Name-Email: %(email)s
"""
#Passphrase: %(passphrase)s
#%%dry-run

# We don't use GnuPGInterface. It has problems: it blocks (os.waitpid), and
# doesn't actually do any of the useful --status-fd parsing that I need. It
# seems good enough to do encryption and decryption, but it doesn't give you
# enough information to do signature verification (except what you can parse
# yourself from stdout). Apart from being a frustratingly low-level
# interface, the options are too sticky (things like a recipients list should
# be set on a per-invocation basis, not on the whole object, unless the
# object is meant to be single-use), the read/write phases are distinct
# (which I think means large messages will cause deadlock), and it can't be
# used in a producer/consumer fashion (with implications for very large
# messages, albeit issues that are probably hard to fix anyway because of
# fundamental petmail design issues).


class GPGProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, passphrase, inputtext, statusfd,
                 acceptNonzeroRC=False, progress=None):
        self.passphrase = None
        if passphrase:
            self.passphrase, self.passphrasefd = passphrase
        self.inputtext = inputtext
        self.statusfd = statusfd
        self.outputtext = ""
        self.errtext = ""
        self.status = ""
        self.rc = None
        self.deferred = defer.Deferred()
        self.acceptNonzeroRC = acceptNonzeroRC
        self.progress = progress
        self.entropyTimer = None

    def watchEntropy(self):
        if not self.progress:
            return
        if self.entropyTimer and self.entropyTimer.active():
            self.entropyTimer.cancel()
        text = open("/proc/sys/kernel/random/entropy_avail", "r").read()
        entropy = float(text)
        text = open("/proc/sys/kernel/random/poolsize", "r").read()
        poolsize = float(text) * 8
        self.progress.set_entropy(entropy / poolsize)
        self.entropyTimer = reactor.callLater(0.5, self.watchEntropy)

    def connectionMade(self):
        if self.passphrase:
            self.transport.writeToChild(self.passphrasefd, self.passphrase)
            self.transport.closeChildFD(self.passphrasefd)

        self.transport.writeToChild(0, self.inputtext) # could use Producer
        self.transport.closeChildFD(0)

    def childDataReceived(self, childFD, data):
        if childFD == 1:
            self.outputtext += data
        if childFD == 2:
            self.errtext += data
        if childFD == self.statusfd:
            if self.progress:
                self.progress.pulse_progress()
            self.status += data

    def processEnded(self, status):
        if self.entropyTimer and self.entropyTimer.active():
            self.entropyTimer.cancel()
        if status.value.signal:
            self.deferred.errback(status)
        else:
            self.rc = status.value.exitCode
            if self.rc and not self.acceptNonzeroRC:
                self.deferred.errback(status)
            else:
                self.deferred.callback((self.outputtext, self.status))

class GnuPGEncryptionAgent(object):
    fake = False
    debug = False
    def __init__(self, basedir):
        self.basedir = basedir
        # in general, we stick to 40-character fingerprints as IDs
        
        # .privkeys holds a list of our private key fingerprints
        self.privkeys = []

        self.createKeyring()

    def log(self, *args):
        log.msg(*args)

    def _getGPG(self):
        # unused. TODO: make sure all GnuPGInterface options have been
        # translated
        gpg = GnuPGInterface.GnuPG()
        gpg.options.homedir = self.basedir
        gpg.options.meta_interactive = 0
        gpg.options.extra_args.append("--no-secmem-warning")
        return gpg

    def _startGPG(self, stdin, args, acceptNonzeroRC=False, progress=None):
        # TODO: allow stdin to be a StringIO, do something Producerish
        # TODO: passphrase
        cmd = ["gpg",
               "--homedir", self.basedir,
               "--batch",
               "--no-secmem-warning",
               "--trust-model", "always",
               "--status-fd", "3",
               ]
        cmd.extend(args)
        pp = GPGProcessProtocol(None, stdin, 3, acceptNonzeroRC,
                                progress=progress)
        d = pp.deferred
        p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                 childFDs={0:"w", 1:"r", 2:"r", 3:"r"})
        if progress:
            if os.path.exists("/proc/sys/kernel/random/entropy_avail"):
                pp.watchEntropy()
        return d, pp

    def createKeyring(self):
        if not os.path.exists(self.basedir):
            os.mkdir(self.basedir)
            os.chmod(self.basedir, 0700)
        # run gpg --list-keys, which will create pubring.gpg

    def generateKey(self, name, progress=None):
        stdin = keygenInput % {'name': name,
                               'comment': "generated for PETmail",
                               'email': "petmail@thingy",
                               'passphrase': "",
                               }

        d, pp = self._startGPG(stdin, ["--gen-key"], progress=progress)

        d.addCallback(self._doneGenerateKey, pp)
        d.addErrback(self._failedGenerateKey, pp)
        
        # status has a lot of "PROGRESS primegen + 0 0" lines
        # stderr has random things like "keyring foo.gpg created"

        return d

    def _doneGenerateKey(self, results, pp):
        outputtext, status = results
        for line in status.split("\n"):
            r = re.search(r'KEY_CREATED\s+(\w+)\s+(\w+)\b', line)
            if r:
                keyid = r.group(2)
                break
        else:
            raise RuntimeError("unable to create key:" + pp.errtext)
        assert len(keyid) == 40
        self.privkeys.append(keyid)
        return PrivateKey(self, keyid)

    def _failedGenerateKey(self, failure, pp):
        log.msg("gpg --gen-key failed: " + pp.errtext)
        return failure


    def addPubkey(self, pubkey):
        # feed key into stdin of --import
        # parse stderr for 'key XX: public key "name <addr>" imported'
        # then figure out keyid, probably with --list-colons on XX
        if self.debug: self.log("addPubkey")

        d, pp = self._startGPG(pubkey, ["--import"])

        d.addCallback(self._doneAddPubkey, pp)
        d.addErrback(self._failedAddPubkey, pp)
        return d

    def _doneAddPubkey(self, results, pp):
        outputtext, status = results
        lines = status.split("\n")
        for line in lines:
            importok = "[GNUPG:] IMPORT_OK "
            if line.startswith(importok):
                fields = line[len(importok):].split()
                reason, keyid = fields
                # could look at reason to check for newkey, unchanged, newsig
                break
        else:
            raise ValueError, "Unable to import key:" + pp.errtext
        assert len(keyid) == 40
        return PublicKey(self, keyid)

    def _failedAddPubkey(self, failure, pp):
        log.msg("gpg --import failed: " + pp.errtext)
        return failure


    def _getLongID(self, otherid):
        if self.debug: self.log("_getLongID")
        d, pp = self._startGPG('',
                               ["--with-colons", "--fixed-list-mode",
                                "--with-fingerprint", "--with-fingerprint",
                                "--list-keys", "0x%s" % otherid])
        d.addCallback(self._doneGetLongID, pp)
        d.addErrback(self._failedGetLongID, pp)
        return d
    def _doneGetLongID(self, results, pp):
        keys, status = results
        stderr = pp.errtext
        longid = None
        for line in keys.split("\n"):
            fields = line.split(":")
            if fields[0] == "fpr":
                # the first 'fpr' line is the fingerprint
                longid = fields[9]
                assert len(longid) == 40
                return longid
        raise KeyError, "id '%s' not in keyring??: %s" % (otherid, stderr)
    def _failedGetLongID(self, failure, pp):
        log.msg("gpg --list-keys failed: " + pp.errtext)
        return failure


    def getKey(self, keyid):
        if self.debug: self.log("getKey")
        d, pp = self._startGPG("", ["--export", "0x%s" % keyid])
        d.addCallback(self._doneGetKey)
        d.addErrback(self._failedGetKey, pp)
        return d
    def _doneGetKey(self, results):
        outputtext, status = results
        return outputtext
    def _failedGetKey(self, failure, pp):
        log.msg("gpg --export failed: " + pp.errtext)
        return failure


    def encryptAndSign(self, recipient, me, plaintext):
        if self.debug: self.log("encryptAndSign")
        assert isinstance(me, PrivateKey)
        assert isinstance(recipient, PublicKey)
        assert isinstance(plaintext, str)

        d, pp = self._startGPG(plaintext, 
                               [
            "--local-user",  "0x%s" % me.getID(),
            "--recipient", "0x%s" % recipient.getID(),
            "--sign",
            "--encrypt",
            ])

        d.addCallback(self._doneEncryptAndSign)
        d.addErrback(self._failedEncryptAndSign, pp)
        return d

    def _doneEncryptAndSign(self, results):
        outputtext, status = results
        return outputtext
    def _failedEncryptAndSign(self, failure, pp):
        log.msg("gpg --sign --encrypt failed: " + pp.errtext)
        return failure

    def encrypt(self, recipient, plaintext):
        if self.debug: self.log("encrypt")
        assert isinstance(recipient, PublicKey)
        assert isinstance(plaintext, str)

        d, pp = self._startGPG(plaintext, 
                               [
            "--recipient", "0x%s" % recipient.getID(),
            "--encrypt",
            ])

        d.addCallback(self._doneEncrypt)
        d.addErrback(self._failedEncrypt, pp)
        return d

    def _doneEncrypt(self, results):
        outputtext, status = results
        return outputtext
    def _failedEncrypt(self, failure, pp):
        log.msg("gpg --encrypt failed: " + pp.errtext)
        return failure


    def store(self, plaintext):
        if self.debug: self.log("store")
        assert isinstance(plaintext, str)
        d, pp = self._startGPG(plaintext, [ "--store" ])
        d.addCallback(self._doneStore)
        d.addErrback(self._failedStore, pp)
        return d

    def _doneStore(self, results):
        outputtext, status = results
        return outputtext
    def _failedStore(self, failure, pp):
        log.msg("gpg --store failed: " + pp.errtext)
        return failure
        

    def decrypt(self, crypttext):
        """Decrypt the given message, returning a tuple of (recipient,
        signer, plaintext). May raise UndecryptableError if the message is
        not encrypted to one of our private keys.

        @type crypttext: C{string}
        @param crypttext: The (binary) encrypted message

        @returns: (recipient, sigID, plaintext)
                  recipient is a C{PrivateKey} to which the message was sent,
                            or None if the message was not encrypted
                  sigID is a C{string} with the ID of the key that signed the
                   message
                  plaintext is a C{string} with the (binary) contents of the
                   message
        """

        if self.debug: self.log("decrypt")
        assert isinstance(crypttext, str) # no StringIO yet
        # TODO: passphrase=""
        d, pp = self._startGPG(crypttext, ["--decrypt"],
                               acceptNonzeroRC=True)
        d.addCallback(self._doneDecrypt, pp)
        d.addErrback(self._failedDecrypt, pp)
        return d

    def _doneDecrypt(self, results, pp):
        plaintext, status = results
        stderr = pp.errtext

        signed = False
        sigid = None
        recipient = None

        for line in status.split("\n"):
            validsig = "[GNUPG:] VALIDSIG "
            if line.startswith(validsig):
                fields = line[len(validsig):].split()
                (fingerprint, sigdate, sigtime, expiretime, sigversion,
                 reserved, pubkeyalgo, hashalgo, sigclass,
                 primaryfingerprint) = fields
                signed = True
                sigid = primaryfingerprint
            errsig = "[GNUPG:] ERRSIG "
            if line.startswith(errsig):
                fields = line[len(errsig):].split()
                (keyid, pubkeyalgo, hashalgo, sigclass,
                 sigtime, rc) = fields
                signed = True
                sigid = None
            encto = "[GNUPG:] ENC_TO "
            if line.startswith(encto):
                fields = line[len(encto):].split()
                (keyid, keytype, keylength) = fields
                recipient = keyid
                # this keyid is a 16-byte id, not the full fingerprint

        if recipient:
            # TODO: we may be able to avoid this gpg call by tracking
            # the shorter keyids in a local hash.
            d = self._getLongID(recipient)
            d.addCallback(self._doneDecrypt2, sigid, plaintext)
            return d
        # not encrypted
        return (None, sigid, plaintext)

    def _doneDecrypt2(self, recipID, sigid, plaintext):
        if sigid: assert len(sigid) == 40
        assert len(recipID) == 40
        return (PrivateKey(self, recipID), sigid, plaintext)

    def _failedDecrypt(self, failure, pp):
        log.msg("gpg --decrypt failed: " + pp.errtext)
        return failure


    def sign(self, me, plaintext):
        if self.debug: self.log("sign")
        assert isinstance(me, PrivateKey)

        assert isinstance(plaintext, str) # no StringIO yet
        # TODO: passphrase=""
        d, pp = self._startGPG(plaintext,
                               ['--local-user', "0x%s" % me.getID(),
                                '--sign'])
        d.addCallback(self._doneSign, pp)
        d.addErrback(self._failedSign, pp)
        return d

    def _doneSign(self, results, pp):
        crypttext, status = results
        stderr = pp.errtext
        if not crypttext:
            raise RuntimeError("gpg --sign failed: " + stderr)
        return crypttext
    def _failedSign(self, failure, pp):
        log.msg("gpg --sign failed: " + pp.errtext)
        return failure


EncryptionAgent = GnuPGEncryptionAgent

class _Waiter:
    done = False
    def __init__(self, d):
        d.addCallback(self.finished)
        d.addErrback(self.error)
    def finished(self, result):
        self.result = result
        self.done = True
    def error(self, why):
        self.result = why
        self.done = True
        raise why

def dr(d):
    from twisted.internet import reactor
    w = _Waiter(d)
    while not w.done:
        reactor.iterate(0.01)
    return w.result

if __name__ == '__main__':
    import sys
    agent = EncryptionAgent("/tmp/homedir")
    agent.createKeyring()
    if 0:
        data = open("/tmp/warner.key","r").read()
        pubkey = dr(agent.addPubkey(data))
        print "KeyID:", pubkey
        key = dr(pubkey.getKey())
        print "key data", len(key), "bytes"
    if 1:
        key = dr(agent.generateKey("testkey"))
        print key
        

