
from twisted.python.components import Interface


class ITransport(Interface):
    """An object used to get a message blob to the receiver. Each ITransport
    handles a single message to a single recipient. It will exist until the
    message has been received safely (if an acknowledgement is expected,
    this object will stick around until after the ack has been received).

    The object is created with a specific recipient and message in mind.
    """

    def __init__(self, trecord, message):
        """
        @type trecord: C{tuple}
        @param trecord: this is the Transport Record that specifies
                        where the message should go. The first element is a
                        string that defines the class of Transport, the rest
                        are like constructor arguments.
        @type message: C{OutboundMessage}
        @param message: the Message to be sent. Through it, the Transport
                        can access .agent and .senderEntry
        """

    def send(self, encrypted):
        """Queue a (binary) encrypted message for delivery. Returns a
        Deferred which fires when the message is out of our hands.

        @type encrypted: C{string}
        @param encrypted: the encrypted message blob
        """
class IUntraceableTransport(ITransport):
    """I add the additional property that delivering the message will not
    reveal my identity or location to the recipient nor to an observer.
    """
    pass


class IInlet(Interface):
    """An object used to accept incoming messages from remote senders. The
    recipient specifies some of these and tells the rest of the world to use
    them. Inlets can be shared amongst multiple Identities in the same Agent.

    Upon receipt of a message, self.parent.receiveMessage(encrypted, source)
    should be called. This hands the message to the Agent, which will figure
    out which identity wants it. The 'source' argument describes how the
    message was received, and is a tuple of (inlet, surbid), where surbid is
    None if it is unknown or the message was not received through a SURB.

    This object should sensibly handle being serialized.
    """

    def start(self):
        """Called when the inlet is first activated (when it is added to the
        application). This is an appropriate time to open listening sockets,
        start timers to poll mailboxes, etc.
        """

    def shutdown(self):
        """Called when the inlet is being removed. Sockets should be closed,
        timers should be canceled.
        """

    def encode(self):
        """Should return a string which describes this transport to the rest
        of the world. This string will be put into the Identity Record.
        """

    def canDecryptType3(self):
        """I return True if my .decryptType3 method works, False if it
        should not be used. It is expected that the Inlet will delegate this
        to its .parent (the Agent).
        """

    def decryptType3(self, body):
        """I accept a string with the incoming message (which should have a
        --BEGIN TYPE III MESSAGE--) header somewhere inside it, and return a
        deferred which fires with a tuple of (text, idstring, surbid). I
        will only be called if self.canDecryptType3() returns True. It is
        expected that the Inlet will delegate this to its .parent (the
        Agent).
        """

class IReceivedMessage(Interface):
    """This is an inbound message."""
    pass

class IAddressBookEntry(Interface):
    pass
class IOwnerEntry(Interface):
    pass

class IPlugin(Interface):
    """When this plugin is a child of the OwnerEntry, it will be notified
    about all incoming and outgoing messages for that owner. When it is a
    child of the AddressBookEntry, it is only told about messages to/from
    that one recipient."""
    def outgoingMessage(self, msg):
        pass
    def inboundMessage(self, msg):
        pass
class IEntryPlugin(IPlugin):
    def entryChanged(self, action, entry, lognode):
        pass
class IOwnerPlugin(IPlugin):
    def entryChanged(self, action, entry, lognode):
        pass
class IAgentPlugin(IPlugin):
    def ownerChanged(self, action, owner, lognode):
        pass
    def entryChanged(self, action, entry, lognode):
        pass


class IIDRecord(Interface):
    pass
class ITicketAgent(Interface):
    """When an outgoing message needs a ticket (msg.requirements is
    non-empty), these agents are asked to satisfy these requirements. This
    usually involves:

     locate and talk to a suitable Ticket Vendor
     presenting their challenge to the user
     conveying the user's response back to the vendor
     accept the vendor's ticket
     add the ticket to the message
     remove the msg.requirements entry
    """

    def handleRequirements(self, msg):
        """Try to satisfy msg.requirements. Should remove all requirements
        thus satisfied before returning.
        """

class IOwnerPermission(Interface):
    """I describe someone else's permission to send a message to me. I am
    held by the receiver.
    """

class IPermissionSlip(Interface):
    """I describe my permission to send a message to someone else. I am held
    by the sender."""

class IProcessor(Interface):
    """I examine messages. Each agent has a list of Processors. Many are
    built-in, but others can come from plugins. Each Processor lives in a
    single agent.
    """

    def setAgent(self, agent):
        """Bind this Processor to the agent it lives in."""

    def getDependencies(self):
        """Return a list of strings, each of which represents a dependency
        of this processor. The Processor will not be run until the
        dependency is fulfilled. Note that it may be fulfilled negatively
        (indicating the required data will never be available)."""

    def getProvides(self):
        """Return a list of strings which indicate all the features which
        this processor will eventually provide (by calling message.provide
        with the feature name).
        """

    def configureSender(self, sender):
        """Hazy: Create a dialog that can be used to set sender-global state
        for this Processor."""

    def configureRecipient(self, recipient):
        """Hazy: add to the address-book page for a single recipient,
        allowing configuration of per-recipient Processor state."""

class IOutboundProcessor(IProcessor):
    """I sign off on outbound messages, and have the ability to modify,
    delay, or reject each one. In general, I will perform a lookup and put
    the results into header nodes inside the Message.

    I will not be asked until all my dependencies are met.

    I will only be asked about the message once. Once all Processors agree,
    the message will be delivered.

    well-known dependencies:
     recipientRecord: fulfilled when message.recipientRecord is either
                      non-None or will never become non-None (?). For messages
                      which are sent using the address book, messages sent to
                      an explicit IDRecord, replies, and most control messages,
                      this will be pre-fulfilled. For messages sent to an
                      Address, this will be fulfilled by the AddressLookup
                      processor.
                      
     recipientEntry: fulfilled when message.recipientEntry is either
                     a valid AddressBookEntry or will never be non-None. For
                     messages sent using the address book, it will be
                     pre-fulfilled. For entries sent to an Address, it will
                     be fulfilled by AddressLookup when it creates the
                     AddressBookEntry.

                     recipientEntry is dependent upon recipientRecord.

     
    """    

    def checkOutbound(self, message):
        """Do stuff to an outbound Message.

        I will not be called until all the features in my .getDependencies
        list are provided (probably by other processors).

        I return True if I am willing to let this message be delivered now.

        I return a Deferred if I have initiated a lookup or other
        asyncronous processing. I will fire the Deferred with 'True' when
        that processing is finished and I am willing to let the message
        pass. I will provide all dependencies that I can before firing the
        deferred.

        If I will refuse to let the message through no matter what the other
        Processors eventually say, I will raise a Unsendable exception. I
        may also so this by firing the deferred's errback.
        """

class IInboundProcessor(IProcessor):
    """For inbound messages, I will look at the Message and extract data
    from it. This will continue until all Processors have indicated that
    they have no more data to extract."""

    def checkInbound(self, receivedMessage):
        """Do stuff to an inbound ReceivedMessage.

        I return True when I am done extracting information from the
        message. Once I return True, I will not be asked about the message
        again.

        I return False if I am not yet finished extracting information
        (perhaps I am waiting for some other Processor to give me a place to
        put it).
        """

class IRequest(Interface):
    pass

class Unsendable(Exception):
    pass
class NotPermitted(Exception):
    """This incoming message was rejected"""
    pass

class UndecryptableError(Exception):
    """The message was not encrypted to our private key."""
class BadSignatureError(Exception):
    """The message had an invalid signature. Note that this does not mean
    the message was signed by an unknown public key. Such signatures cannot
    be verified, but are identified by returning None as a sigID rather than
    raising a BadSignatureError."""

class CatOutOfBagError(Exception):
    """The Identity has already used non-secure means, and can no longer be
    safely made pseudonymous.
    """

# remote interfaces: that is, interfaces which provide functionality to
# clients of the Petmail Agent (clients which live in an MUA). These
# Interfaces are usable by code which runs in the same process as the Agent,
# but it is expected that PB/XML-RPC/whatever interfaces will provide
# corresponding functionality.

class IProperties(Interface):
    properties = {} # maps property-name to property-type
    ro_properties = {}

    def getProperties():
        """Return the property map."""

    def getReadonlyProperties():
        """Return the readonly properties."""

    def getProperty(name):
        pass

    def setProperty(name, value):
        pass


class IAgent(IProperties):
    """I am a remote representation of an agent.Agent object. This is the
    root object, the first reference given to an MUA-side client."""

    # agent-level preferences
    properties = {}

    def newIdentity(): # DEFERRED
        # TODO: progress= argument?
        """Create a new OwnerEntry, to be added to this Agent. This method
        returns an IRemoteIdentity object, which can be manipulated further.
        It will not actually be added to the Agent (and therefore cannot be
        used to send messages) until this object is passed to my
        .addIdentity() method."""

    def addIdentity(identity): # DEFERRED
        """Add an IRemoteOwnerEntry object (previously obtained from
        .newIdentity) to the Agent. Once added, it can be used to send and
        receive messages. The Identity will be published immediately."""

    def deleteIdentity(identity):
        """Delete this Identity. Heavy."""

    def getIdentity(id):
        """Look up an OwnerEntry by ID Number, and return its
        IRemoteIdentity object, which can be used to further manipulate the
        OwnerEntry."""

    def getAllIdentities():
        """Return a list of IRemoteIdentity objects, representing all
        Identities known by the Agent."""

    def newPermission(ptype):
        """Return a new Permission object of the given type. This object can
        be manipulated by setting various properties, then added to an
        Identity or a Correspondent."""

    def newInlet(itype, **args):
        """Create a new Inlet object of the given type. This is suitable for
        use in IIdentity.setInlets() ."""

    def newMessage(mtype="rfc822"):
        """Create and return a new Message object (IOutboundMessage) which
        can be used to send a message of the given type. The Message must be
        filled in with things like recipients and contents, then can be sent
        with its .send method."""

    def getPlugins():
        """Return control objects for all Agent Plugins which provide remote
        control over their state."""


class IIdentity(IProperties):
    """I represent an owner.OwnerEntry object."""

    properties = { "myName": str,
                   "name": str, # TODO: improve someday after some use cases
                   "addresses": None, # ListOf(String)
                   "expirationTime": int,
                   }
    ro_properties = { "id": str,
                      "pseudonymous": bool,
                      }

    # .genericPermission object
    # assorted plugins: use formless

    def getInlets():
        """Return a list of Inlets used by this Identity. Each Inlet
        implements a subclass of IInlet2, with properties that depend upon
        the particular type of inlet."""
    def setInlets(inlets): # DEFERRED
        """Take a list of Inlet objects (each provided by IAgent.newInlet)
        and start using it for this Identity. If the new Inlets are
        equivalent to the old ones, nothing will be changed."""


    def getGenericPermission(ptype):
        """Retrieve the InboundPermission object for the given type of
        permission. This object can be manipulated in-place. This method
        will return None if this permission has not been created (which is
        more likely for Correspondents than for Identities)."""

    def setGenericPermission(ptype, perm):
        """Starting using a Permission object (as obtained from
        Agent.newPermission) as the Generic Permission of the given type."""
        # TODO: do not share Permission objects: they also track per-user
        # status (like when the last message was received).

    def getAddressBook():
        pass

    def getArmored():
        """Return a string with the armored form of the Identity."""

    def getPlugins():
        """Return control objects for all Owner Plugins which provide remote
        control over their state."""

    def subscribeMailbox(watcher, mtype="rfc822"):
        """Register a watcher which will be called as watcher(mtype, msg)
        with accepted inbound messages of the given type. mtype=None means to
        accept all message types. If there are any messages queued, they will
        be delivered immediately. Each one will implement IInboundMessage.
        Each watcher function may only be registered once."""
    def unsubscribeMailbox(watcher):
        """Unregister a watcher that was previously registered with
        subscribeMailbox."""

class IInlet2(IProperties):
    def getType():
        """Return a string indicating the type of the Inlet."""

class IInlet2_Dummy(IInlet2):
    # type = Dummy
    ro_properties = { "name": str }

class IInlet2_PB(IInlet2):
    ro_properties = { "hostname": str, "portnum": int}

class IInlet2_Maildir(IInlet2):
    ro_properties = { "basedir": str, "address": str}
class IInlet2_MailSocket(IInlet2):
    ro_properties = { "portnum": int, "address": str}

class IInlet2_IMAP(IInlet2):
    ro_properties = { "hostname": str, "portnum": int,
                      "username": str, "password": str,
                      "pollinterval": int,
                      "folder": str, # could be empty string
                      "address": str}
    def check():
        """Do an immediate poll."""

class IInlet2_POP3(IInlet2):
    ro_properties = { "hostname": str, "portnum": int,
                      "username": str, "password": str,
                      "pollinterval": int,
                      "address": str}
    def check():
        """Do an immediate poll."""

class IInlet2_Jabber(IInlet2):
    ro_properties = { "server": str, "username": str, "passwd": str,
                      "host": str, "port": int,
                      }


class IInboundPermission_rfc822(IProperties):
    properties = { "minIMG": int, # or None for "unlimited"
                   "maxsize": int, # or None for "unlimited"
                   "acceptHTML": bool,
                   "canGrantPermission": bool,
                   }
    # TODO: pedigree

    def listTicketTakerURLs():
        """Return a list of URLs for the accepted Ticket Takers."""
    def addTicketTaker(taker):
        """Add a new TicketTaker. REQUIRES LOCAL ACCESS (to create the
        'taker' object)."""

class IOutboundPermission_rfc822(IProperties):
    ro_properties = { "minIMG": int,
                      "maxsize": int,
                      "acceptHTML": bool,
                      "canGrantPermission": bool,
                      }
    def listTicketTakerURLs():
        """Return a list of URLs for the accepted Ticket Takers."""

class IAddressBook(Interface):
    """I am a remote representation of an addressbook.AddressBook object.
    The address book contains Correspondents."""

    def getOwner():
        """Return the IRemoteIdentity object of the Owner of this
        AddressBook."""

    def subscribe(watcher):
        """Subscribe a watcher. This callable will be invoked (with a single
        argument: this IAddressBook instance) when a Correspondent is added,
        removed, or modified."""
    def unsubscribe(watcher):
        """Remove the watcher that was added by subscribe()."""

    # attempt to add an entry
    def addAddress(addr): # DEFERRED
        """Use an Address Server to retrieve the IDRecord pointed to by this
        address, then add it as a correspondent. Fires with the
        ICorrespondent which has just been added."""

    def addArmoredIDRecord(record): # DEFERRED
        """Given a text (ascii-armored) form of an IDRecord, add it."""

    def deleteCorrespondent(correspondent): # DEFERRED
        """Remove the given correspondent, revoking any Specific Permission
        they were previously granted."""

    # find an existing entry
    def lookupID(id):
        """Lookup an ID Number, return the IRemoteEntry object associated
        with it (or None)."""

    def lookupName(name):
        """Lookup a pet name."""

    def lookupAddress(addr):
        """Lookup by address."""

    def countEntries():
        """Return the number of entries currently in the addressbook."""

class ICorrespondent(IProperties):
    """I am a remote representation of an addressbook.AddressBookEntry
    object."""

    properties = { "petname": str, # or None
                   "vouchFor": bool,
                   "grantReturnPermission": bool,
                   }
    ro_properties = { "id": str,
                      "name": str, # TODO: match Identity.name
                      "addresses": [str],
                      "transports": [str], # informational
                      }

    def getAddressBook():
        """Return the IRemoteAddressBook object of the book this entry lives
        inside."""

    # plugins
    def getPlugins():
        """Return control objects for all Entry Plugins which provide remote
        control over their state."""

    # idRecord data (read-only)
    
    def getOurPermission(ptype):
        """Return the (read-only) OutboundPermission object that we must
        obey when sending messages to this recipient. This will either be
        our Specific Permission, or the Generic Permission published by the
        currespondent."""

    def getOurPermissionType(ptype):
        """Return 'specific' or 'generic' depending upon what kind of
        permission we have available to us."""

    def subscribe(watcher):
        """Subscribe a callable to be invoked (with this ICorrespondent as
        the sole argument) whenever one of its attributes changes (perhaps
        because the remote end has sent an update message). This includes
        changes to any of our Permission records."""
    def unsubscribe(watcher):
        """Remove a callable that was previously subscribed."""


    def setTheirPermission(ptype, perm):
        """Starting using a Permission object (as obtained from
        Agent.newPermission) as this correspondent's Specific Permission of
        the given type."""
        # TODO: do not share Permission objects: they also track per-user
        # status (like when the last message was received).

    def revokeTheirPermission(ptype):
        """Completely revoke the correspondent's inbound Permission."""

    def getTheirPermission(ptype):
        """Return the (read-write) InboundPermission object that the
        recipient must obey when sending messages to us. This object can be
        manipulated in-place by getting/setting various properties. Returns
        None if we have not granted them Specific Permission."""


class IOutboundMessage(Interface):
    """This is an outbound message."""
    def setSender(sender):
        """Set the sender of this message, which must be an IIdentity as
        returned by agent.getIdentity or agent.getAllIdentities ."""

    def send(): # DEFERRED
        """Send the message. Once the message has been sent, it may no
        longer be manipulated."""

class IOutboundMessage_rfc822(IOutboundMessage):
    """I represent a single outbound rfc822 message."""

    def addRecipient_petname(name):
        """Add a recipient by looking up the name in the Owner's
        addressbook."""

    def addRecipient_id(id):
        """Add a recipient by ID number."""

    def addRecipient_correspondent(correspondent):
        """Add a recipient by providing an ICorrespondent instance."""

    def addReference(correspondent):
        """Add a reference to the given ICorrespondent."""

    def addRFC822(msg):
        # TODO: rename to 'setRFC822'
        """Set the message headers and body from a string containing both."""

class IInboundMessage(Interface):
    def getFromName(): # DEFERRED
        """Return a string describing the sender of the message. This will be
        the pet name that has been assigned by the recipient, or the sender's
        claimed name in quotes."""

    def getRecipient():
        """Return the IIdentity object for the recipient of this message."""

    def getSender():
        """Return the ICorrespondent entry for the sender of this message, if
        they are known. If it is from an unknown sender (they are not in the
        recipient's address book), or if it is anonymous, this will return
        None."""

    def getSenderID():
        """Return the ID number of the sender of this message, or None if the
        message was anonymous."""

    def getSenderIDRecord():
        """Return the (ascii-armored) ID Record of the sender of this
        message. If it is not available, return None."""

    def canAddSender():
        """Return True if the sender of this message is not already in the
        address book AND if there is enough information to add them. This
        should be used to enable/provide an 'Add This Sender' button to the
        displayed message."""

    def addSender(): # DEFERRED
        """Add the sender of this message to the recipient's address book.
        Returns the ICorrespondent entry that results."""

    def canGrantInboundPermission():
        """Returns True if the sender of this message is known to the
        recipient but does not currently have inbound permission. This should
        be used to enable a 'Grant Inbound Permission' button on the
        displayed message."""

    def grantInboundPermission():
        """Grant the sender of this message permission to send more."""

    def getReason():
        """Return a string that indicates what kind of permission the
        sender used to get this message delivered."""

class IInboundMessage_rfc822(IInboundMessage):

    def canAddReferences():
        """Returns a list of References (TODO: format? dict of some sort)
        which could usefully be added to the recipient's address book. This
        should be used to enable/provide some 'Add This Reference' buttons to
        the displayed message."""

    def addReference(ref): # DEFERRED
        """Add a referenced identity to the recipient's address book. Returns
        the ICorrespondent entry that results."""

    def canReply():
        """Return True if this message can be replied to (i.e.: it is not
        anonymous)."""

    def reply():
        """Create a new message object, just like agent.newMessage(), with
        the sender and destination already filled in."""

    def getMessage():
        """Return a string with the header and body of the enclosed rfc822
        message."""
