
import os
from petmail import util

class LogBundle(object):

    def __init__(self):
        self.logs = {}
    def addLog(self, log):
        self.logs[log.name] = log

class Entry(object):
    """This is the object returned by methods like Log.getItem() . At the
    moment it is also the type of objects stored in the event log, but in the
    future Log will probably use a more compact data type instead.
    """
    children = None

    def __init__(self, timestamp, message):
        self.time = timestamp
        self.message = message # tuple of args

    def addChild(self, entry):
        if self.children is None:
            self.children = [entry]
        else:
            self.children.append(entry)

    def getMessage(self):
        return self.message
    def getChildren(self):
        if self.children is None:
            return []
        return self.children

    def dumpTree(self, indent=0):
        print (" " * indent) + ",".join(self.message)
        if self.children:
            for child in self.children:
                child.dumpTree(indent+2)

class ChildLog(object):
    def __init__(self, root, node):
        self.root = root
        self.node = node
    def log(self, *args):
        return self.root.log(node=self.node, *args)

class Log(object):
    """A size-limited record of events.

    Each event has a priority. Each priority class has a separate size limit,
    so that low-priority (minor or routine) events do not last as long as
    high-priority (significant or unusual) ones. The events are presented in
    chronological order, but there may be gaps where some old low-priority
    events were dropped.

    Each event is a tree. Each node has some text (or other data types which
    can be rendered at output time). The log presentation code is expected to
    show only the top-level node at first, allowing the user to expand that
    node to see more detail if desired.
    
    """

    def __init__(self, name, numClasses=1, maxItems=100):
        self.name = name
        self.items = [ [] for i in range(numClasses) ]
        self.allItems = []
        self.observers = []
        if type(maxItems) in (tuple, list):
            self.maxItems = maxItems
        else:
            self.maxItems = (maxItems,) * numClasses
        if os.environ.get("LOGMEBABY"):
            self.observers.append(self.printEvent)

    def __getstate__(self):
        d = self.__dict__.copy()
        d['observers'] = []
        return d

    def addObserver(self, observer):
        """Add a log observer. The callable will be invoked as follows:

        observer(lognode, pclass=0, parent=None)

        If parent!=None then this is supplemental information, and a strictly
        chronological log should probably not just simply append it to a
        list.

        The observer can use lognode.getMessage() to get the tuple which
        contains the actual logged information.
        """
        self.observers.append(observer)
    def removeObserver(self, observer):
        self.observers.remove(observer)

    def printEvent(self, lognode, pclass, parent):
        if parent != None:
            print "CHILD", ' '.join(map(str, lognode.getMessage()))
            return
        print ' '.join(map(str, lognode.getMessage()))

    def log(self, *args, **kwargs):
        pclass = kwargs.get("pclass", 0)
        node = kwargs.get("node", None)
        if node:
            assert isinstance(node, Entry)
        entry = Entry(util.now(), args)
        if node:
            node.addChild(entry)
        else:
            self.items[pclass].append(entry)
            self.allItems.append(entry)
            if len(self.items[pclass]) > self.maxItems[pclass]:
                dropped = self.items[pclass].pop(0)
                self.allItems.remove(dropped)
        for o in self.observers:
            o(entry, pclass, node)
        return ChildLog(self, entry)

    def addEvent(self, eventDict):
        """This method can be given to twisted.python.log"""
        # doing t.p.log.msg(*args, **kwargs) results in an eventDict with
        #  ['message'] = *args
        #  ['time'] = timestamp
        #  and other keys merged in from kwargs and a "context" dict
        #
        # if ['isError'] is set, then there might be a ['failure'] key
        #
        # we'll just behave kind of like the standard twisted
        # FileLogObserver. This code follows FLO.emit()

        if eventDict['message']:
            text = ''.join(map(str, eventDict['message']))
        elif eventDict['isError'] and eventDict.has_key('failure'):
            text = eventDict['failure'].getTraceback()
        else:
            text = "unloggable t.p.log event"
        pclass = 0
        self.log(text, pclass=pclass)

    def getItems(self, pclass=None):
        if pclass == None:
            return self.allItems
        else:
            return self.items[pclass]

    def getItem(self, which, pclass=None):
        return self.getItems(pclass)[which]

    def getMessage(self, which, pclass=None):
        return self.getItem(which, pclass).getMessage()

    def dumpTree(self):
        for i in self.allItems:
            i.dumpTree()

class NullLog:
    # behave just like Log except throw out everything
    def log(self, *args, **kwargs):
        return self
    add = log

