
import StringIO

from twisted.internet import protocol, defer
from twisted.protocols import pop3, smtp
from twisted.cred import portal, checkers, credentials, error
from twisted.python import log, failure
IMailbox = pop3.IMailbox

from petmail import message

class Mailbox:
    __implements__ = pop3.IMailbox

    def __init__(self, owner):
        self.owner = owner
        self.messages = []
        self.deleted = []
        self.observers = []

    def __getstate__(self):
        d = self.__dict__.copy()
        d['observers'] = []
        return d

    def notifyOnChange(self, observer):
        self.observers.append(observer)
    def dontNotifyOnChange(self, observer):
        self.observers.remove(observer)

    def addMessage(self, m):
        self.messages.append(m)
        for o in self.observers:
            o(self)

    def _getLength(self, i):
        if i >= len(self.messages):
            return 0
        if i in self.deleted:
            return 0
        m = self.messages[i]
        if m:
            return len(m.getNode('rfc822')['body'])
        return 0

    def listMessages(self, i=None):
        if i is None:
            return map(self._getLength, range(len(self.messages)))
        return self._getLength(i)

    def getMessage(self, i):
        if i in self.deleted:
            body = ""
        else:
            body = self.messages[i].getNode('rfc822')['body']
        return StringIO.StringIO(body)

    def getUidl(self, i):
        return i

    def deleteMessage(self, i):
        self.deleted.append(i)

    def undeleteMessages(self):
        self.deleted = []
        
    def sync(self):
        newmessages = []
        for i in range(len(self.messages)):
            if i not in self.deleted and self.messages[i]:
                newmessages.append(self.messages[i])
        self.messages = newmessages
        self.deleted = []
        for o in self.observers:
            o(self)

class MyChecker:
    __implements__ = checkers.ICredentialsChecker
    credentialInterfaces = (credentials.IUsernamePassword,
        credentials.IUsernameHashedPassword)

    def __init__(self, agent):
        self.agent = agent

    def _cbPasswordMatch(self, matched, username):
        if matched:
            return username
        else:
            return failure.Failure(error.UnauthorizedLogin())

    def requestAvatarId(self, credentials):
        try:
            owner = self.agent.getOwnerByName(credentials.username)
        except KeyError:
            return failure.Failure(error.UnauthorizedLogin())
        password = "secret" # TODO: get from OwnerEntry
        d = defer.maybeDeferred(credentials.checkPassword, password)
        d.addCallback(self._cbPasswordMatch, credentials.username)
        return d

class POP3Factory(protocol.Factory):
    protocol = pop3.POP3
    def __init__(self, portal):
        self.portal = portal

    def buildProtocol(self, addr):
        log.msg("POP3 connection from", addr)
        p = protocol.Factory.buildProtocol(self, addr)
        p.portal = self.portal
        return p

def makePortal(agent):
    p = portal.Portal(agent)
    p.registerChecker(MyChecker(agent))
    return p

# SMTP proxy stuff

class Message:
    __implements__ = smtp.IMessage,

    def __init__(self, user):
        owner = user.orig.owner
        agent = owner.parent
        m = message.OutboundMessage(agent)
        m.setSenderEntry(owner)
        print "dest", user.dest, type(user.dest)
        m.setMessageRecipient(str(user.dest), owner)
        self.message = m
        self.body = ""

    def lineReceived(self, line):
        self.body += line + "\n"

    def eomReceived(self):
        self.message.addRFC822(self.body)
        d = self.message.send()
        return defer.succeed("righto")

    def connectionLost(self):
        print "connectionLost, discarding"

class Deliverer:
    __implements__ = smtp.IMessageDelivery,

    def __init__(self, agent):
        self.agent = agent

    def validateFrom(self, helo, origin):
        # find an Owner that corresponds to the origin
        user = origin.local
        try:
            owner = self.agent.getOwnerByName(user)
        except KeyError:
            raise smtp.SMTPBadSender(origin, resp="No such Petmail Identity")
        origin.owner = owner
        return origin

    def validateTo(self, user):
        # just accept everything for now
        # can raise smtp.SMTPBadRcpt if the dest addr is not accepted
        creator = lambda: Message(user)
        return defer.succeed(creator)

    def receivedHeader(self, helo, origin, recipients):
        return "Received: by petmail agent" # the \n is added elsewhere

class SMTPFactory(smtp.SMTPFactory):
    def __init__(self, delivery):
        smtp.SMTPFactory.__init__(self)
        self.delivery = delivery
    def buildProtocol(self, addr):
        log.msg("SMTP connection from", addr)
        p = smtp.SMTPFactory.buildProtocol(self, addr)
        p.delivery = self.delivery
        return p
