
from twisted.python.components import implements
from twisted.internet import defer
from twisted.python import failure, log

from petmail import util, crypto, interfaces
from petmail.idrecord import IDRecord
from petmail.log import NullLog
NotPermitted = interfaces.NotPermitted
Unsendable = interfaces.Unsendable

class InvalidMessage(Exception):
    """The message was mangled somehow"""

class BaseMessage(object):
    logPriority = 0
    debug = False
    lognode = None

    def __init__(self, agent):
        self.agent = agent
        # delivery attributes
        self.when = util.now()  # the nominal creation time of the message
        self.eventlog = [] # log of what processors have done
        self.doDump = True
        self.provides = []
        # recipientEntry, recipientRecord, grantedPermission
        # senderEntry, acceptedRFC822
        self.senderEntry = None

    def startProcessing(self):
        if self.agent.debugProcessing:
            print "startProcessing"
        self.processors = self.getProcessors()
        self.processing = []
        self.done = False
        self.failed = False
        d = self.processingDeferred = defer.Deferred()
        self.process()
        return d
        
    def process(self):
        """Do whatever can be done to the message."""
        # this function will not raise Unsendable exceptions. Instead, it
        # (or a callback) will run self.processingFailed
        debug = self.agent.debugProcessing
        if debug:
            text = "process:\n"
            text += " processors %s\n" % self.processors
            text += " provides %s\n" % self.provides
            text += " processing %s\n" % self.processing
            self.log(text)
        if not self.processors and not self.processing:
            if debug:
                self.log("all done")
            return self.processingComplete()
        runnable = self.findRunnableProcesses()
        if not runnable and not self.processing:
            text = "no processors are runnable"
            text += " (processors: %s)" % self.processors
            text += " (provides: %s)" % ",".join(self.provides)
            text += " (processing: %s)" % self.processing
            if debug:
                self.log(text)
                print " no processors are runnable"
            f = failure.Failure
            why = f(interfaces.Unsendable(text))
            return self.processingFailed(why)

        if debug and not runnable:
            self.log("weird, nothing to do")
            print " weird, nothing to do"
        self.processing.extend(runnable)
        for p in runnable:
            lognode = self.log(p.__class__.__name__)
            d = defer.maybeDeferred(p.checkMessage, self, lognode)
            if self.debug:
                self.log("%s gave %s" % (p, d))
            d.addCallback(self.processorComplete, p)
            d.addErrback(self.processorFailed, p)
        if debug:
            self.log(" done")

    def findRunnableProcesses(self):
        # which processors can be run? Removes them from self.processors
        runnable = []
        need = []
        for p in self.processors[:]:
            canRun = True
            for d in p.getDependencies():
                if not d in self.provides:
                    if d not in need: need.append(d)
                    canRun = False
                    break
            needToRun = False
            # empty provides means it must always be run
            processorProvides = p.getProvides()
            if not processorProvides:
                needToRun = True
            else:
                for f in processorProvides:
                    if not f in self.provides:
                        needToRun = True
            if not needToRun:
                if self.agent.debugProcessing:
                    print "don't need to run", p
                self.processors.remove(p)
            elif canRun:
                self.processors.remove(p)
                runnable.append(p)
        if self.agent.debugProcessing:
            self.log("runnable: %s" % ",".join(map(str, runnable)))
            print " runnable:", runnable
        return runnable

    def processorComplete(self, res, processor):
        if self.agent.debugProcessing:
            self.log("processorComplete", processor)
            print "processorComplete", processor
        self.processing.remove(processor)
        if not self.failed:
            self.process()
    def processorFailed(self, why, processor):
        if self.agent.debugProcessing:
            self.log("processorFailed: %s" % processor)
            self.log("failure:", why)
            print "processorFailed", processor, why
        self.processing.remove(processor)
        self.processingFailed(why)

    def processingComplete(self):
        # this should only be called once, when the last processor succeeds.
        assert not self.done
        self.done = True
        self.processingDeferred.callback(self)
    def processingFailed(self, why):
        # this may be called multiple times, once per failing processor
        log.msg("processing step failed:", why)
        if self.failed:
            if self.agent.debugProcessing:
                print "processingFailed >1 times"
            return
        self.done = self.failed = True
        self.processingDeferred.errback(why)

    def provide(self, which):
        if which not in self.provides:
            self.provides.append(which)

    def hasNode(self, name):
        return self.headers.has_key(name)
    def getNode(self, name, default=None):
        return self.headers.get(name, default)

    def startLogging(self, msg):
        if hasattr(self.agent, "eventLog"):
            eventLog = self.agent.eventLog
            self.lognode = eventLog.log(msg, pclass=self.logPriority)
        else:
            self.lognode = NullLog()

    def log(self, *args, **kwargs):
        lognode = kwargs.get("parent")
        if not lognode:
            lognode = self.lognode
        return lognode.log(*args)

    def stopLogging(self):
        self.lognode.log("logging stopped, message complete")


class OutboundMessage(BaseMessage):
    """This object represents a single message being sent to some individual
    recipient. Its life-cycle is as follows:

     1: Created when the user hits the 'Send' button on their MUA,
        containing an rfc822-formatted header+body pair.
     2: The recipient's AddressBookEntry is looked up.
     3: .encode() is called with the AddressBookEntry. This returns a string
        which should be given to the transport layer for delivery to the
        recipient.
     4: The Message is stored in a list of outstanding messages (indexed by
        messageID). An ack from the far end will remove it from the list. If
        it stays on the list for too long (5 days), the user is notified
        that the message did not make it through.


    Useful attributes of this object:

    .senderEntry: the OwnerEntry which is sending the message

    .anonymous: True for anonymous (unsigned) messages

    .recipientEntry: at the start of processing, this is only set if the
                     user is writing to a correspondent in their
                     addressbook. After the AddressLookup processor is run,
                     it will be set (and the entry added to the user's
                     AddressBook) if an rfc822 message is being sent to
                     them.

    .recipientRecord: Set as a destination (with .setRecipientRecord), or
                      when agent.sendMessageToAddress turns the address into
                      an idRecord. This is what we actually need to deliver
                      the message.
    
    .transport: an ITransport object capable of sending this message

    .provides is a list of things that have been done to this message:

    recipient: .recipientRecord and .recipientEntry have been populated
    grantedPermission: return permission has been granted

    """
    __implements__ = (interfaces.IOutboundMessage,
                      interfaces.IOutboundMessage_rfc822)
    recipientRecord = None
    recipientEntry = None
    transport = None # the transport used to send the message
    anonymous = False

    def __init__(self, agent):
        BaseMessage.__init__(self, agent)
        self.sent = None # the time this message was sent. TODO: set it
        self.headers = {} # indexed by name
        self.references = []
        self.startLogging("Outbound Message created")
        self.messageID = util.generateMessageID()
        self.addNode("messageID", self.messageID)


    def setRecipientRecord(self, r):
        assert implements(r, interfaces.IIDRecord)
        self.recipientRecord = r
        self.addNode("recipientID", r.ID)
        self.provide("recipient")

    def setRecipientEntry(self, e):
        assert implements(e, interfaces.IAddressBookEntry)
        self.recipientEntry = e
        self.setRecipientRecord(e.idRecord)
        self.provide("recipient")

    def setSenderEntry(self, e, anonymous=False):
        # TODO: deprecate
        assert implements(e, interfaces.IOwnerEntry), e
        self.senderEntry = e
        self.anonymous = anonymous

    # IOutboundMessage

    def setSender(self, sender):
        self.setSenderEntry(sender)

    def send(self):
        return self.agent.sendMessage(self)

    # IOutboundMessage_rfc822

    def addRecipient_petname(self, name):
        assert self.senderEntry
        book = self.senderEntry.getAddressBook()
        recip = book.lookupName(name)
        assert recip
        self.setRecipientEntry(recip)

    def addRecipient_id(self, ID):
        assert self.senderEntry
        book = self.senderEntry.getAddressBook()
        recip = book.lookupID(ID)
        assert recip
        self.setRecipientEntry(recip)

    def addRecipient_correspondent(self, correspondent):
        self.setRecipientEntry(correspondent)

    def addReference(self, entry):
        """Add a reference to the given AddressBookEntry. This allows the
        recipient to add the referenced Identity to their own address
        book."""

        assert implements(entry, interfaces.IAddressBookEntry)
        self.references.append(entry)

    def addRFC822(self, body): #, us, recipient, body, addrbook=None):
        """Add an RFC822 message section to the Message."""

        # construct the header
        header = {}
        #header['bodyHash'] = util.sha_hash(body)
        #header['bodyLen'] = len(body)
        ##header['senderID'] = us.ID
        ##header['messageID'] = util.generateMessageID()
        ##header['timestamp'] = self.when
        header['tickets'] = []
        header['body'] = body

        self.addNode("rfc822", header)

    # internal stuff

    def getProcessors(self):
        return self.agent.getOutboundProcessors()

    def logOutboundMessage(self):
        assert self.senderEntry
        if self.senderEntry:
            self.log(".senderEntry is %s" % self.senderEntry.getName())
        self.log(".anonymous is %s" % self.anonymous)
        if self.recipientEntry:
            self.log(".recipientEntry is %s" % self.recipientEntry.getName())
        elif self.recipientRecord:
            self.log(".recipientRecord is %s" % self.recipientRecord.name)
        else:
            self.log("I wonder who this is going to")

    def startProcessing(self):
        self.logOutboundMessage()
        d = BaseMessage.startProcessing(self)
        d.addCallback(self.handlePlugins)
        d.addCallback(lambda res: self)
        return d

    def handlePlugins(self, res):
        # give all outbound plugins a look at the message
        d = defer.succeed(None)
        if self.recipientEntry:
            recip = self.recipientEntry
            d.addCallback(recip.runPluginsD, "outboundMessage", self)
        if self.senderEntry:
            sender = self.senderEntry
            d.addCallback(sender.runPluginsD, "outboundMessage", self)
        if self.agent:
            d.addCallback(self.agent.runPluginsD, "outboundMessage", self)
        return d

    def addNode(self, name, value):
        assert (name not in self.headers)
        self.headers[name] = value

    def addTicket(self, ticket):
        # ticket is a signed blob
        assert(self.headers['rfc822'])
        self.headers['rfc822']['tickets'].append(ticket)

    def encodeHeaders(self):
        return util.encode(self.headers)

    def deliver(self):
        # we expect that mailing list software will sign once (and not
        # encrypt), then send the same message body to each subscriber with
        # a different transport. Therefore we encrypt before delivery here,
        # anticipating .recipientRecord becoming a list or something.
        # However, this is at odds with, say, a SURBSupply that gets drained
        # in the course of preparing the Transport and wants to tell the
        # other side to send more by adding a node to the already-encrypted
        # message. (on the other hand it isn't clear that this is useful,
        # the receiving end isn't going to send more SURBs until it sees the
        # old ones are used).

        self.log("Delivering message")
        self.log("headers: %s" % self.headers.keys())
        if self.anonymous:
            self.log("message is anonymous")
            us = None
        else:
            us = self.senderEntry.privkey
        recipient = self.recipientRecord.pubkey
        d = self.encrypt(us, recipient)
        d.addCallback(self._didEncrypt)
        return d

    def _didEncrypt(self, encrypted):
        # now is the time when we determine a transport. It must get the
        # message to the recipient, and it must obey the sender's policy
        # (secure or not).

        # TODO: allow reply messages to have their own Transport,
        # independent of the sender or recipient

        if not self.recipientRecord:
            raise Unsendable("cannot get a Transport without an IDRecord")
        self.log("getting a transport")
        wantSecure = self.senderEntry.pseudonymous
        t = None
        if self.recipientEntry:
            # give the AddressBookEntry a try, because we may have a
            # SURBSupply for this recipient
            t = self.recipientEntry.getTransport(self, wantSecure)
        if not t:
            t = self.recipientRecord.getTransport(self, wantSecure)
        #d.addCallback(self._gotTransport, encrypted)
        #return d
        return self._gotTransport(t, encrypted)

    def _gotTransport(self, transport, encrypted):
        assert implements(transport, interfaces.ITransport)
        if self.senderEntry.pseudonymous:
            if not implements(transport, interfaces.IUntraceableTransport):
                self.log("Hey, transport is not secure", transport)
                raise Unsendable("transport is not secure")
        self.log("transport:", transport)
        return transport.send(encrypted)


    def encrypt(self, us, recipient):
        """Returns a Deferred with the encrypted form of this message.
        If 'us' is present, sign the message.
        If 'recipient' is present, encrypt the message.
        """

        if us:
            if not isinstance(us, crypto.PrivateKey):
                print "us", us
            assert isinstance(us, crypto.PrivateKey)
            self.log("signing with %s" % us.getID())

        if recipient:
            assert isinstance(recipient, crypto.PublicKey)
            self.log("encrypting to %s" % recipient.getID())
        else:
            self.log("not encrypting")
        encoded = util.encode(self.headers)

        if recipient:
            if us:
                return recipient.encryptAndSign(us, encoded)
            else:
                return recipient.encrypt(encoded)
        else:
            if us:
                return us.sign(encoded)
            else:
                return self.agent.store(encoded)

    def dump(self):
        if not self.doDump:
            return
        log = self.agent.log
        log("Outbound message dump:")
        keys = self.headers.keys()
        keys.sort()
        log("", ",".join(keys))
        for m in self.eventlog:
            log(" %s" % m)
        log(" message dump finished")


class ReceivedMessage(BaseMessage):
    """A ReceivedMessage object is created for each incoming message.

    The .process method is run on it to check permissions. Once this
    succeeds, the following attributes will be available on the
    ReceivedMessage object:

    TODO: add the permission with which the message was accepted

    Useful attributes of this object which are set early:
    .agent: the agent which is receiving the message
    .encrypted: a file object (or StringIO) with the original encrypted form
    .recipientEntry: the OwnerEntry that is the target of this message
    .headers: the unserialized contents of the message

    Attributes which are set by various processing steps (and the .provides
    tags which indicate they are available):

    .allowRFC822 [checkedRFC822]: a boolean indicating that we are willing
    to accept an rfc822 message from this ReceivedMessage

    .senderEntry: always present, but only non-None for messages which have
    a valid signature from a previously-known sender. It may be set later
    (after InboundProcessing is complete) if the user chooses to add the
    sender to their address book.
    
    """

    __implements__ = (interfaces.IReceivedMessage,
                      interfaces.IInboundMessage,
                      interfaces.IInboundMessage_rfc822)

    def __init__(self, agent, encrypted):
        """If any part of the message is invalid, this will raise
        InvalidMessage or UndecryptableError. If the sender does not have
        permission, then it will raise NotPermitted."""

        BaseMessage.__init__(self, agent)
        self.encrypted = encrypted
        self.startLogging("Inbound Message created")

    def decrypt(self):
        """I decrypt my crypttext and set the following attributes:

        .recipientEntry: the OwnerEntry of the message recipient (to whom
        the message was encrypted, or the 'recipientID' node if the message
        was not encrypted)

        .headers: the plaintext message dictionary

        .senderEntry: this will point to an AddressBookEntry if the message
        has a valid signature from a known sender (someone who is already in
        our AddressBook). It will be None if the message had no signature or
        was signed by a sender we don't already know (such as a new sender).

        .senderIDRecord: if .senderEntry is None, the AcceptRFC822
        processing step uses .createIDRecord() to populate this with the
        unverified IDRecord of the otherwise-unknown sender. It is only
        really useful to extract a name for presentation to the user.

        I return a Deferred which will fire with the .recipientEntry .
        """

        ca = self.agent.crypto
        d = ca.decrypt(self.encrypted)
        d.addCallback(self._decrypt2)
        return d

    def _decrypt2(self, results):
        recipientKey, senderID, plaintext = results

        self.headers = util.decode(plaintext)
        self.log("headers: %s" % self.headers.keys())

        try:
            if recipientKey:
                recipientID = recipientKey.getID()
            else:
                recipientID = self.getNode("recipientID")
            self.recipientEntry = self.agent.getSender(recipientID)
        except KeyError:
            self.log("message was meant for someone else")
            raise

        self.senderID = senderID
        self.senderIDRecord = None
        self.senderEntry = None
        if senderID:
            book = self.agent.getAddressBook(recipientID)
            self.senderEntry = book.lookupID(senderID)

        if self.senderEntry:
            who = self.senderEntry.getName()
            self.log("message sent by known sender %s" % who)
        else:
            self.log("message sent by unknown sender")

        return self.recipientEntry

    def getProcessors(self):
        return self.agent.getInboundProcessors()

    def startProcessing(self):
        # return a Deferred which fires with self
        d = BaseMessage.startProcessing(self)
        d.addCallback(self.handlePlugins)
        d.addCallback(lambda res: self)
        return d

    def handlePlugins(self, res):
        # give all inbound plugins a look at the message
        d = defer.succeed(None)
        if self.senderEntry:
            sender = self.senderEntry
            d.addCallback(sender.runPluginsD, "inboundMessage", self)
        if self.recipientEntry:
            recip = self.recipientEntry
            d.addCallback(recip.runPluginsD, "inboundMessage", self)
        if self.agent:
            d.addCallback(self.agent.runPluginsD, "inboundMessage", self)
        return d

    def dump(self):
        if not self.doDump:
            return
        log = self.agent.log
        log("Inbound message dump:")
        if self.senderEntry:
            log(" From:", self.senderEntry.getName())
        if "checkedRFC822" in self.provides and self.allowRFC822 and \
            self.getNode("rfc822"):
            l = len(self.getNode('rfc822')['body'])
            log(" RFC822 len(%d) accepted" % l)
        for m in self.eventlog:
            log(" %s" % m)
        log(" message dump finished")

    # IInboundMessage, IInboundMessage_rfc822

    def getReason(self):
        return self.allowRFC822_reason

    def getRecipient(self):
        return self.recipientEntry

    def getSender(self):
        return self.senderEntry # could be None (if anonymous or unknown)

    def getSenderID(self):
        return self.senderID # could be None (if anonymous)

    def getSenderIDRecord(self): # the armored form
        if self.senderEntry:
            r = self.senderEntry.idRecord.signed
        else:
            r = self.getNode("senderIDRecord")
        if r:
            r = util.armor(r, "IDRECORD")
        return r

    def getIDRecord(self):
        if self.senderIDRecord:
            return defer.succeed(self.senderIDRecord)
        return self.createIDRecord()

    def createIDRecord(self):
        if self.senderEntry and self.senderEntry.idRecord:
            self.senderIDRecord = self.senderEntry.idRecord
            return defer.succeed(self.senderIDRecord)
        signed = self.getNode("senderIDRecord")
        if not signed:
            self.senderIDRecord = None
            return defer.succeed(None)
        r = IDRecord()
        d = r.createFromSigned(signed, self.agent.crypto)
        d.addCallback(self._createIDRecord)
        return d
    def _createIDRecord(self, record):
        self.senderIDRecord = record
        return record

    # canAddSender, canGrantInboundPermission, and canReply are only meant
    # to be called on rfc822-accepted messages

    def canAddSender(self):
        """I check to see if the sender of this message can be added to the
        recipient's AddressBook (i.e.: they are not anonymous, they are not
        already listed). If not, I return True to indicate that the user
        should be offered the option of adding them. To actually add them,
        just call my addSender method."""
        
        if self.senderEntry:
            return False # already there

        book = self.agent.getAddressBook(self.recipientEntry.ID)
        if not book:
            return False # nowhere to put it
        if not self.getNode("senderIDRecord"):
            return False # they didn't give us enough information to add one
        return True

    def addSender(self, reason=None):
        # returns a Deferred that fires with the entry that was added
        self.log("asked to add sender entry")
        book = self.agent.getAddressBook(self.recipientEntry.ID)
        signed = self.getNode("senderIDRecord")
        assert signed
        if not reason:
            reason = self
        d = book.addFromSigned(signed, reason)
        d.addCallback(self._addedSender1)
        return d

    def _addedSender1(self, entry):
        self.senderEntry = entry
        self.log("added sender entry")
        self.grantInboundPermission()
        self.log("granted them inbound permission")
        # now we must update our outbound permission
        slip = self.getNode("yourPermission")
        if not slip:
            self.senderEntry.modified()
            return entry # no specific permission
        d = self.senderEntry.updateOurPermission(slip)
        d.addCallback(self._addedSender2, entry)
        return d

    def _addedSender2(self, results, entry):
        self.senderEntry.modified()
        self.log("added our permission")
        return entry

    def canAddReferences(self):
        """I check to see if there are any References in the message
        (pointers to other people that we don't already know about), so the
        user can be offered the option of adding them. I return a list of
        refnodes, one per new Reference. This list might be empty.

        Each refnode is a dictionary. It has a ['name'] node which indicates
        how the sender refers to them, and it may be useful to display this
        to the user. To add the reference, give the refnode to my
        .addReference method.
        """

        if not hasattr(self, 'references'):
            return []
        book = self.agent.getAddressBook(self.recipientEntry.ID)
        newnodes = [r for r in self.references if not book.lookupID(r['id'])]
        return newnodes

    def addReference(self, refnode):
        """I add a reference to our addressbook. If the sender can help us
        get Permission to the referenced recipient, we will sent a request to
        start the process.

        I return a Deferred which will fire when the AddressBookEntry has
        been added. It fires with that entry. Any permission-requesting is
        performed asynchronously.
        """

        lognode = self.log("asked to addReference for %s (%s)" % \
                           (refnode['name'], refnode['id']))
        book = self.agent.getAddressBook(self.recipientEntry.ID)
        reason = self
        d = book.addFromSigned(refnode['idrecord'], reason)
        d.addCallback(self._addedReference1, refnode, lognode)
        return d

    def _addedReference1(self, entry, refnode, lognode):
        self.log("added reference", parent=lognode)
        if refnode['offerPermission']:
            self.log("can get Permission to %s" % entry.getName(),
                     parent=lognode)
            # send permission request
            req = OutboundMessage(self.agent)
            req.setSenderEntry(self.recipientEntry)
            req.setRecipientEntry(entry)
            request = { 'voucherID': self.senderIDRecord.ID }
            req.addNode("request-permission", request)
            # TODO: make sure the message includes our IDRecord
            # send the request and
            # TODO: tell the agent that we await the response
            d = self.agent.sendMessage(req)
            # ignore the Deferred it returns and any errors caused
            d.addErrback(lambda why: None)
        return entry
        

    def canGrantInboundPermission(self):
        """I check to see if the sender of the message already has inbound
        permission. If not, I return True to indicate that the user should
        be offered the option of granting it to them. To actually grant it,
        just call my grantInboundPermission method.
        """
        entry = self.senderEntry
        if not entry:
            return False # must know who they are to grant them permission
        if entry.theirPermission:
            return False # they already have it
        if entry.doNotGrantReturnPermission:
            return False # we don't want to give it to them
        return True

    def grantInboundPermission(self, reason=None):
        if not reason:
            reason = self
        self.recipientEntry.grantPermissionTo(self.senderEntry, reason)

    def getFromName(self):
        """I return a Deferred which fires with something reasonable to put
        in the 'from' box. This is the user-specified pet name, or the
        sender-specified name (in quotes), or their ID number as a last
        resort. If there was no sender at all (the message was sent
        anonymously) I will return '-anonymous-'."""

        if self.senderEntry:
            return defer.succeed(self.senderEntry.getName())
        d = self.getIDRecord()
        d.addCallback(self._getFromName)
        return d
    def _getFromName(self, senderIDRecord):
        if senderIDRecord:
            if senderIDRecord.name:
                return '"%s"' % senderIDRecord.name
            return senderIDRecord.ID
        return "-anonymous-" # TODO: maybe None to indicate anonymous?

    def canReply(self):
        # TODO: you can reply to idRecords too
        if self.senderEntry:
            return True
        return False

    def reply(self):
        # TODO: you can reply to idRecords too
        assert self.senderEntry
        m = self.agent.newMessage("rfc822")
        m.setSender(self.recipientEntry)
        m.addRecipient_correspondent(self.senderEntry)
        return m

    def getMessage(self):
        return self.getNode("rfc822")['body']
