
"""Utilities to support Mixminion usage."""

import re, os.path, time, calendar
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from twisted.python import failure
from twisted.internet import protocol, defer, reactor
from twisted.application import service, internet
from petmail import util, interfaces, plugin

class HelperPP(protocol.ProcessProtocol):
    debug = 0
    def __init__(self, inputs, outputs=[]):
        # inputs is a dict which maps fd numbers to strings
        self.inputs = inputs
        # outputs is a list of fds which should be read
        self.outputs = {}
        for fd in outputs:
            self.outputs[fd] = "" 
        # this Deferred will be fired (with the outputs dict) when the
        # process is complete
        self.deferred = defer.Deferred()

    def connectionMade(self):
        if self.debug: print "connectionMade"
        for fd, text in self.inputs.items():
            self.transport.writeToChild(fd, text)
            self.transport.closeChildFD(fd)

    def childDataReceived(self, childFD, data):
        if self.debug: print "childDataReceived", data
        self.outputs[childFD] += data

    def processEnded(self, status):
        if self.debug: print "processEnded"
        if status.value.signal:
            self.deferred.errback(status)
        else:
            self.rc = status.value.exitCode
            if self.rc:
                self.deferred.errback(status)
            else:
                self.deferred.callback(self.outputs)

class VersionGetter(object):
    debug = 0
    def get(self, mixminion_name):
        mixminion_name = os.path.expanduser(mixminion_name)
        cmd = (mixminion_name, "version")
        if self.debug: print "cmd:",cmd
        pp = HelperPP({}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={0:"w", 1:"r", 2:"r"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._done)
        d.addErrback(self._error, pp)
        return d

    def _done(self, outputs):
        if self.debug: print "VersionGetter.done"
        stdout = outputs[1]
        r = re.search(r'Mixminion version (\S+)$', stdout, re.M)
        if not r:
            return (None, stdout)
        return (r.group(1), None)

    def _error(self, why, pp):
        if self.debug: print "VersionGetter._error"
        return (None, pp.outputs[2])


def getVersion(mixminion_name="mixminion"):
    """I attempt to run mixminion and determine what version is available. By
    default I use a bare 'mixminion' command (which means your $PATH will be
    searched), but you can override this to use /usr/local/bin/mixminion or
    ~/bin/mixminion or whatever.

    I return a Deferred which fires with a tuple, either (version, None),
    or (None, errortext)."""

    return VersionGetter().get(mixminion_name)

class Decoder(object):
    def decode_plaintext(self, encoded, mm_path, rcfile):
        cmd = (mm_path, "decode", "--input", "-")
        if rcfile:
            cmd = cmd + ("--config", rcfile)
        pp = HelperPP({0:encoded}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={0:"w", 1:"r", 2:"r"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._done)
        return d

    def _done(self, outputs):
        stdout = outputs[1]
        return stdout

def decode_plaintext(encoded, mm_path="mixminion", rcfile=None):
    return Decoder().decode_plaintext(encoded, mm_path, rcfile)

class SURB(object):
    """I represent a single SURB.
    """
    # TODO: make this smaller, there could be thousands of them
    granted = False
    redeemed = False
    expired = False

    def __init__(self, identity, expires):
        self.identity = identity
        self.expires = expires
        self.watchers = []

    def setSURBid(self, surbid):
        self.surbid = surbid
    def setReplyBlock(self, text):
        self.text = text

    def get(self):
        """Retrieve the text of the SURB, suitable for giving to someone
        else."""
        assert not self.granted
        self.granted = True
        return self.text

    def redeem(self):
        assert self.granted
        assert not self.redeemed
        self.redeemed = True

    def notifyOnExpiration(self, watcher):
        self.watchers.append(watcher)
    def dontNotifyOnExpiration(self, watcher):
        self.watchers.remove(watcher)

    def expire(self):
        assert not self.redeemed
        assert not self.expired
        self.expired = True
        for w in self.watchers:
            w(self)

class MixminionContext(service.MultiService):
    """I maintain a Mixminion context on behalf of an Agent.
    This context has its own .mixminionrc and a separate home directory.

    I keep track of all SURBs created in this context, from minting to final
    redemption or expiration.
    """

    startedTimer = False
    pollinterval = 3600 # seconds

    def __init__(self, basedir):
        service.MultiService.__init__(self)
        """Create the context. 'basedir' should be a subdirectory of the
        parent agent's base directory.
        """
        self.basedir = basedir
        if not os.path.isdir(basedir):
            os.mkdir(basedir, 0700)
        self.mmrcfile = os.path.join(basedir, "mixminion.rc")
        self.mmrcfile = os.path.abspath(self.mmrcfile)
        f = open(self.mmrcfile, "w")

        cfg = """
[Host]

[DirectoryServers]
DirectoryTimeout: 1 minute

[User]
UserDir: %s

[Security]

[Network]
Timeout: 2 minutes
"""

        f.write(cfg % os.path.abspath(self.basedir))
        self.surbids = {} # keys are surbids, values are SURBs
        self.outstanding = []
        self.passphrase = "Sekrit" # TODO: generate randomly

    def set_mm_path(self, mm_path):
        self.mm_path = mm_path

    def startService(self):
        if not self.startedTimer:
            # note that the TimerService is not pickleable until it has been
            # stopped.
            # TODO: consider using the deferred to prevent overlapping checks
            # TODO: SURB expiration is quantized to midnight GMT, so this
            #  could be done less frequently. Remember that the SURB could
            #  be in flight when it expires.
            timer = internet.TimerService(self.pollinterval, self.check)
            timer.setServiceParent(self)
            self.startedTimer = True
        service.MultiService.startService(self)
        self.check()

    def remove(self, s):
        self.outstanding.remove(s)
        if self.surbids.get(s.surbid):
            del self.surbids[s.surbid]

    def expire(self, s):
        self.remove(s)
        s.expire()

    def redeem(self, surbid):
        """I should be called when a message arrives that used the given
        SURB. I mark the SURB as having been redeemed."""
        s = self.surbids.get(surbid)
        if s:
            self.remove(s)
            s.redeem()

    def check(self):
        # TODO: sort .outstanding by expiration date to make this faster
        now = util.now()
        expired = []
        for s in self.outstanding:
            if s.expires and s.expires > now:
                expired.append(s)
        for s in expired:
            self.expire(s)

    def checkDirectoryDownload(self):
        """Utility method to see if mixminion can reach its directory
        servers. It needs to do this the first time it is run, and every
        time it thinks its list is out of date.

        This method returns a Deferred which fires back with True or False.
        If false, mixminion cannot be used (probably because we are not
        currently online). """
        
        cmd = (self.mm_path, "list-servers",
               "--config=%s" % self.mmrcfile,
               "-D", "yes", # force download
               )
        pp = HelperPP({}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={1:"r", 2:"r"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._checkDD_done)
        d.addErrback(self._checkDD_failed)
        return d

    def _checkDD_done(self, outputs):
        #print outputs[2]
        return True
    def _checkDD_failed(self, why):
        return False

    def mint(self, dest, idstring, lifetime=7):
        """Create a new SURB for the given identity-string. Each
        identity-string requires mixminion to maintain a separate secret
        key, so don't go nuts with them. These strings are strictly local,
        so it is safe to use the Owner's ID number for them.
        """

        # per the protocol, the SURB will expire 'lifetime' days after the
        # previous GMT midnight. This routine taken from mixminion.Common
        when = util.now() + 24*60*60*lifetime
        yyyy,MM,dd = time.gmtime(when)[0:3]
        expires = calendar.timegm((yyyy,MM,dd,0,0,0,0,0,0))

        surb = SURB(idstring, expires)

        cmd = (self.mm_path, "generate-surb",
               "--config=%s" % self.mmrcfile,
               "--to=%s" % dest,
               "--identity=%s" % idstring,
               "--lifetime=%d" % lifetime,
               "--passphrase-fd=%d" % 3,
               )
        pp = HelperPP({3:self.passphrase}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={1:"r", 2:"r", 3:"w"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._mint_done, surb)
        d.addErrback(self._mint_failed, pp)
        return d

    def _mint_done(self, outputs, surb):
        stdout = outputs[1]
        #print "stderr", outputs[2]
        surbid = "unknown" # TODO: need MM --status-fd API to get this
        surb.setReplyBlock(stdout)
        surb.setSURBid(surbid)

        self.outstanding.append(surb) # TODO: sort by expiration date
        self.surbids[surbid] = surb
        return surb

    def _mint_failed(self, why, pp):
        print "mint failed"
        print pp.outputs[2]
        return why

    def decrypt(self, message):
        """Decrypt a mixminion message. Returns a Deferred which fires with
        a tuple (plaintext, identity, surbid). If the message was encrypted
        with a SURB, the SURB's 'identity' string and surbid are returned,
        and the SURB is marked as having been used. If the message was not
        encrypted (message-type: plaintext), 'identity' and 'surbid' will be
        None."""

        cmd = (self.mm_path, "decode",
               "--config=%s" % self.mmrcfile,
               "--verbose", # want INFO lines
               "--input=-",
               "--passphrase-fd=%d" % 3,
               )
        pp = HelperPP({0:message, 3:self.passphrase}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={0:"w", 1:"r", 2:"r", 3:"w"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._decrypt_done)
        d.addErrback(self._decrypt_failed, pp)
        return d

    def _decrypt_done(self, outputs):
        stdout = outputs[1]
        stderr = StringIO(outputs[2])
        surbid, idstring = None, None
        for line in stderr.readlines():
            r = re.search(r"\[INFO\] Decoded reply message to identity '(\w*)'",
                          line)
            if r:
                surbid = None # TODO: need MM --status-fd API to get this
                idstring = r.group(1)
        if surbid:
            self.redeem(surbid)
        return (stdout, idstring, surbid)

    def _decrypt_failed(self, why, pp):
        print "mixminion decode failed"
        print why
        print pp.outputs[2]
        return why

    def sendToDest(self, body, dest, subject, queue=False):
        assert queue # safety net during development
        cmd = (self.mm_path, "send",
               "--config=%s" % self.mmrcfile,
               "--to", dest,
               "--subject", subject,
               )
        if queue:
            cmd = cmd + ("--queue",)
        pp = HelperPP({0: body}, [1,2])
        d = pp.deferred
        p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                 childFDs={0:"w", 1:"r", 2:"r"})
        d.addCallback(self._sent)
        d.addErrback(self._sendError, pp)
        return d

    def sendToSURBs(self, body, surb, subject, queue=False):
        assert queue # safety net during development
        assert type(surb) == type("")
        cmd = (self.mm_path, "send",
               "--config=%s" % self.mmrcfile,
               "--reply-block-fd", "3",
               "--subject", subject,
               )
        if queue:
            cmd = cmd + ("--queue",)
        # surb may have multiple SURBs concentated together
        pp = HelperPP({0: body, 3:surb}, [1,2])
        d = pp.deferred
        p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                 childFDs={0:"w", 1:"r", 2:"r", 3:"w"})
        d.addCallback(self._sent)
        d.addErrback(self._sendError, pp)
        return d

    def _sent(self, outputs):
        return True

    def _sendError(self, why, pp):
        print "mixminion send failed"
        print why
        print "stdout:", pp.outputs[1]
        print "stderr:", pp.outputs[2]
        return interfaces.Unsendable("mixminion send failed")

    def countSURBs(self, body):
        cmd = (self.mm_path, "count-packets",
               "--config=%s" % self.mmrcfile,
               )
        pp = HelperPP({0: body}, [1,2])
        d = pp.deferred
        p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                 childFDs={0:"w", 1:"r", 2:"r"})
        d.addCallback(self._counted)
        d.addErrback(self._countError, pp)
        return d

    def _counted(self, outputs):
        stdout = StringIO(outputs[1])
        for line in stdout.readlines():
            r = re.search(r"Message of length (\d+) will require (\d+) SURBs",
                          line)
            if r:
                count = int(r.group(2))
                return count
        # something went wrong
        raise RuntimeError("couldn't run count-packets")

    def _countError(self, why, pp):
        print "mixminion count-packets failed"
        print why
        print "stdout:", pp.outputs[1]
        print "stderr:", pp.outputs[2]
        raise RuntimeError("couldn't run count-packets")


class SURBSupply(object):
    """I keep track of SURBs issued to us for a single recipient. I am
    responsible for providing SURBs on demand to send messages to that
    recipient, and for noticing when we are running low to ask for more.

    Each anonymous correspondent has a SURBSupply in their AddressBookEntry,
    and (TODO: somehow) their Transport record points to it.
    """

    def add(self, surbs):
        pass

    def get(self, count):
        if count > len(self.supply):
            raise interfaces.Unsendable("Message is too large, we do not have enough SURBs")
        used = self.supply[:count]
        self.supply = self.supply[count:]
        # TODO: tell the other side if we need more, maybe
        return used

class SURBSupplyAgentPlugin(plugin.AgentPlugin):
    def inboundMessage(self, msg):
        if not msg.hasNode("surb-transport"):
            return
        if not msg.senderEntry:
            return # we aren't keeping track of them, don't bother

        # there must not be an EntryPlugin to take the SURB. Install one.
        p = SURBSupplyEntryPlugin()
        msg.senderEntry.addPlugin(p)
        return p.getSURBsFromMessage(msg)

class SURBSupplyEntryPlugin(plugin.AddressBookEntryPlugin):

    def __init__(self):
        self.supply = SURBSupply()

    def inboundMessage(self, msg):
        if not msg.hasNode("surb-transport"):
            return
        return self.getSURBsFromMessage(msg)
    def getSURBsFromMessage(self, msg):
        surblist = msg.getNode("surb-transport")
        # TODO: turn the SURB blocks into mm_util.SURB objects
        d = self.supply.add(surbs)
        return d


class SURBStocker(object):
    """I keep track of SURBs that we have issued to a single sender. I am
    responsible for keeping them stocked with a reasonable supply of SURBs.
    This means noticing when they are running low and sending them new
    batch. I am also responsible for responding to their requests for more
    SURBs. I keep track of the total number of SURBs outstanding at any
    given time, so it can be kept low enough to avoid a burst attack.

    If we are anonymous, each correspondent in our address book will have a
    SURBStocker instance in their AddressBookEntry. In addition each SURB
    Server we use will have a SURBStocker instance.
    """
    pass


def test():
    d = getVersion()
    def done(results):
        version, error = results
        if version:
            print "good!:", version
        else:
            print "bad!:", error
        reactor.stop()
        # this could be called before reactor.run() is
    def oops(why):
        print why
        reactor.stop()
    d.addCallback(done)
    d.addErrback(oops)
    print "running"
    reactor.run()

if __name__ == '__main__':
    test()
