#! /usr/bin/python

from petmail import pbutil
from petmail.interfaces import IOwnerEntry
from petmail.idrecord import IDRecord

from twisted.python.components import implements
from twisted.python import log
from twisted.internet import defer

import os


class FakeNameServer(object):
    def __init__(self, args):
        self.names = {}
        self.IDs = {}
    def publish(self, record, signed):
        assert implements(record, IOwnerEntry)
        for address in record.addresses:
            self.names[address] = signed
        self.IDs[record.ID] = signed

    def lookup(self, agent, name):
        return defer.succeed(self.names.get(name))
    def lookupID(self, agent, ID):
        return defer.succeed(self.IDs.get(ID))
    def describe(self):
        return "Fake"

class DiskNameServer(object):
    def __init__(self, args):
        self.mapdir = args

    def _getAllRecords(self, agent, records):
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        defs = []
        for f in os.listdir(mapdir):
            s = open(os.path.join(mapdir, f), "r").read()
            r = IDRecord()
            d = r.createFromSigned(s, agent.crypto)
            d.addCallback(self._gotARecord, records)
            d.addErrback(self._notGotARecord, f)
            defs.append(d)
        dl = defer.DeferredList(defs)
        return dl

    def _gotARecord(self, record, records):
        records.append(record)
    def _notGotARecord(self, failure, f):
        log.err(failure)
        return None

    def lookup(self, agent, address):
        """Take a name and return a (single) ID Record, using a server of
        some sort."""
        records = []
        d = self._getAllRecords(agent, records)
        d.addCallback(self._doLookup, records, address)
        return d
    def _doLookup(self, results, records, address):
        for r in records:
            if address in r.addresses:
                return r.signed
        return None

    def lookupID(self, agent, ID):
        """Take an ID number and return a (single) ID Record, using a server
        of some sort."""
        records = []
        d = self._getAllRecords(agent, records)
        d.addCallback(self._doLookup, records, ID)
        return d
    def _doLookupID(self, results, records, ID):
        for r in records:
            if ID == r.ID:
                return r.signed
        return None

    def publish(self, record, signed):
        assert implements(record, IOwnerEntry)
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        f = open(os.path.join(mapdir, record.ID), "w")
        f.write(signed)
        f.close()

    def describe(self):
        return "Disk:%s" % self.mapdir

class PBNameServer(object):
    def __init__(self, url):
        self.hostname, portnum = url.split(":")
        self.portnum = int(portnum)

    def lookup(self, agent, address):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookup", address)
        return d

    def lookupID(self, agent, ID):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookupID", ID)
        return d

    def _lookup2(self, server, method, address):
        d = server.callRemote(method, address)
        return d

    def publish(self, record, signed):
        assert implements(record, IOwnerEntry)
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._publish2, signed)
        return d
    def _publish2(self, server, signed):
        d = server.callRemote("publish", signed)
        return d

    def describe(self):
        return "PB:%s:%d" % (self.hostname, self.portnum)    

def getLookupAgentFromString(args):
    scheme, rest = args.split(":", 1)
    fac = {"Fake": FakeNameServer,
           "Disk": DiskNameServer,
           "PB": PBNameServer,
           }
    return fac[scheme](rest)
