
from twisted.application import service
from twisted.python.components import implements
from twisted.python import log
from twisted.internet import defer, reactor
from twisted.persisted import styles

from petmail import crypto, util, interfaces, permission, plugin
from petmail import message
from petmail.interfaces import NotPermitted


# standard Owner plugins

class UpdateTheirIDRecord(plugin.OwnerPlugin):
    """If we aren't remembering them, they probably don't know about us, so
    send them a copy of our record."""

    def outboundMessage(self, msg):
        lognode = msg.log("owner-UpdateTheirIDRecord")
        if msg.hasNode("senderIDRecord"):
            return # already there
        if msg.anonymous:
            lognode.log("anonymous, not sending our IDRecord")
            return
        if msg.recipientEntry:
            # they know about us, leave it up to the plugin in the Entry
            return
        d = msg.senderEntry.sign()
        d.addCallback(self._gotSigned, msg, lognode)
        return d

    def _gotSigned(self, signed, msg, lognode):
        msg.addNode("senderIDRecord", signed)
        lognode.log("included copy of our IDRecord")


class HandleReferences(plugin.OwnerPlugin):
    # scan for any inbound references. This is kind of dull right now,
    # beacuse we put off doing anything with the references until/unless the
    # user decides to add them to the address book. It might make sense to
    # skip this translation step altogether.
    def inboundMessage(self, msg):
        refs = msg.getNode("references")
        if not refs:
            return
        references = []
        for r in refs:
            msg.log("reference: offer=%d %s %s" % (r['offerPermission'],
                                                   r['name'], r['id']))
            references.append(r)
        msg.references = references


class HandlePermissionRequest(plugin.OwnerPlugin):
    def inboundMessage(self, msg):

        req = msg.getNode("request-permission")
        if not req:
            return
        # msg.senderIDRecord wants permission to send to us.
        voucherID = req['voucherID']
        # they claim that this person vouches for them
        book = self.agent().getAddressBook(msg.recipientEntry.ID)
        voucher = book.lookupID(voucherID)
        if not voucher:
            return
        if not voucher.theirPermission:
            return
        if not voucher.theirPermission.canGrantPermission:
            return
        # we trust this voucher to recommend senders. Send them a query to
        # see if they really do vouch for this person; if we get back a
        # signed message stating that they do, we'll accept them.
        
        # we need the requestor's ID, which probably hasn't been extracted
        # yet
        if msg.senderIDRecord:
            d = defer.succeed(msg.senderIDRecord)
        else:
            d = msg.createIDRecord()
        d.addCallback(self._gotIDRecord, msg, voucher, lognode)
        return d

    def _gotIDRecord(self, record, msg, voucher, lognode):
        requestorID = record.ID

        req = message.OutboundMessage(self.agent())
        req.setSenderEntry(msg.recipientEntry)
        req.setRecipientEntry(voucher)
        request = { 'requestorID': requestorID }
        req.addNode("check-voucher", request)

        d = self.agent().sendRequest(req)
        # slow: this response will take a roundtrip to the voucher
        d.addCallback(self._gotVouchResponse,
                      requestorID, voucher, msg, lognode)
        return d

    def _gotVouchResponse(self, reply,
                          requestorID, voucher, msg, lognode):
        owner = msg.recipientEntry
        # the response must be from the voucher to whom we sent the request
        if not reply.senderEntry:
            return # message was unsigned
        if (reply.senderEntry.idRecord.ID != voucher.idRecord.ID):
            return # message was forged: signed by the wrong party
        replynode = reply.getNode('check-voucher-reply')
        if replynode['requestorID'] != requestorID:
            return # weird, they answered the wrong question
        if not replynode['doVouch']:
            return # they said no

        # ok, we are now willing to grant them permission. We have to
        # add them to the addressbook first
        reason = reply
        d = msg.addSender(reason)
        d.addCallback(self._added, msg, lognode)
        return d

    def _added(self, entry, msg, lognode):
        reason = "recommended"
        msg.grantInboundPermission(reason)
        return


class CheckRFC822(plugin.OwnerPlugin):
    """If an incoming message has an rfc822 node with valid permission,
    deliver it."""
    #provides = ["checkedRFC822"]

    def inboundMessage(self, msg):
        a = self.agent()
        debug = a.debugProcessing

        if not msg.getNode("rfc822"):
            return

        # determine which Permission record to use

        # they use genericPermission if:
        #  A: we don't know who they are
        #  B: we know them but haven't given them specific permission
        p = None
        if msg.senderEntry:
            ptype = "specific"
            p = msg.senderEntry.getTheirPermission('rfc822')
        if not p:
            ptype = "generic"
            p = msg.recipientEntry.getGenericPermission('rfc822')
        if not p:
            msg.log("rfc822 rejected: no permission")
            return self.reject(msg, NotPermitted("no permission"))

        msg.log("they have %s permission" % ptype)

        # TODO: remove this .maybeDeferred, but only after turning the
        # exception-raises in checkInboundMessage into something like
        # 'return defer.fail(why)'
        d = defer.maybeDeferred(p.checkInboundMessage, msg)
        d.addCallback(self.acceptMessage, ptype, p, msg)
        d.addErrback(self.checkFailed, msg)
        return d

    def acceptMessage(self, why, ptype, p, msg):
        why = "(%s) %s" % (ptype, why)
        # if we get here, the message is accepted

        # TODO: do this now? or when message is delivered?
        p.acceptInboundMessage(msg)

        msg.log("will accept rfc822 message: %s" % why)
        msg.allowRFC822 = True
        msg.allowRFC822_reason = why

        # try to create some kind of sender information
        if msg.senderEntry:
            d = defer.succeed(None)
        else:
            msg.log("extracting IDRecord for sender name first")
            d = msg.createIDRecord()
        d.addCallback(self.deliver, msg, why)
        return d

    def deliver(self, results, msg, why):
        # tell the Identity to deliver the body to the user. At this point,
        # message.senderName() will work.
        msg.recipientEntry.deliverMessage("rfc822", msg)

    def checkFailed(self, f, msg):
        f.trap(NotPermitted)
        why = f.value
        return self.reject(msg, why)

    def reject(self, msg, why):
        msg.log("rejected rfc822 message: %s" % why)
        msg.allowRFC822 = False
        msg.agent.rejectRFC822(msg, why)



## end plugins


class OwnerEntry(service.MultiService,
                 plugin.PluginHolderMixin,
                 styles.Versioned):
    """I hold information about one of the owner's identities. The agent
    holds a collection of these objects. Encoded copies will be published to
    the Name Servers.

    All OwnerEntry objects are children of the Agent, so they can get at it
    with self.parent
    """

    persistenceVersion = 1
    __implements__ = (interfaces.IOwnerEntry, interfaces.IIdentity,
                      service.MultiService.__implements__)
    grantReturnPermission = True # tested by processors.GrantReturnPermission
    grantInboundPermission = True #  by processors.GrantInboundPermission
    pseudonymous = False # use mixminion, never use direct TCP
    fromEmail = None
    acceptablePluginInterface = interfaces.IOwnerPlugin
    standardPlugins = [UpdateTheirIDRecord, HandleReferences, CheckRFC822]

    doPublish = True
    dirty = False
    publishTimer = None

    def __init__(self):
        service.MultiService.__init__(self)
        # .name, .running, and .parent are reserved for use by MultiService

        self.myName = None # local pet name for this identity
        self.inlets = []
        self.lastModified = util.now()
        self.privkey = None
        self.signed = None # cache of our signed form
        self.hasBeenRevealed = False # TODO: set whenever direct TCP is used
        self.subscribers = {}
        self.queuedMessages = []

        # fields which get published
        self.ID = None # derived from the pubkey
        self.names = "" # human names, affiliations: multiline text string
        self.addresses = [] # addresses which should resolve to this record
        # fields which are published in some indirect form
        self.genericPermission = None # default is no permission
        self.expirationTime = 120 # one month is more reasonable

        # extensions
        self.extensions = {}
        for p in self.standardPlugins:
            self.addPlugin(p())

    def __repr__(self):
        return "<%s id=%s names=%s>" % (self.__class__.__name__,
                                        self.ID, self.names)

    def __getstate__(self):
        state = service.MultiService.__getstate__(self)
        state = styles.Versioned.__getstate__(self, state)
        state['subscribers'] = {}
        state['publishTimer'] = None # TODO: publish-on-save??
        return state

    def upgradeToVersion1(self): # post 0.1.2
        print "OwnerEntry.upgradeToVersion1"
        self.pseudonymous = False
        self.hasBeenRevealed = True
        self.extensions = {}

    def setServiceParent(self, parent):
        service.MultiService.setServiceParent(self, parent)
        for inlet in self.inlets:
            self.parent.addInlet(self, inlet)
        if self.dirty:
            self.publish()

    def disownServiceParent(self):
        d = service.MultiService.disownServiceParent(self)
        for inlet in self.inlets:
            self.parent.removeInlet(self, inlet) # TODO: could return Deferred?
        return d

    def getAgent(self):
        return self.parent

    def getProperty(self, pname):
        if pname == "id":
            return self.ID
        if pname == "name":
            return self.names
        if pname in ("myName", "addresses", "expirationTime", "pseudonymous"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))

    def setProperty(self, pname, value):
        assert pname in ("myName", "name", "addresses", "expirationTime")
        if pname == "myName":
            assert type(value) is str
            self.myName = value
            return # not published, do not mark as modified
        if pname == "name":
            # TODO: enhance this after some use-cases
            assert type(value) is str
            self.names = value
        if pname == "addresses":
            assert type(value) in (list, tuple)
            for n in value:
                assert type(n) is str
            self.addresses = value
        if pname == "expirationTime":
            assert type(value) is int
            self.expirationTime = value
        self.modified()

    def getAddressBook(self):
        # TODO: why didn't I just make the addressbook a child of the Owner?
        return self.getAgent().getAddressBook(self.ID)

    def modified(self):
        self.lastModified = util.now()
        self.signed = None
        agent = self.getAgent()
        if agent:
            self.dirty = True
            agent.ownerChanged("change", self)
            if self.doPublish:
                # TODO: think about this
                if self.publishTimer:
                    # TODO: sometimes this is an Ephemeral. Why??
                    self.publishTimer.cancel()
                self.publishTimer = reactor.callLater(5,
                                                      self.publishTimerFired)

    def publishTimerFired(self):
        self.publishTimer = None
        self.publish()

    def publish(self):
        """Publish my record to my agent's publishers, using all of my
        addresses. Returns a Deferred that fires (with a meaningless value)
        when publishing is complete."""
        if self.publishTimer:
            self.publishTimer.cancel()
            self.publishTimer = None
        d = self.sign()
        d.addCallback(self._publish2)
        return d

    def _publish2(self, signed):
        agent = self.getAgent()
        if not agent:
            log.msg("want to publish, but can't because we have no Agent")
            return # oh well
        self.dirty = False
        defs = []
        for p in self.getAgent().addressPublishAgents:
            d = p.publish(self, signed)
            if isinstance(d, defer.Deferred):
                defs.append(d)
        if defs:
            return defer.DeferredList(defs)

    def setPseudonymous(self, state, yesReally=False):
        if state and self.hasBeenRevealed and not yesReally:
            raise interfaces.CatOutOfBagError
        self.pseudonymous = state

    def setKey(self, privkey):
        assert isinstance(privkey, crypto.PrivateKey)
        self.privkey = privkey
        self.ID = privkey.getID()
        self.modified()

    def changeKey(self):
        raise NotImplemented
        self.modified()

    def setName(self, name):
        # TODO: deprecate this
        self.setProperty("name", name)

    def getName(self):
        # used to generate a suitably descriptive string
        if self.myName:
            return self.myName
        if self.names and self.names != "":
            return self.names
        return self.ID

    def addAddress(self, addr):
        self.addresses.append(addr)
        self.modified()

    def setGenericPermission(self, ptype, perm):
        """Set .genericPermission to None to indicate no permission at all,
        or to an OwnerPermission to set up TicketServers and such."""
        assert ptype == "rfc822"
        if perm:
            assert implements(perm, interfaces.IOwnerPermission)
        self.genericPermission = perm
        self.modified()

    def getGenericPermission(self, ptype):
        if ptype != 'rfc822':
            raise IndexError("unknown permission type %s" % (ptype,))
        return self.genericPermission

    def grantPermissionTo(self, sender, message=None):
        assert implements(sender, interfaces.IAddressBookEntry)
        assert sender.getOwner() == self
        if not self.parent:
            return # we were deleted
        if not sender.addressbook:
            return # they were deleted
        log.msg("OwnerEntry.grantPermissionTo(%s)" % sender.getName())
        # TODO: This gives full permission. The default ought to be
        # configurable.
        # TODO: record 'message' as the reason for the permission
        if sender.theirPermission:
            log.msg("OwnerEntry.grantPermissionTo: they already have it")
            return
        p = permission.OwnerPermission()
        sender.setTheirPermission("rfc822", p)
        sender.modified()

    def getInlets(self):
        return self.inlets

    def setInlets(self, inlets):
        old_inlets = self.inlets[:]
        new_inlets = []
        modified = False
        for ni in inlets:
            if ni in old_inlets:
                old_inlets.remove(ni)
            else:
                new_inlets.append(ni)
        if old_inlets or new_inlets:
            modified = True

        ds = []
        for i in old_inlets:
            self.inlets.remove(i)
            if self.parent:
                d = defer.maybeDeferred(self.parent.removeInlet, self, i)
                ds.append(d)
        dl = defer.DeferredList(ds)
        dl.addCallback(self._setInlets2, new_inlets, modified)
        return dl

    def _setInlets2(self, res, new_inlets, modified):
        for i in new_inlets:
            assert implements(i, interfaces.IInlet)
            self.inlets.append(i)
            if self.parent:
                self.parent.addInlet(self, i)
            # otherwise this will be done in setServiceParent
        if modified:
            self.modified()

    def encode(self):
        """I return a Deferred which will fire with the encoded form of this
        record.
        """
        d = self.privkey.getKey()
        d.addCallback(self._encode)
        return d

    def _encode(self, pubkey):
        d = {'ID': self.ID,
             'pubkey': pubkey,
             'name': self.names,
             'addresses': self.addresses,
             'permission': None,
             'expires': util.now() + self.expirationTime,
             'transports': [inlet.encode() for inlet in self.inlets],
             'extensions': {},
             }
        if self.genericPermission:
            d['permission'] = self.genericPermission.encode()
        image = self.extensions.get('image')
        if image:
            d['extensions']['image'] = image
        return util.encode(d)

    def sign(self):
        """Return a Deferred which fires with a signed string ready for
        uploading to a server or including in a a message. This is the
        serialized form of the IDRecord, and can be passed to an IDRecord
        instance to create such an object."""

        if self.signed:
            return defer.succeed(self.signed)
        d = self.encode()
        d.addCallback(self._sign1)
        return d

    def _sign1(self, encoded):
        d = self.privkey.sign(encoded)
        d.addCallback(self._sign2)
        return d

    def _sign2(self, signed):
        self.signed = signed
        return signed

    def getArmored(self):
        d = self.sign()
        d.addCallback(self._armor2)
        return d

    def _armor2(self, signed):
        return util.armor(signed, "IDRECORD")


    def deliverMessage(self, mtype, msg):
        subscribers = self.subscribers.get(mtype, [])
        if not subscribers:
            self.queuedMessages.append((mtype, msg))
        else:
            for w in subscribers:
                w(mtype, msg)

    def deliverQueuedMessages(self):
        msgs = self.queuedMessages
        self.queuedMessages = []
        for mtype, msg in msgs:
            self.deliverMessage(mtype, msg)

    def subscribeMailbox(self, watcher, mtype="rfc822"):
        assert callable(watcher)
        assert type(mtype) == str or mtype is None
        if not self.subscribers.has_key(mtype):
            self.subscribers[mtype] = []
        self.subscribers[mtype].append(watcher)
        self.deliverQueuedMessages()

    def unsubscribeMailbox(self, watcher):
        for k in self.subscribers.keys():
            if watcher in self.subscribers[k]:
                self.subscribers[k].remove(watcher)



    # plugins
    def entryChanged(self, action, entry, lognode=None):
        return self.runPlugins("entryChanged", action, entry, lognode)

