
import cPickle
from StringIO import StringIO
from twisted.python.components import implements
from twisted.internet import defer

from petmail import util, interfaces, ticket
NotPermitted = interfaces.NotPermitted

# TODO: PermissionSlips need to point to what, exactly, they are giving
# permission *to*. This means an OwnerEntry, where the existence of the slip
# indicates permission to deliver rfc822 messages to that recipient. Need to
# decide if AddressBookEntries are scoped to a single OwnerEntry. If so,
# they can continue to have a single .theirPermission entry. If not, they
# need a list, and some way to indicate what each one is for.

# The code that makes this obvious is PermissionSlip.signed(), which needs
# to take a "us" argument to know which private key should sign the message.

# I will punt on this for now. Eventually I think PermissionSlips should be
# standalone objects that are signed independently, so that party A can
# prove to party B that they have some permission granted by party C. That
# may be useful for helping C recover some lost state, or to convince B that
# they should grant some related permission. To do that properly requires
# that the PermissionSlips be labeled meaningfully, and I don't yet know how
# that should work. So for now PermissionSlips are unlabeled, and derive all
# of their authority from the (signed) message that contain them.

class _PermissionCommon(object):

    def checkSizeAndIMG(self, msg, lastMsgTime):
        node = msg.getNode('rfc822')
        size = len(node['body'])
        contentType = "??"
        if self.maxsize and size > self.maxsize:
            why = "message is too large (%d > %d)" % (size, self.maxsize)
            raise NotPermitted, why
        # TODO: check self.contentTypes
        if self.minIMG and lastMsgTime:
            lapsed = msg.when - lastMsgTime
            if self.debug:
                print " IMG", self.minIMG
                print " lapsed", lapsed
            if lapsed < self.minIMG:
                raise NotPermitted("last message was sent too recently")


class OwnerPermission(_PermissionCommon):

    """The receiver uses this record to describe the permission it grants to
    a sender. This object appears as the .theirPermission attribute of an
    AddressBookEntry describing the given sender. An additional object,
    which describes permission granted to unknown senders, is stored in the
    OwnerEntry's .genericPermission attribute.

    A newly-created OwnerPermission (using __init__) gives blanket permission
    to everybody.
    """

    __implements__ = (interfaces.IOwnerPermission,
                      interfaces.IInboundPermission_rfc822)
    debug = 0

    def __init__(self):
        # this slip allows a sender to send mail to a receiver. These two
        # identities are implicit, and depend upon where this slip was found
        # (as the .theirPermission attribute of an AddressBookEntry, or as
        # the .genericPermission attribute of an IDRecord).

        self.lastModified = util.now()

        # message attributes of the permission
        #self.maxrate = None # unlimited
        self.minIMG = None # inter-message-gap (seconds), None == unlimited
        self.maxsize = None # unlimited
        self.contentTypes = ["*/*"] # either that or ["text/plain"], for now
        self.ticketTakers = {}

        # other attributes
        self.canGrantPermission = 1 # 3rd-party permission
        self.pedigree = None # where did this permission come from?

        # history, used to enforce maxrate
        self.lastReceived = None # time last message was received

    def copy(self):
        p = OwnerPermission()
        p.minIMG = self.minIMG
        p.maxsize = self.maxsize
        p.contentTypes = self.contentTypes[:]
        p.ticketTakers = self.ticketTakers.copy()
        p.canGrantPermission = self.canGrantPermission
        p.pedigree = self.pedigree
        return p

    def derive(self):
        """Create an OwnerPermission object that derives from us. In it's
        pedigree, we are the parent.
        """
        # TODO
        raise UnimplementedError()

    def modified(self):
        self.lastModified = util.now()

    def __repr__(self):
        url_string = self.getTicketTakersString()
        return "<OwnerPermission at 0x%x msize=%s URLs=%s>" \
               % (id(self), self.maxsize, url_string)

    def getProperty(self, pname):
        if pname == "acceptHTML":
            if "*/*" in self.contentTypes or "text/html" in self.contentTypes:
                return True
            return False
        if pname in ("minIMG", "maxsize", "canGrantPermission"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))

    def setProperty(self, pname, value):
        # TODO: don't call modified() if the value is being left alone?
        assert pname in ("minIMG", "maxsize", "acceptHTML",
                         "canGrantPermission")
        if pname == "minIMG":
            assert type(value) is int or value is None
            self.minIMG = value
        if pname == "maxsize":
            assert type(value) is int or value is None
            self.maxsize = value
        if pname == "acceptHTML":
            assert type(value) is bool
            if value:
                self.contentTypes = ["*/*"]
            else:
                self.contentTypes = ["text/plain"]
        if pname == "canGrantPermission":
            assert type(value) is bool
            self.canGrantPermission = value
        self.modified()

    def listTicketTakerURLs(self):
        urls = self.ticketTakers.keys()
        urls.sort()
        return urls


    def setMinIMG(self, img):
        # TODO: deprecate
        self.setProperty("minIMG", img)

    def setMaxSize(self, kb):
        # TODO: deprecate
        self.setProperty("maxsize", kb)

    def setContentTypes(self, types):
        # TODO: rethink, generalize, refactor
        self.contentTypes = types[:]
        self.modified()

    def encode(self):
        """Return a string which encodes this record."""
        d = {
            'IMG': self.minIMG,
            'maxsize': self.maxsize,
            'contentTypes': self.contentTypes,
            'ticketVendors': self.ticketTakers.keys(),
            'canGrantPermission': self.canGrantPermission,
            }
        return util.encode(d)

    def getTicketTakersString(self):
        return ",".join(self.listTicketTakerURLs())

    def setTicketTakersFromString(self, agent, str):
        new = []
        str = str.strip()
        for url in str.split(","):
            url.strip()
            if not url:
                continue
            if self.debug:
                print "setTicketTakersFromString: iter: '%s'" % url
            new.append(url)
            if url not in self.ticketTakers.keys():
                tt = ticket.TicketTaker(url, agent)
                self.addTicketTaker(tt)
        for url in self.ticketTakers.keys():
            if not url in new:
                self.removeTicketTaker(url)

    def addTicketTaker(self, taker):
        """The recipient calls this to add a new TicketTaker object to this
        slip. The sender who has access to this Permission Slip is thus
        allowed to earn entry to the mailbox by providing a ticket from the
        named URL."""
        self.ticketTakers[taker.url] = taker
        if self.debug:
            print "adding", taker
        self.modified()

    def removeTicketTaker(self, url):
        if self.debug:
            print "Removing", self.ticketTakers[url]
        del(self.ticketTakers[url])
        self.modified()

    def checkInboundMessage(self, msg):
        """Test an inbound message against our permissions. Raise an
        NotPermitted exception if the message is rejected for any reason.
        Verify that any necessary tickets are present.
        """
        assert implements(msg, interfaces.IReceivedMessage)
        node = msg.getNode('rfc822')
        self.checkSizeAndIMG(msg, self.lastReceived)

        why = "no ticket required"
        if self.ticketTakers:
            d = ticket.extractTickets(node['tickets'], msg.agent.crypto)
            d.addCallback(self.checkInboundMessage2, msg)
            return d
        # message is ok
        return defer.succeed(why)

    def checkInboundMessage2(self, tickets, msg):
        validTicket = 0
        if self.debug:
            print "checkInboundMessage: tickets:", tickets
        for (url, sigID, t) in tickets:
            try:
                taker = self.ticketTakers.get(url)
                if not taker:
                    why = "The ticket was issued by an unknown " + \
                          "vendor (%s)" % url
                    raise ticket.InvalidTicket, why
                # checkTicket raises InvalidTicket if it isn't valid
                taker.checkTicket(sigID, t)
                # ticket is valid
                validTicket = 1
                break
            except ticket.InvalidTicket, why:
                print "ticket was invalid:", why.args[0]
        if not validTicket:
            raise NotPermitted, "ticket required but not supplied"
        why = "valid ticket from %s" % url
        return why

    def acceptInboundMessage(self, msg):
        """This method should be invoked when a message is actually accepted.
        It will update the send history. Note that we should not track
        history on generic permissions, only on sender-specific ones.
        """
        self.lastReceived = msg.when

class PermissionSlip(_PermissionCommon):

    """The sender uses this record to describe the permission it has been
    granted to send mail to some particular receiver. This object appears as
    the .ourPermission attribute on the AddressBookEntry describing the
    recipient. Another instance appears as the .genericPermission attribute
    on the IDRecord, which is used when the sender does not have specific
    permission.

    This record is created by deserializing an IDRecord or the
    'yourPermission' node in an incoming message.
    """

    __implements__ = (interfaces.IPermissionSlip,
                      interfaces.IOutboundPermission_rfc822)
    debug = 0

    def __init__(self):
        self.watchers = []

        # message attributes of the permission
        self.minIMG = 0 # inter-message-gap (seconds), 0 means unlimited.
        self.maxsize = None # unlimited
        self.contentTypes = ["*/*"]
        self.ticketURLs = []

        # other attributes
        self.canGrantPermission = 1 # 3rd-party permission

        # history, used to enforce maxrate
        self.lastSent = None # time last message was sent

    def __getstate__(self):
        d = self.__dict__.copy()
        d['watchers'] = []
        return d

    def __repr__(self):
        url_string = ",".join(self.ticketURLs)
        url_string += "[%d]" % len(self.ticketURLs)
        return "<PermissionSlip msize=%s URLs=%s>" % (self.maxsize,
                                                      url_string)

    def subscribe(self, callback):
        self.watchers.append(callback)
    def unsubscribe(self, callback):
        self.watchers.remove(callback)

    def modified(self):
        for w in self.watchers:
            w(self)

    def getProperty(self, pname):
        if pname == "acceptHTML":
            if "*/*" in self.contentTypes or "text/html" in self.contentTypes:
                return True
            return False
        if pname in ("minIMG", "maxsize", "canGrantPermission"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))

    def listTicketTakerURLs(self):
        return self.ticketURLs

    def setFromEncoded(self, encoded):
        assert encoded
        d = util.decode(encoded)
        self.minIMG = d['IMG']
        self.maxsize = d['maxsize']
        self.contentTypes = d['contentTypes']
        self.ticketURLs = d['ticketVendors']
        self.canGrantPermission = d['canGrantPermission']
        self.modified()

    def checkOutboundMessage(self, msg):
        """Test an outbound message against these permissions. Raise an
        NotPermitted exception if the message cannot go through at all (it
        is too big, or too many have been sent recently). If tickets are
        needed, add the necessary information to the message's .requirements
        list.
        """
        assert implements(msg, interfaces.IOutboundMessage)
        requirements = []
        self.checkSizeAndIMG(msg, self.lastSent)
        if self.debug:
            print "checkOutboundMessage:", self.ticketURLs
        if self.ticketURLs:
            # note that this message will need a ticket
            note = ('ticket', self.ticketURLs)
            requirements.append(note)
        return requirements

    def sendOutboundMessage(self, msg):
        """This method should be invoked when a message is actually sent. It
        will update the send history.
        """
        self.lastSent = msg.when
