
"""
Define several built-in Processors to handle messages.
"""

"""
Message Processing (outbound):

Messages are created and populated with nodes, then handed to the Agent. The
Agent gives them a list of Processors that get to run (stored in
.processors).

Each processor has .dependencies and .provides lists to provide a sequence
in which they want to be run. Any given processor should be skipped if
everything it .provides is already present in the message. A processor
should not be run until its .dependencies are all provided by the message.

Each Processor can do one of three things (this might be merged into just
one thing later):

 return True, indicating that the processor has done its job immediately
 return a Deferred, which will callback or errback when the job is done
 raise Unsendable

If the Processor returns a Deferred, it is added to the message's
.processing list. When the Deferred completes, it will be removed from that
list.

Message.process figures out all Processors which can be run and starts them.
It

When the message begins processing, the agent puts in on a .processing list. 
"""


from twisted.internet import defer

from petmail import interfaces, util
from petmail.interfaces import Unsendable, NotPermitted
from petmail.idrecord import IDRecord

class BaseProcessor(object):
    __implements__ = interfaces.IProcessor
    dependencies = []
    provides = []
    def getDependencies(self):
        return self.dependencies
    def getProvides(self):
        return self.provides
    def configureSender(self, sender):
        raise Unimplemented
    def configureRecipient(self, recipient):
        raise Unimplemented
    def assertDependencies(self, message):
        for d in self.dependencies:
            if not d in message.provides:
                raise AssertionError("'%s' not in .provides=%s" \
                                     % (d, message.provides))

class OutboundProcessor(BaseProcessor):
    __implements__ = interfaces.IOutboundProcessor

class InboundProcessor(BaseProcessor):
    __implements__ = interfaces.IInboundProcessor






class RFC822Permission(OutboundProcessor):
    dependencies = ['recipient']
    def checkMessage(self, message, lognode=None):
        """Outbound rfc822 messages require Permission."""
        self.assertDependencies(message)

        node = message.getNode("rfc822")
        if not node:
            return True
        if not message.recipientRecord:
            raise Unsendable, "need ID Record to determine Permission"
        e = message.recipientEntry
        # do we need any kind of permission for this message?
        p = None
        if e:
            p = e.getOurPermission("rfc822")
            ptype = e.getOurPermissionType("rfc822")
        else:
            p = message.recipientRecord.genericPermission
            ptype = "(generic)"

        message.log("using %s permission" % ptype, parent=lognode)
        if message.agent.debugProcessing:
            print "Using %s Permission, IMG %s" % (ptype, p.minIMG)

        if not p:
            message.log("no permission, rejected", parent=lognode)
            message.log("rfc822 rejected: no permission")
            raise Unsendable, "this recipient does not accept new senders"

        try:
            reqs = p.checkOutboundMessage(message)
        except NotPermitted, why:
            message.log("permission check failed", why, parent=lognode)
            raise
        assert type(reqs) == type([])
        # checkMsg will raise a NotPermitted exception if a size limit is
        # violated. It will return requirements if there are tickets we must
        # obtain before sending
        if not reqs:
            p.sendOutboundMessage(message)
            message.log(ptype + " rfc822 allowed through: no tickets needed",
                        parent=lognode)
            return True
        # we need to fulfill this requirement
        message.log("must fulfill requirements", reqs, parent=lognode)
        if message.agent.debugProcessing:
            print "have reqs", reqs
        dl = []
        tas = message.agent.getTicketAgents()
        if not tas:
            raise Unsendable, ptype + \
                  " message needs tickets but we have no agents to get them"
        for ta in tas:
            deferreds = ta.handleRequirements(reqs, message)
            for d in deferreds:
                if isinstance(d, defer.Deferred):
                    d.addCallback(self.gotTicket, message, p, lognode)
                    d.addErrback(self.failed, message)
                    dl.append(d)
        # TODO: make sure the requirements were actually handled. This
        # probably involves having each ticketAgent tell us which
        # requirements they took care of (or started taking care of)
        if dl:
            # when all these Deferreds fires, the message is ready to go
            dlist = defer.DeferredList(dl, fireOnOneErrback=True)
            # the DeferredList is taking ownership of errors in the
            # individual Deferreds. To make sure no errors get left over,
            # add an errback which will drop them.
            [d.addErrback(lambda why: None) for d in dl]
            return dlist
        return True

    def gotTicket(self, ticket, message, p, lognode):
        message.log("adding ticket", parent=lognode)
        if message.agent.debugProcessing:
            print "gotTicket", ticket
        node = message.getNode('rfc822')
        if not node.has_key("tickets"):
            node["tickets"] = []
        node["tickets"].append(ticket) # TODO: maybe serialize here?
        p.sendOutboundMessage(message)
        return True
    def failed(self, why, message):
        return why


allOutbound = [RFC822Permission]

############################################################

# .senderEntry is looked up first, but set to None if they are unknown



allInbound = []

# outbound:
#  make sure there is a MessageID header node
#  require that the message has a recipientRecord
#  if the message has a recipientAddress but not a recipientRecord, do an
#   address lookup
#  make sure there is a Transport, get one if not
#  if their copy of the ID Record is out of date, include a copy
#  make sure there is a senderID node (or use the crypto envelope for that?)
#  if their permission record is out of date, include a copy
#  rfc822 messages need permission
#  return permission: if we're sending an rfc822 node to them, and they don't
#   have return permission, and the address book entry doesn't specifically
#   say we should deny them permission, then add Permission to their entry
#  3rd-party permission: if the message references other addresses to whom we
#   have permission, ask the user and then request 3rd-party permission for
#   them (keep track of the answer in the address book entry)

# inbound:
#  if they sent us an rfc822 message that was accepted, make sure there is
#   an address book entry for them
#  if they included an ID Record, make sure our copy is up to date (maybe ack)
#  if they included a Permission Slip, update our copy (maybe ack)
#  if there is an rfc822 node, verify the permission, and deliver to the user
#   (maybe ack)
#  if their message was not deemed undesireable, and they don't already have
#   permission, grant it (probably with a configurable timer)
#  if they requested 3rd-party permission, maybe grant it

# other inbound
# for address servers:
#   respond to addressLookup nodes
# for transport servers:
#   respond to transportLookup nodes
