
import urllib

from twisted.application import service, internet
from twisted.spread import pb
from twisted.python import log
from twisted.web import server, resource, util
from twisted.web.error import NoResource
from twisted.internet import defer

from petmail import pbutil, crypto
from petmail.idrecord import IDRecord
from petmail.util import armor

# code for address servers


class AddressServer(object):
    """This is a simple address server which uses PB as a wire protocol.
    """

    def __init__(self):
        self.crypto = crypto.EncryptionAgent("as")
        self.IDs = {}

    def lookup(self, address):
        log.msg("lookup(%s)" % address)
        for entry in self.IDs.values():
            if address in entry[0].addresses:
                return entry
        return None

    def lookupID(self, ID):
        log.msg("lookupID(%s)" % ID)
        return self.IDs.get(ID)

    def publish(self, s):
        log.msg("publish")
        r = IDRecord()
        d = r.addFromSigned(s, self.crypto)
        d.addCallback(self._publish2, s)
        return d

    def _publish2(self, r, s):
        log.msg(" (%s)" % r.addresses)
        entry = (r, s)
        self.IDs[r.ID] = entry
        return None

    def delete(self, ID):
        r = self.IDs.get(ID)
        if not r:
            log.msg("attempt to delete non-existent ID '%s'" % ID)
            return
        del self.IDs[ID]

    def allRecords(self):
        return self.IDs.values()

class PBAddressServer(pb.Root):
    """This is a simple address server which uses PB as a wire protocol.
    """

    def __init__(self, as):
        self.server = as

    def remote_lookup(self, address):
        entry = self.server.lookup(address)
        if entry:
            (r,s) = entry
            return s
        return None

    def remote_lookupID(self, ID):
        entry = self.server.lookupID(ID)
        if entry:
            (r,s) = entry
            return s
        return None

    def remote_publish(self, s):
        return self.server.publish(s)

class HtmlResource(resource.Resource):
    contentType = "text/html"

    def render(self, request):
        d = defer.maybeDeferred(self.content, request)
        d.addCallback(self._render, request)
        return server.NOT_DONE_YET
    def _render(self, data, request):
        request.setHeader("content-type", self.contentType)
        if request.method == "HEAD":
            request.setHeader("content-length", len(data))
        else:
            request.write(data)
        request.finish()

    title = "Dummy"
    def content(self, request):
        d = defer.maybeDeferred(self.body, request)
        d.addCallback(self._content)
        return d

    def _content(self, body):
        data = "<html>\n<head><title>" + self.title + "</title></head>\n"
        data += "<body vlink=\"#800080\">\n"
        data += body
        data += "</body></html>\n"
        return data

    def body(self, request):
        return "Dummy\n"

class WebAllRecords(HtmlResource):
    title = "PETMail ID Records"
    def __init__(self, as):
        HtmlResource.__init__(self)
        self.server = as

    def body(self, request):
        data = ""
        data += "<h1>All ID Records</h1>\n"

        records = self.server.allRecords()
        if records:
            data += "<ul>\n"
            for (r,s) in records:
                data += "<li>"
                link = urllib.quote(request.childLink(r.ID))
                data += "<a href=\"%s\">%s</a>" % (link, r.ID)
                if r.name:
                    data += " (%s)" % r.name
                data += "</li>\n"
            data += "</ul>\n"
        else:
            data += "<p>no records</p>\n"
        return data

class WebID(HtmlResource):
    title = "ID Record"
    def __init__(self, r, root):
        HtmlResource.__init__(self)
        self.record = r
        self.root = root

    def body(self, request):
        d = self.record.pubkey.getKey()
        d.addCallback(self._gotKey, request)
        return d
    
    def _gotKey(self, key, request):
        r = self.record
        data = ""
        data += "<h1>ID Record %s</h1>\n" % r.ID
        data += "<ul>\n"
        if r.name:
            data += "<li>Name:  %s</li>\n" % r.name
        if r.addresses:
            data += "<li>Addresses:  %s</li>\n" % ", ".join(r.addresses)
        if r.transports:
            data += "<li>Transport:  %s</li>\n" % r.transports
        data += "</ul>\n"

        gp = r.genericPermission
        if gp:
            data += "<h2>Generic Permission</h2>\n"
            data += "<ul>\n"
            if gp.maxsize == None:
                data += "<li>MaxSize:  Unlimited</li>\n"
            else:
                data += "<li>MaxSize:  %d</li>\n" % gp.maxsize
            data += "<li>Content-Types: %s</li>\n" % gp.contentTypes
            if gp.ticketURLs:
                data += "<li>Ticket URLs:  %s</li>\n" \
                        % ", ".join(gp.ticketURLs)
            data += "</ul>\n"

        data += "<h2>Armored Record</h2>\n"
        data += "<pre>\n%s</pre>\n" % armor(r.signed, "IDRECORD")

        data += "<h2>Delete</h2>\n"
        link = urllib.quote(request.childLink("delete"))
        data += "<a href=\"%s\">(delete this entry)</a>" % link

        return data

    def getChild(self, path, request):
        if path == "delete":
            self.root.server.delete(self.record.ID)
            return util.Redirect("../..")
        return NoResource("Unknown URL '%s'" % path)

class WebRoot(resource.Resource):
    def __init__(self, as):
        resource.Resource.__init__(self)
        self.server = as
        self.putChild("", WebAllRecords(self.server))
        self.putChild("all", WebAllRecords(self.server))

    def render(self, request):
        request.redirect(request.prePathURL() + '/')
        request.finish()

    def getChild(self, path, request):
        entry = self.server.lookupID(path)
        if entry:
            (r,s) = entry
            return WebID(r, self)
        return NoResource("No record with ID '%s'" % path)



def makeApp(pbport, webport):
    as = AddressServer()

    pbService = pbutil.listen(PBAddressServer(as), pbport)
    webService = internet.TCPServer(webport, server.Site(WebRoot(as)))

    a = service.Application("address-server")
    pbService.setServiceParent(a)
    webService.setServiceParent(a)
    return a
