
import os.path

from twisted.application import service
from twisted.spread import pb

from petmail import util, pbutil

# code for ticket issuers

class Ticket(object):
    def __init__(self, url, when=None, ticketID=None):
        self.url = url
        if when == None:
            self.timestamp = util.now()
        else:
            self.timestamp = when
        if ticketID == None:
            self.ticketID = util.generateMessageID()
        else:
            self.ticketID = ticketID

    def encode(self):
        d = {'URL': self.url,
             'Timestamp': self.timestamp,
             'TicketID': self.ticketID,
             }
        return util.encode(d)
    def signed(self, ourkey):
        """Returns a Deferred which fires with a signed version of the
        ticket."""
        return ourkey.sign(self.encode())


class BadResponse(pb.Error):
    """I will not grant a ticket because the response to my challenge was
    incorrect."""
    pass

import Image, ImageDraw, ImageFont, random, string
from cStringIO import StringIO

class Request(pb.Referenceable):
    def __init__(self, server):
        self.server = server
        letters = string.letters + string.digits
        length = 6
        self.text = "".join([random.choice(letters) for i in range(length)])

    def getChallenge(self):
        print "challenge:", self.text
        image = Image.new("RGB", (200, 100), "white")
        draw = ImageDraw.ImageDraw(image)
        font = ImageFont.truetype("font.ttf", 50)
        draw.setfont(font)
        draw.setink("blue")
        draw.text((10, 10), self.text) # top-left corner
        draw.line([(5,5), (190, 50)])
        #return image.tostring("png")
        # of course not
        f = StringIO()
        image.save(f, "PNG")
        return f.getvalue()

    def remote_respond(self, response):
        if response == self.text:
            return self.server.dispenseTicket()
        raise BadResponse

class FakeRequest(Request):
    def getChallenge(self):
        return self.text


class PBTicketServer(pb.Root):
    """This is a prototype Ticket Server which uses PB as a wire protocol
    and presents images (not distorted at all, yet) for the user to read.
    """

    def __init__(self, url, privkey, fake=False):
        self.url = url
        self.privkey = privkey
        self.fake = fake

    def remote_getTicket(self):
        print "remote_getTicket"
        if self.fake:
            r = FakeRequest(self)
        else:
            r = Request(self)
        challenge = r.getChallenge()
        return (r, challenge)

    def dispenseTicket(self):
        """I return a Deferred which fires with the signed/encoded form of
        the ticket, suitable for inclusion in a Message node."""
        print "dispenseTicket"
        t = Ticket(self.url)
        return t.signed(self.privkey)

    def remote_getPublicKey(self):
        print "remote_getPublicKey"
        return self.privkey.getKey()
    
    
from twisted.internet import defer

class PBTicketClient:
    """I am a client for PBTicketServer. You must provide the .displayImage
    method, and need to run .sendResponse with the string that the user has
    typed in."""

    def __init__(self, url):
        self.url = url

    def getTicket(self):
        self.deferred = defer.Deferred()
        PB, host, portnum = self.url.split(":")
        assert PB == "PB"
        d = pbutil.connect(host, int(portnum))
        d.addCallback(self.connected).addErrback(self.failed)
        return self.deferred

    def connected(self, perspective):
        d = perspective.callRemote("getTicket")
        d.addCallback(self.gotChallenge).addErrback(self.failed)

    def gotChallenge(self, (request, challenge)):
        self.remote = request
        self.displayImage(challenge)

    def displayImage(self, image):
        raise NotImplementedError

    def sendResponse(self, response):
        d = self.remote.callRemote("respond", response)
        d.addCallback(self.gotTicket).addErrback(self.failed)

    def gotTicket(self, ticket):
        self.deferred.callback(ticket)

    def failed(self, why):
        self.deferred.errback(why)

class PBTicketKeyClient:
    """I retrieve public keys from a PBTicketServer."""

    def __init__(self, url):
        self.url = url

    def getKey(self):
        self.deferred = defer.Deferred()
        PB, host, portnum = self.url.split(":")
        assert PB == "PB"
        d = pbutil.connect(host, int(portnum))
        d.addCallback(self.connected).addErrback(self.failed)
        return self.deferred

    def connected(self, perspective):
        d = perspective.callRemote("getPublicKey")
        d.addCallback(self.gotKey).addErrback(self.failed)

    def gotKey(self, key):
        self.deferred.callback(key)

    def failed(self, why):
        self.deferred.errback(why)

def makeApp(url):
    assert os.path.exists("font.ttf")
    from petmail import crypto
    from twisted.internet import reactor
    PB, hostname, portnum = url.split(":")
    assert PB == "PB"
    print "generating key.."
    cagent = crypto.EncryptionAgent("ts")
    d = cagent.generateKey("ticketserver")
    results = []
    def done(x):
        results.append(x)
        reactor.stop()
    d.addBoth(done)
    reactor.run()
    privkey = results[0]
    assert isinstance(privkey, crypto.PrivateKey)
    s = pbutil.listen(PBTicketServer(url, privkey), int(portnum))
    print "ticket server will run at url", url

    a = service.Application("ticket-server")
    s.setServiceParent(a)
    return a

if __name__ == '__main__':
    r = Request(None)
    img = r.getChallenge()
    open("/tmp/foo.png", "w").write(img)
    print "saved challenge image to /tmp/foo.png"

