
import sys, re, os
from twisted.trial import unittest
from twisted.python import log
from twisted.python.components import implements
from twisted.internet import reactor, defer
def dr(d, timeout=5):
    return unittest.deferredResult(d, timeout)
def de(d, timeout=5):
    return unittest.deferredError(d, timeout)

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import addressbook, permission, agent, interfaces, message, \
     owner, crypto, util, processors, names
from petmail.transport import dummy

class FakeAgent(agent.Agent):
    doNotScheduleEvents = True

class Inbox:
    def __init__(self, owner):
        self.messages = []
        owner.subscribeMailbox(self.deliver, "rfc822")
    def deliver(self, mtype, msg):
        self.messages.append(msg)

body = "A message body goes here"

class SingleAgent(unittest.TestCase):

    def setUp(self):
        self.oldUpdateHoldoffTime = None

    def tearDown(self):
        self.cleanAgentTimer()
        if self.oldUpdateHoldoffTime != None:
            old = self.oldUpdateHoldoffTime
            addressbook.AddressBookEntry.updateHoldoffTime = old

    def cleanAgentTimer(self):
        if self.agent.scheduledEventTimer:
            self.agent.scheduledEventTimer.cancel()
            self.agent.scheduledEventTimer = None

    def buildAgent(self):
        server = names.FakeNameServer()
        self.agent = FakeAgent('singleAgent')
        self.agent.populateProcessors()
        self.agent.crypto = crypto.EncryptionAgent()
        self.agent.registerAddressLookupAgent(server)
        self.agent.inbox = []

    def buildIdentity(self, name, address):
        assert " " not in name  # also used as a directory for generateKey

        us = dr(self.agent.createIdentity())
        us.setName(name)
        us.addAddress(address)
        us.setGenericPermission("rfc822", permission.OwnerPermission())
        us.setInlets([dummy.DummyInlet(name)])
        us.grantReturnPermission = False
        self.agent.addIdentity(us)
        return us

    def build(self):
        self.buildAgent()
        alice = self.buildIdentity("alice", "alice@example.com")
        bob = self.buildIdentity("bob", "bob@example.com")
        return (alice, bob)

    def addToAddressBook(self, owner, dest):
        signed = dr(dest.sign())
        book = self.agent.addressbooks[owner]
        entry = dr(book.addFromSigned(signed, "no reason"))
        return entry

    def buildMessage1(self, alice, bob):
        to_bob = self.addToAddressBook(alice, bob)
        
        m = message.OutboundMessage(self.agent)
        m.setSenderEntry(alice)
        m.setRecipientEntry(to_bob)
        m.addRFC822(body)

        return m

    # TODO: test sending to an entryless address
    #  test sending to an entryless IDRecord
    #  assert that the pre-encoded wire message has the right nodes
    #  assert that the 'why' argument to agent.acceptRFC822 is correct
    #  test agent.retrieveAddress, especially with multiple servers

    def test1(self):
        alice, bob = self.build()
        inbox = Inbox(bob)
        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        # at this point the message has been fully processed (both by the
        # sender and, due to the DummyTransport, also by the receiver).

        self.failUnlessEqual(len(inbox.messages), 1)
        m1 = inbox.messages[0]
        #self.failUnless("checkedRFC822" in m1.provides)
        self.failUnless(m1.allowRFC822)
        self.failUnless(m1.headers['rfc822'])
        self.failUnless(m1.headers['rfc822']['body'] == body)

        # bob doesn't know who the message is from
        self.failIf(m1.senderEntry)

        # alice should not have given permission to bob
        to_bob = self.agent.addressbooks[alice].lookupID(bob.ID)
        self.failUnless(to_bob)
        self.failIf(to_bob.theirPermission)

        # make sure bob can add the entry
        self.failUnless(m1.canAddSender())
        dr(m1.addSender())
        self.failUnless(m1.senderEntry)

    def test2(self):
        # verify that alice gives permission to Bob
        alice, bob = self.build()
        alice.grantReturnPermission = True
        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        to_bob = self.agent.addressbooks[alice].lookupID(bob.ID)
        self.failUnless(to_bob)
        self.failUnless(to_bob.theirPermission)

        
    def test3(self):
        raise unittest.SkipTest("this functionality was changed")
        # verify that bob gives inbound permission to alice
        alice, bob = self.build()

        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        back_to_alice = self.agent.addressbooks[bob].lookupID(alice.ID)
        self.failUnless(back_to_alice)
        self.failIf(back_to_alice.theirPermission)

        self.failUnless(len(self.agent.scheduledEvents), 1)
        ev = self.agent.scheduledEvents[0][1]
        self.failUnless("granting them permission to send more" in ev.why())
        #print "will fire in", self.agent.scheduledEvents[0][0] - util.now()
        
        #print "waiting for inbound permission holdoff"

        # now wait an additional three seconds for the holdoff-timer: 1 for
        # the timer, 1 for the "wake up a bit late" in checkScheduledEvents,
        # 1 for margin
        then = util.now() + 3
        while util.now() < then:
            reactor.iterate(0.1)
            
        self.failUnless(back_to_alice.theirPermission)

        # and there should be a scheduled event to update them
        self.failUnlessEqual(len(self.agent.scheduledEvents), 1)
        ev = self.agent.scheduledEvents[0][1]
        self.failUnless("Must update recipient" in ev.why())

    def test4(self):
        raise unittest.SkipTest("this functionality was changed")
        # verify that alice accepts that inbound permission
        alice, bob = self.build()
        bob.grantInboundPermission = True
        for p in self.agent.inboundProcessors:
            if isinstance(p, processors.GrantInboundPermission):
                p.permissionGrantHoldoffTime = 1
        # accelerate the update process. Must set this on the class because
        # a new instance will be created for the inbound message
        old = addressbook.AddressBookEntry.updateHoldoffTime
        self.oldUpdateHoldoffTime = old
        addressbook.AddressBookEntry.updateHoldoffTime = 1
        
        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        back_to_alice = self.agent.addressbooks[bob].lookupID(alice.ID)
        self.failUnless(back_to_alice)
        self.failIf(back_to_alice.theirPermission)

        self.failUnless(len(self.agent.scheduledEvents), 1)
        ev = self.agent.scheduledEvents[0][1]
        self.failUnless("granting them permission to send more" in ev.why())
        #print "will fire in", self.agent.scheduledEvents[0][0] - util.now()
        
        #print "waiting for inbound permission holdoff"

        # now wait an additional three seconds for the holdoff-timer: 1 for
        # the timer, 1 for the "wake up a bit late" in checkScheduledEvents,
        # 1 for margin
        then = util.now() + 3
        while util.now() < then:
            if back_to_alice.theirPermission:
                break
            reactor.iterate(0.1)
        self.failUnless(back_to_alice.theirPermission)

        to_bob = self.agent.addressbooks[alice].lookupID(bob.ID)
        self.failIf(to_bob.ourPermission)
        # now wait three seconds for the return message to propogate
        then = util.now() + 3
        while util.now() < then:
            if to_bob.ourPermission:
                break
            reactor.iterate(0.1)
        self.failUnless(to_bob.ourPermission)

class NewAgentTest(unittest.TestCase):
    def buildAgent(self, basedir, nameserver):
        a = agent.Agent(basedir)
        a.crypto = crypto.EncryptionAgent()
        a.registerAddressLookupAgent(nameserver)
        a.populateProcessors()
        a.doNotScheduleEvents = True
        return a

    def buildIdentity(self, agent, name, address, acceptMail=True):
        assert " " not in name  # also used as a directory for generateKey

        perm = None
        if acceptMail:
            perm = permission.OwnerPermission()

        us = dr(agent.createIdentity(perm=perm))
        us.setName(name)
        us.addAddress(address)
        us.setInlets([dummy.DummyInlet(name)])
        #us.grantReturnPermission = False
        agent.addIdentity(us)
        return us

    def setUp(self):
        nameserver = names.FakeNameServer()

        a1 = self.a1 = self.buildAgent("a1", nameserver)
        a1o1 = self.a1o1 = self.buildIdentity(a1, "a1o1", "a1o1@")
        a1o2 = self.a1o2 = self.buildIdentity(a1, "a1o2", "a1o2@")

        a2 = self.a2 = self.buildAgent("a2", nameserver)
        a2o1 = self.a2o1 = self.buildIdentity(a2, "a2o1", "a2o1@")
        a2o2 = self.a2o2 = self.buildIdentity(a2, "a2o2", "a2o2@")
        a2o3 = self.a2o3 = self.buildIdentity(a2, "a2o3", "a2o3@",
                                              acceptMail=False)

        a3 = self.a3 = self.buildAgent("a3", nameserver)
        a3o1 = self.a3o1 = self.buildIdentity(a3, "a3o1", "a3o1@")

    def tearDown(self):
        # push out any pending publishes, to clear the dirtyTimers
        dr(self.a1o1.publish())
        dr(self.a1o2.publish())
        dr(self.a2o1.publish())
        dr(self.a2o2.publish())
        dr(self.a2o3.publish())
        dr(self.a3o1.publish())

    def addToAddressBook(self, owner, dest):
        signed = dr(dest.sign())
        book = self.agent.addressbooks[owner]
        entry = dr(book.addFromSigned(signed, "no reason"))
        return entry

    def addAddress(self, owner, addr):
        agent = owner.getAgent()
        signed = dr(agent.retrieveAddress(addr))
        book = agent.getAddressBook(owner.ID)
        entry = dr(book.addFromSigned(signed))
        self.failUnless(entry)
        self.failUnless(entry is book.lookupID(entry.idRecord.ID))
        return entry

    def buildMessage1(self, recip, anonymous=False):
        sender = recip.getOwner()
        agent = sender.getAgent()
        m = message.OutboundMessage(agent)
        m.setSenderEntry(sender, anonymous=anonymous)
        m.setRecipientEntry(recip)
        return m

    def buildMessage2(self, sender, anonymous=False):
        agent = sender.getAgent()
        m = message.OutboundMessage(agent)
        m.setSenderEntry(sender, anonymous=anonymous)
        return m
        
    def test1(self):
        # test Agent.sendMessage
        #  and that the message has a messageID node
        inbox = Inbox(self.a2o1)
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        #self.a1.debugProcessing = True
        #self.a2.debugProcessing=True
        #log.startLogging(sys.stderr)
        dr(self.a1.sendMessage(m))
        self.failUnless(len(inbox.messages) == 1)
        rm = inbox.messages[0]
        self.failUnlessEqual(rm.getNode("rfc822")['body'], "message body")
        self.failUnless(rm.getNode("senderIDRecord"))
        self.failUnless(rm.getNode("messageID"))

    def testSendMessageToAddress(self):
        # tests Agent.sendMessageToAddress
        # also tests CreateSenderEntry code in Agent._sendMessageToAddress2
        inbox = Inbox(self.a2o1)

        tid = self.a2o1.ID
        book = self.a1.getAddressBook(self.a1o1.ID)

        m = self.buildMessage2(self.a1o1) # no .recipientEntry
        dr(self.a1.sendMessageToAddress(m, "a2o1@"))
        # no rfc822 node, so no Entry should be created
        self.failIf(book.lookupID(tid))

        # now include an rfc822 node. The act of sending should add an
        # AddressBookEntry

        m = self.buildMessage2(self.a1o1) # no .recipientEntry
        m.addRFC822("message body")
        dr(self.a1.sendMessageToAddress(m, "a2o1@"))
        # the act of sending should add an AddressBookEntry
        entry = book.lookupID(tid)
        self.failUnless(entry)
        self.failUnlessEqual(entry.idRecord.name, self.a2o1.names)
        self.failUnless(entry.theirPermission) # grants default
        # TODO: verify other fields of the new entry

        # make sure the rfc822 got delivered
        self.failUnless(len(inbox.messages) == 1)
        rm = inbox.messages[0]
        self.failUnlessEqual(rm.getNode("rfc822")['body'], "message body")
        self.failUnless(rm.getNode("senderIDRecord"))

    def testFailingAddressLookup(self):
        # bad addresses are unsendable
        tid = self.a2o1.ID
        m = self.buildMessage2(self.a1o1) # no .recipientEntry
        m.addRFC822("message body")
        e = de(self.a1.sendMessageToAddress(m, "bogus@"))
        self.failUnless(e.check(interfaces.Unsendable))

    def test3(self):
        # recipient did not grant Generic Permission for rfc822 messages
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o3@")
        inbox = self.a2.getInbox(self.a2o3)

        allMessages = []
        self.a2.inboundMessageHandlers = [allMessages.append]

        # messages without rfc822 nodes should be ok
        m1 = self.buildMessage1(dest)
        dr(self.a1.sendMessage(m1))
        self.failUnlessEqual(len(allMessages), 1)
        #rm = allMessages[0] ; print rm.headers


        # but messages with rfc822 nodes should be rejected on the sender
        # side
        m2 = self.buildMessage1(dest)
        m2.addRFC822("message body")
        e = de(self.a1.sendMessage(m2))
        self.failUnless(e.check(interfaces.Unsendable))
        self.failUnlessEqual(e.value.args[0],
                             "this recipient does not accept new senders")

        # recipients which require a ticket should cause a ticket request


    def testUpdateTheirIDRecord1(self):

        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]

        # sending a message through an IDRecord (no entry) should always
        # include a copy. The easiest way to do this is to send a non-rfc822
        # message through an address. This hits the OwnerPlugin
        # UpdateTheirIDRecord
        m = self.buildMessage2(self.a1o1)
        dr(self.a1.sendMessageToAddress(m, "a2o1@"))
        m = inbox.pop()
        self.failUnless(m.getNode("senderIDRecord"))

        # unless, of course, the message is anonymous
        m = self.buildMessage2(self.a2o1, anonymous=True)
        m.setRecipientRecord(dest.idRecord)
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failIf(m.getNode("senderIDRecord"))


    def testUpdateTheirIDRecord2(self):
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]

        # now we send through the Entry. This hits the Entry's plugin. The
        # first case is where we don't have specific permission from the
        # recipient, which means they probably don't have us in their
        # address book

        # sending through the Entry can still be done anonymously, and of
        # course should not include a copy
        m = self.buildMessage1(dest, anonymous=True)
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failIf(m.getNode("senderIDRecord"))

        # the first non-anonymous message should always include a copy
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failUnless(m.getNode("senderIDRecord"))

        # the second should too, because we don't have specific permission
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failUnless(m.getNode("senderIDRecord"))


    def testUpdateTheirIDRecord3(self):
        # this time, we do have specific permission
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        dest.ourPermission = permission.PermissionSlip()
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]

        # sending anonymously does not include a copy
        # the timestamp (so the next send should still include a copy)
        m = self.buildMessage1(dest, anonymous=True)
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failIf(m.getNode("senderIDRecord"))

        # the first non-anonymous message should always include a copy
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failUnless(m.getNode("senderIDRecord"))

        # the second message should not, because we remember sending it to
        # them just a moment ago
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failUnless(not m.getNode("senderIDRecord"),
                        "unnecessary senderIDRecord was included")

        # modifying our Owner data should schedule an update, which will
        # piggyback on any other messages heading their way
        util.warpForward(1)
        self.a1o1.modified()
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        m = inbox.pop()
        self.failUnless(m.getNode("senderIDRecord"))

        # TODO: test the fact that the send gets scheduled

    def testUpdateMyIDRecord1(self):
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]
        sender = self.a1o1

        # send a message, the recipient adds the entry
        m = self.buildMessage2(sender)
        dr(self.a1.sendMessageToAddress(m, "a2o1@"))
        rm = inbox.pop()
        self.failUnless(rm.getNode("senderIDRecord"))

        self.failUnless(rm.canAddSender())
        dr(rm.addSender())
        recip = self.a2.getAddressBook(self.a2o1.ID).lookupID(self.a1o1.ID)
        self.failUnless(recip)
        self.failUnlessEqual(recip.idRecord.name, sender.names)

        # update the record
        self.a1o1.setName("bob")
        assert sender.names == "bob"
        # send message, should update
        m = self.buildMessage2(sender)
        dr(self.a1.sendMessageToAddress(m, "a2o1@"))
        rm = inbox.pop()
        self.failUnless(rm.getNode("senderIDRecord"))
        self.failUnlessEqual(recip.idRecord.name, sender.names)

    def testReturnPermission(self):
        # tests the ReturnPermission entry-plugin

        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]

        # sending a non-rfc822-bearing message should not auto-grant
        # permission
        m = self.buildMessage1(dest)
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failIf(dest.theirPermission)
        self.failIf(rm.getNode("yourPermission"))

        # but the act of sending an rfc822 message *should* grant it
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(dest.theirPermission)
        self.failUnless(implements(dest.theirPermission,
                                   interfaces.IOwnerPermission))
        # it should also include a copy of their permission slip
        self.failUnless(rm.hasNode("yourPermission"))
        # check that it looks vaguely correct
        perm = permission.PermissionSlip()
        perm.setFromEncoded(rm.getNode("yourPermission"))
        self.failUnlessEqual(perm.maxsize, dest.theirPermission.maxsize)

        # and sending a second one shouldn't blow up or anything. In
        # particular it should not modify the existing permission, say by
        # re-granting a new one. We tag the object to make sure.
        dest.theirPermission.tagyourit = 34
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(dest.theirPermission)
        self.failUnless(dest.theirPermission.tagyourit == 34)
        # it should also not re-send the permission
        self.failIf(rm.hasNode("yourPermission"))

        # changing the permission should make it re-send
        util.warpForward(1)
        dest.theirPermission.setMaxSize(12) # does .modified
        m = self.buildMessage1(dest)
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(rm.hasNode("yourPermission"))
        # check that it looks vaguely correct
        perm = permission.PermissionSlip()
        perm.setFromEncoded(rm.getNode("yourPermission"))
        self.failUnlessEqual(perm.maxsize, 12)


        # the entry's "never auto-grant permission" mark should be honored
        dest = self.addAddress(self.a1o1, "a2o2@")
        dest.doNotGrantReturnPermission = True
        m = self.buildMessage1(dest)
        m.addRFC822("message body")
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failIf(dest.theirPermission)
        self.failIf(rm.hasNode("yourPermission"))


    def testUpdateOurPermission(self):
        # tests the UpdateOurPermission entry-plugin

        tid = self.a2o1.ID
        sender = self.a1o1
        sender.setGenericPermission("rfc822", None)
        dest = self.addAddress(self.a1o1, "a2o1@")
        sender.grantPermissionTo(dest)
        dest.getTheirPermission("rfc822").setMaxSize(1234)
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]

        # send a message, the recipient adds the entry
        m = self.buildMessage1(dest)
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(rm.hasNode("yourPermission"))
        self.failUnless(rm.canAddSender())
        dr(rm.addSender())
        recip = self.a2.getAddressBook(self.a2o1.ID).lookupID(sender.ID)
        self.failUnless(recip)
        self.failUnlessEqual(recip.idRecord.name, sender.names)
        perm = recip.getOurPermission("rfc822")
        self.failUnless(perm)
        self.failUnlessEqual(perm.maxsize, 1234)
        self.failUnlessEqual(recip.getOurPermissionType("rfc822"), "specific")

        util.warpForward(1)
        dest.getTheirPermission("rfc822").setMaxSize(8877)
        # send a message, permission should be updated
        m = self.buildMessage1(dest)
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(rm.hasNode("yourPermission"))
        perm = recip.getOurPermission("rfc822")
        self.failUnlessEqual(perm.maxsize, 8877)
        self.failUnlessEqual(recip.getOurPermissionType("rfc822"), "specific")

        util.warpForward(1)
        dest.revokeTheirPermission("rfc822")
        dest.modified()
        # send a message, permission should be revoked
        m = self.buildMessage1(dest)
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(rm.hasNode("yourPermission"))
        perm = recip.getOurPermission("rfc822")
        self.failIf(perm)
        self.failUnlessEqual(recip.getOurPermissionType("rfc822"), "generic")

    def testReferences(self):
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        referenced = self.addAddress(self.a1o1, "a3o1@")
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]
        # TODO: need to test vouchers too (['offerPermission']=True)

        m = self.buildMessage1(dest)
        m.addReference(referenced)
        dr(self.a1.sendMessage(m))
        rm = inbox.pop()
        self.failUnless(rm.hasNode("references"))
        refs = rm.getNode("references")
        self.failUnlessEqual(len(refs), 1)
        ref = refs[0]
        self.failIf(ref['offerPermission'])
        self.failUnlessEqual(ref['name'], '"a3o1"')
        self.failUnlessEqual(ref['id'], referenced.idRecord.ID)

        # TODO: it might make sense to get rid of owner.HandleReferences
        self.failUnless(hasattr(rm, 'references'))
        self.failUnlessEqual(len(rm.references), 1)
        tref = rm.references[0]
        self.failIf(ref['offerPermission'])
        self.failUnlessEqual(ref['name'], '"a3o1"')
        self.failUnlessEqual(ref['id'], referenced.idRecord.ID)


    def testPing(self):
        tid = self.a2o1.ID
        dest = self.addAddress(self.a1o1, "a2o1@")
        inbox = []
        self.a2.inboundMessageHandlers = [inbox.append]

        d = self.a1.pingRecipient(dest)
        pong = dr(d)

        self.failUnless(pong)
        self.failUnless(pong.hasNode("pong"))
        self.failUnlessEqual(len(inbox), 1)
        rm = inbox.pop()
        self.failUnlessEqual(rm.getNode("ping"), pong.getNode("pong"))

# Local Variables:
# compile-command: "make -C ../.. test-emacs"
# End:
