
import os
from StringIO import StringIO

from twisted.internet import reactor
import signal
from twisted.trial import unittest
dr = unittest.deferredResult

from petmail import util, crypto

# the real key generation setup is expensive (10-20 seconds), so only do it
# once. When using the resulting agents, be careful to not modify them in any
# way (like adding pubkeys) that will influence later tests.

def setup(cryptoClass):
    agent1 = cryptoClass("test_crypto_1")
    priv1 = dr(agent1.generateKey("key1"), 60)
    keyid1 = priv1.getID()
    pub1 = crypto.PublicKey(agent1, keyid1)

    agent2 = cryptoClass("test_crypto_2")
    priv2 = dr(agent2.generateKey("key2"), 60)
    keyid2 = priv2.getID()
    pub2 = crypto.PublicKey(agent2, keyid2)

    return (agent1, agent2, priv1, pub1, priv2, pub2)

gnupgStuff = None
def setupGnuPG():
    # TODO: pay attention to how this gets cleaned up (if at all)
    global gnupgStuff
    if not gnupgStuff:
        gnupgStuff = setup(crypto.GnuPGEncryptionAgent)
    return gnupgStuff

fakeStuff = None
def setupFake():
    global fakeStuff
    if not fakeStuff:
        fakeStuff = setup(crypto.FakeEncryptionAgent)
    return fakeStuff

class SignalMixin:
    # This class is necessary for any code which wants to use Processes
    # outside the usual reactor.run() environment. It is copied from
    # Twisted's twisted.test.test_process
    sigchldHandler = None
    
    def setUpClass(self):
        # make sure SIGCHLD handler is installed, as it should be on
        # reactor.run(). problem is reactor may not have been run when this
        # test runs.
        if hasattr(reactor, "_handleSigchld") and hasattr(signal, "SIGCHLD"):
            self.sigchldHandler = signal.signal(signal.SIGCHLD,
                                                reactor._handleSigchld)
    
    def tearDownClass(self):
        if self.sigchldHandler:
            signal.signal(signal.SIGCHLD, self.sigchldHandler)

class CryptoTestMixin(SignalMixin):

    def dr(self, d):
        return unittest.deferredResult(d, 60)

    def test1st(self):
        # this is run first, and exists to make it visually clear where/when
        # the key-generation takes place
        pass

    def testExport(self):
        # export a new to a new agent
        agent = self.cryptoClass("export2")

        key = self.dr(self.priv1.getKey())
        newkey = self.dr(agent.addPubkey(key))
        self.failUnlessEqual(self.priv1.getID(), newkey.getID())

        # importing a known key should not change anything
        newkey = self.dr(agent.addPubkey(key))
        self.failUnlessEqual(self.priv1.getID(), newkey.getID())


    def testSign1(self): # covers sign/verify
        # signed by a key we know
        agent = self.cryptoClass("sign1")
        agent.addPubkey(self.dr(self.priv1.getKey()))

        # test encoding while we're at it, might catch a text-vs-binary bug
        d = {'Foo': "blah", 'Bar': "bloo boo"}
        plaintext = util.encode(d)
        signed = self.dr(self.priv1.sign(plaintext))
        (recip, sigID, p2) = self.dr(agent.decrypt(signed))
        self.failUnlessEqual(recip, None)
        self.failUnlessEqual(sigID, self.priv1.getID())
        d2 = util.decode(p2)
        self.failUnlessEqual(d, d2)

    def testSign2(self):
        # signed by a key we don't know
        agent = self.cryptoClass("sign2")

        plaintext = "as plain as the nose on your face"
        signed = self.dr(self.priv1.sign(plaintext))
        (recip, sigID, plaintext2) = self.dr(agent.decrypt(signed))
        self.failUnlessEqual(recip, None)
        self.failUnlessEqual(sigID, None)
        self.failUnlessEqual(plaintext, plaintext2)

        # now add the key and make sure it starts reporting the sigID
        newkey = self.dr(agent.addPubkey(self.dr(self.priv1.getKey())))
        self.failUnlessEqual(self.priv1.getID(), newkey.getID())

        signed = self.dr(self.priv1.sign(plaintext))
        (recip, sigID, plaintext2) = self.dr(agent.decrypt(signed))
        self.failUnlessEqual(recip, None)
        self.failUnlessEqual(sigID, self.priv1.getID())
        self.failUnlessEqual(plaintext, plaintext2)

    def testSign3(self): # covers sign/decrypt
        # signed by a key we know
        agent = self.cryptoClass("sign1")
        agent.addPubkey(self.dr(self.priv1.getKey()))

        plaintext = "plane nonsense"
        signed = self.dr(self.priv1.sign(plaintext))
        (recip, sigID, p2) = self.dr(agent.decrypt(signed))
        self.failUnlessEqual(recip, None)
        self.failUnlessEqual(sigID, self.priv1.getID())
        self.failUnlessEqual(p2, plaintext)


    def testEncrypt1(self):
        # covers encryptAndSign/encrypt/decrypt

        # encrypt to ourselves
        plaintext = "plain plain mary jane"
        encrypted = self.dr(self.pub1.encryptAndSign(self.priv1, plaintext))
        (recip, sigID, p2) = self.dr(self.agent1.decrypt(encrypted))
        self.failUnlessEqual(recip.getID(), self.priv1.getID())
        self.failUnlessEqual(sigID, self.priv1.getID())
        self.failUnlessEqual(p2, plaintext)

        # encrypt to ourselves, no signature
        plaintext = "metallic birds bright upon the horizon"
        encrypted = self.dr(self.pub1.encrypt(plaintext))
        (recip, sigID, p2) = self.dr(self.agent1.decrypt(encrypted))
        self.failUnlessEqual(recip.getID(), self.priv1.getID())
        self.failUnlessEqual(sigID, None)
        self.failUnlessEqual(p2, plaintext)

    def testEncrypt2(self):
        # encrypted to us but from an unknown sender

        # Tell agent1 about pub2, so it can encrypt to priv2, or verify
        # signatures from priv2
        key2 = self.dr(self.agent1.addPubkey(self.dr(self.pub2.getKey())))

        plaintext = "spain, mostly"
        # encrypt to priv2
        encrypted = self.dr(key2.encryptAndSign(self.priv1, plaintext))

        # now we play the role of agent2. sigID should be None because we
        # don't yet know about the sender
        (recip, sigID, p2) = self.dr(self.agent2.decrypt(encrypted))
        self.failUnlessEqual(recip.getID(), self.priv2.getID())
        self.failUnlessEqual(sigID, None)
        self.failUnlessEqual(p2, plaintext)

        # now add the key and make sure sigID starts showing up
        key = self.dr(self.pub1.getKey())
        newkey = self.dr(self.agent2.addPubkey(key))
        self.failUnlessEqual(self.priv1.getID(), newkey.getID())

        (recip, sigID, p2) = self.dr(self.agent2.decrypt(encrypted))
        self.failUnlessEqual(recip.getID(), self.priv2.getID())
        self.failUnlessEqual(sigID, self.priv1.getID())
        self.failUnlessEqual(p2, plaintext)

    def testStore1(self):
        # no encryption, just compress/encode/armor
        plaintext = "plain bagel, not toasted"
        encrypted = self.dr(self.agent1.store(plaintext))
        (recip, sigID, p2) = self.dr(self.agent2.decrypt(encrypted))
        self.failUnlessEqual(recip, None)
        self.failUnlessEqual(sigID, None)
        self.failUnlessEqual(p2, plaintext)


class GnuPGCryptoTest(CryptoTestMixin, unittest.TestCase):
    cryptoClass = crypto.GnuPGEncryptionAgent

    def setUp(self):
        (self.agent1, self.agent2,
         self.priv1, self.pub1,
         self.priv2, self.pub2) = setupGnuPG()

        
class FakeCryptoTest(CryptoTestMixin, unittest.TestCase):
    cryptoClass = crypto.FakeEncryptionAgent

    def setUp(self):
        (self.agent1, self.agent2,
         self.priv1, self.pub1,
         self.priv2, self.pub2) = setupFake()

if os.environ.get("PETMAIL_SKIP_GPG"):
    GnuPGCryptoTest.skip = "skipping long gpg crypto test"
