
import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import agent, util, permission, ticket, crypto
from petmail import interfaces
from petmail.owner import OwnerEntry
from petmail.idrecord import IDRecord

from twisted.trial import unittest
dr = unittest.deferredResult

class TestOwnerEntry(unittest.TestCase):
    def setUp(self):
        self.crypto = crypto.EncryptionAgent()
        self.agent = agent.Agent("test_idrecord")

    def dr(self, d):
        return unittest.deferredResult(d, 60)

    def build1(self):
        e = OwnerEntry()
        e.doPublish = False
        self.agent.addOwner(e)
        e.setKey(dr(self.crypto.generateKey("alice"), 60))
        e.setName("alice")
        e.addAddress("alice@foo.com")
        e.setGenericPermission("rfc822", None) # no new senders are accepted

        e.setInlets([self.agent.newInlet("dummy", name="alice"),
                     self.agent.newInlet("dummy", name="alice2")])
        e.setInlets([self.agent.newInlet("dummy", name="alice2")])
        e.setInlets([self.agent.newInlet("dummy", name="alice"),
                     self.agent.newInlet("dummy", name="alice2")])
        i = e.getInlets()
        i.sort()
        i2 = [self.agent.newInlet("dummy", name="alice"),
              self.agent.newInlet("dummy", name="alice2")]
        i2.sort()
        self.failUnlessEqual(i, i2)
        return e

    def build2(self):
        e = self.build1()
        p0 = permission.OwnerPermission()
        p0.minIMG = 3600
        p0.maxsize = 10**6
        tt_privkey = dr(self.crypto.generateKey("ticketserver.com"), 60)
        tt_pubkey = tt_privkey.pubkey()
        v = ticket.TicketTaker("http://ticketserver.com", None, tt_pubkey)
        p0.addTicketTaker(v)
        e.setGenericPermission("rfc822", p0)
        return e

    def build3(self):
        e = self.build2()
        i0 = self.agent.newInlet("pb", hostname="localhost", portnum=9999)
        e.setInlets([i0])
        return e

    # OwnerEntries are serialized into strings, and then unserialized into
    # IDRecords.
    def testEncode1(self):
        e = self.build1()
        r0 = dr(e.sign())

        recip, sigID, d = dr(self.crypto.decrypt(r0), 60)
        r1 = util.decode(d)
        self.failUnlessEqual(sigID, r1['ID'])
        pk = dr(e.privkey.getKey())
        self.failUnlessEqual(pk, r1['pubkey'])
        self.failUnlessEqual(r1['name'], "alice")
        self.failUnlessEqual(r1['addresses'], ["alice@foo.com"])
        self.assertApproximates(r1['expires'], util.now()+120, 60)
        self.failUnlessEqual(r1['transports'], [("Dummy", "alice"),
                                                ("Dummy", "alice2")])

    def testEncode2(self):
        e = self.build2()
        r0 = dr(e.encode())

        r1 = util.decode(r0)

        gp = util.decode(r1['permission'])
        self.failUnlessEqual(gp['IMG'], 3600)
        self.failUnlessEqual(gp['maxsize'], 10**6)
        self.failUnlessEqual(gp['contentTypes'], ["*/*"])
        self.failUnlessEqual(gp['canGrantPermission'], 1)

    def testEncode3(self):
        e = self.build3()
        r0 = dr(e.sign())

        recip, sigID, d = dr(self.crypto.decrypt(r0), 60)
        r1 = util.decode(d)

        self.failUnlessEqual(r1['transports'], [("PB", "localhost", 9999)])

    def testEncode4(self):
        e = self.build1()
        r0 = dr(e.sign())
        r1 = dr(e.sign()) # should be cached and therefore identical
        self.failUnlessEqual(r0, r1)

        e.setName("bob") # should invalidate the cache
        r2 = dr(e.sign())
        self.failIfEqual(r1, r2)

    def testArmor1(self):
        e = self.build1()
        a = dr(e.getArmored())
        self.failUnless(a.find("BEGIN PETMAIL IDRECORD") != -1)

    def testDecode1(self):
        r0 = self.build1()
        e = dr(r0.sign())

        r1 = IDRecord()
        d = r1.createFromSigned(e, self.crypto)
        self.dr(d)
        self.failUnless(isinstance(r1, IDRecord))
        self.failUnlessEqual(r0.ID, r1.ID)
        self.failUnlessEqual(r0.names, r1.name)
        self.failUnlessEqual(r0.addresses, r1.addresses)

    def testDecode2(self):
        r0 = self.build1()
        e = dr(r0.sign())

        r1 = IDRecord()
        d = r1.addFromSigned(e, self.crypto)
        self.dr(d)
        self.failUnless(isinstance(r1, IDRecord))
        self.failUnlessEqual(r0.ID, r1.ID)
        self.failUnlessEqual(r0.privkey.getID(), r1.pubkey.getID())
        self.failUnlessEqual(r0.names, r1.name)
        self.failUnlessEqual(r0.addresses, r1.addresses)

    def testPseudonymous(self):
        e = self.build1()

        e.setPseudonymous(True)
        self.failUnless(e.pseudonymous)

        e.setPseudonymous(False)
        self.failIf(e.pseudonymous)

        e.hasBeenRevealed = True
        self.failUnlessRaises(interfaces.CatOutOfBagError,
                              e.setPseudonymous, True)
        self.failIf(e.pseudonymous)

        e.setPseudonymous(False) # should be ok
        e.setPseudonymous(True, yesReally=True) # overrides warning
        self.failUnless(e.pseudonymous)

        
