
from petmail import log
from twisted.trial import unittest
import time

class LogTest(unittest.TestCase):
    def testBasics(self):
        l = log.Log("foo", maxItems=3)
        l.log("one")
        l.log("two")
        l.log("three")
        self.failUnlessEqual(len(l.getItems()), 3)
        l.log("four") # should displace "one"
        items = l.getItems()
        self.failUnlessEqual(len(items), 3)
        self.failUnlessEqual(items[0], l.getItem(0))
        m0 = items[0].getMessage()
        self.failUnlessEqual(type(m0), tuple)
        self.failUnlessEqual(m0[0], "two")
        self.failUnlessEqual(l.getMessage(0)[0], "two")
        self.failUnlessEqual(l.getMessage(-1)[0], "four")

    def testOnePriority(self):
        l = log.Log("foo", numClasses=1, maxItems=3)
        l.log("one", pclass=0)
        l.log("two", pclass=0)
        l.log("three", pclass=0)
        l.log("four", pclass=0)
        items = l.getItems()
        self.failUnlessEqual(len(items), 3)
        self.failUnlessEqual(l.getMessage(0), ("two",))
        self.failUnlessEqual(l.getMessage(-1), ("four",))

    def testPriorities(self):
        l = log.Log("foo", numClasses=3, maxItems=3)
        l.log("a", pclass=0) # 0a
        self.failUnlessEqual(len(l.getItems(None)), 1)
        self.failUnlessEqual(len(l.getItems(0)), 1)
        self.failUnlessEqual(len(l.getItems(1)), 0)
        self.failUnlessEqual(l.getMessage(0, None), ("a",))
        self.failUnlessEqual(l.getMessage(0, 0), ("a",))
        self.failUnlessRaises(IndexError, l.getMessage, 0, 1)

        l.log("b", pclass=0) # 0a 0b
        l.log("c", pclass=1) # 0a 0b 1c
        l.log("d", pclass=2) # 0a 0b 1c 2d
        l.log("e", pclass=0) # 0a 0b 1c 2d 0e
        l.log("f", pclass=1) # 0a 0b 1c 2d 0e 1f
        self.failUnlessEqual(len(l.getItems(None)), 6)
        self.failUnlessEqual(len(l.getItems(0)), 3)
        self.failUnlessEqual(len(l.getItems(1)), 2)
        self.failUnlessEqual(len(l.getItems(2)), 1)
        self.failUnlessEqual(l.getMessage(-1, None), ("f",))
        self.failUnlessEqual(l.getMessage(-1, 0), ("e",))
        self.failUnlessEqual(l.getMessage(-1, 1), ("f",))
        self.failUnlessEqual(l.getMessage(-1, 2), ("d",))
        self.failUnlessEqual(l.getMessage(0, None), ("a",))
        self.failUnlessEqual(l.getMessage(0, 0), ("a",))
        self.failUnlessEqual(l.getMessage(0, 1), ("c",))
        self.failUnlessEqual(l.getMessage(0, 2), ("d",))

        l.log("g", pclass=0) # 0b 1c 2d 0e 1f 0g
        l.log("h", pclass=2) # 0b 1c 2d 0e 1f 0g 2h
        l.log("i", pclass=1) # 0b 1c 2d 0e 1f 0g 2h 1i
        l.log("j", pclass=1) # 0b 2d 0e 1f 0g 2h 1i 1j
        l.log("k", pclass=0) # 2d 0e 1f 0g 2h 1i 1j 0k
        self.failUnlessEqual(len(l.getItems(None)), 8)
        self.failUnlessEqual(len(l.getItems(0)), 3)
        self.failUnlessEqual(len(l.getItems(1)), 3)
        self.failUnlessEqual(len(l.getItems(2)), 2)
        self.failUnlessEqual(l.getMessage(-1, None), ("k",))
        self.failUnlessEqual(l.getMessage(-1, 0), ("k",))
        self.failUnlessEqual(l.getMessage(-1, 1), ("j",))
        self.failUnlessEqual(l.getMessage(-1, 2), ("h",))
        self.failUnlessEqual(l.getMessage(0, None), ("d",))
        self.failUnlessEqual(l.getMessage(0, 0), ("e",))
        self.failUnlessEqual(l.getMessage(0, 1), ("f",))
        self.failUnlessEqual(l.getMessage(0, 2), ("d",))

    def testTree(self):
        l = log.Log("foo")
        # a
        #  b
        #   c
        #   d
        #  e
        #   f
        # g

        a = l.log("a")
        b = a.log("b")
        b.log("c")
        b.log("d")
        e = a.log("e")
        e.log("f")
        l.log("g")

        items = l.getItems()
        self.failUnlessEqual(len(items), 2)
        A = items[0]
        self.failUnlessEqual(A.getMessage(), ("a",))
        Akids = A.getChildren()
        self.failUnlessEqual(len(Akids), 2)
        B,E = Akids
        self.failUnlessEqual(B.getMessage(), ("b",))
        Bkids = B.getChildren()
        self.failUnlessEqual(len(Bkids), 2)
        self.failUnlessEqual(Bkids[1].getMessage(), ("d",))
        Dkids = Bkids[1].getChildren()
        self.failIf(Dkids) # should be None

    def testObservers(self):
        l = log.Log("foo")
        events = []
        def observe(lognode, pclass, parent, events=events):
            events.append(lognode.getMessage())
        l.addObserver(observe)

        self.failIf(events)
        l.log("a")
        self.failUnlessEqual(events, [("a",)])
        l.log("b")
        self.failUnlessEqual(events, [("a",), ("b",)])
