
from twisted.trial import unittest
from twisted.internet import reactor
dr = unittest.deferredResult

import os
from petmail import mm_util
from petmail.transport import mail


class TestMixminion(unittest.TestCase):
    mm_path = None # set by the module-level setup() function

    def getMM(self):
        mm_path = os.environ.get("PETMAIL_TEST_MIXMINION", "mixminion")
        d = mm_util.getVersion(mm_path)
        version, error = dr(d)
        if not version:
            msg = "couldn't locate mixminion. Install it on your PATH " + \
                  "or define PETMAIL_TEST_MIXMINION=/path/to/mixminion"
            raise unittest.SkipTest(msg)
        return mm_path

    def setUp(self):
        mm_path = self.getMM()
        self.mmc = mm_util.MixminionContext("test_mm")
        self.mmc.set_mm_path(mm_path)
        self.mmc.passphrase = "foo"
        
    def testMint(self):
        surb1 = dr(self.mmc.mint("dest@example.com", "id1"))
        self.failUnless(surb1)
        self.failIf(surb1.granted)
        self.failIf(surb1.redeemed)
        self.failIf(surb1.expired)
        self.failUnlessEqual(surb1.identity, "id1")

        text = surb1.get()
        self.failUnless(surb1.granted)
        self.failUnless("BEGIN TYPE III REPLY BLOCK" in text)

        surb2 = dr(self.mmc.mint("dest@example.com", "id1"))
        self.failUnlessEqual(surb1.identity, surb2.identity)
        #self.failIfEqual(surb1.surbid, surb2.surbid) # need MM API first

    def testExpire(self):
        surb1 = dr(self.mmc.mint("dest@example.com", "id1"))
        results = []
        def expired(s, results=results):
            results.append(s)
        surb1.notifyOnExpiration(expired)
        self.failIf(results)
        self.mmc.expire(surb1)
        self.failUnless(surb1.expired)
        self.failUnlessEqual(results, [surb1])

    def testSend(self):
        queuedir = os.path.join(self.mmc.basedir, "queue")
        oldfiles = []
        if os.path.isdir(queuedir):
            oldfiles = os.listdir(queuedir)
        msg = "test message\n"
        res = dr(self.mmc.sendToDest(msg, "dest@example.com",
                                     "PETMAIL MESSAGE",
                                     queue=True))
        newfiles = os.listdir(queuedir)
        for f in oldfiles:
            newfiles.remove(f)
        self.failUnless(len(newfiles) == 2)
        for f in newfiles:
            os.unlink(os.path.join(queuedir, f))

    def testSendToSURB(self):
        queuedir = os.path.join(self.mmc.basedir, "queue")
        oldfiles = []
        if os.path.isdir(queuedir):
            oldfiles = os.listdir(queuedir)
        surb1 = dr(self.mmc.mint("dest@example.com", "id1"))
        msg = "test message\n"
        res = dr(self.mmc.sendToSURBs(msg, surb1.get(),
                                      "PETMAIL MESSAGE",
                                      queue=True))
        newfiles = os.listdir(queuedir)
        for f in oldfiles:
            newfiles.remove(f)
        self.failUnless(len(newfiles) == 2)
        for f in newfiles:
            os.unlink(os.path.join(queuedir, f))
        

    def receiveMessage(self, msg, source):
        self.messages.append((msg, source))

    def testDecodePlaintext(self):
        # this one actually spawns mixminion to unwrap 'plaintext' messages

        message = "message\x00\x01\x02" * 40
        encoded = mail.encapsulateInText(message)
        
        mmencoded = "\n" + encoded
        mmencoded = mmencoded.replace("\n--", "\n- --")
        mmencoded = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----\n" + \
                    "Message-type: plaintext\n" + \
                    mmencoded + \
                    "-----END TYPE III ANONYMOUS MESSAGE-----\n"

        i = mail.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return True
        i.canDecryptType3 = canDecryptType3
        def decryptType3(encoded, self=self):
            return self.mmc.decrypt(encoded)
        i.decryptType3 = decryptType3

        self.messages = []
        dr(i.gotText(mmencoded))
        self.failUnless(len(self.messages), 1)
        out, (inlet, idstring, surbid) = self.messages[0]

        self.failUnlessEqual(out, message)
        self.failUnlessEqual(inlet, i)
        self.failIf(idstring)
        self.failIf(surbid)


    def testDecodeEncrypted(self):
        # this test spawns mixminion to unwrap 'encrypted' messages

        # need multiple things to do this:
        #  must create a valid encrypted message (make SURB, send-to-SURB)
        #  must get the encrypted message (i.e. before it gets sent out)
        #  must hand message to the same MM environment which made the SURB
        
        message = "message\x00\x01\x02" * 40
        encoded = mail.encapsulateInText(message)

        # populate the secrets
        keyring_txt = \
"""-----BEGIN TYPE III KEYRING-----
Version: 0.1

S0VZUklORzIAzGzxQP2HLczwZQ/7FHWa7ISLqHOw4gUON3Wn1zYgFjw8uMr/42dP
KuNw6ABkSu/PIexxK6Lqle57przeiBlrMgM92oVxWzCdI0NjEou+29/AZZQFqYXn
1ml4WDj14RB7czhyfn4Ggm2PSRnGN2CNUW/tXebpwUVylkJ1da0yhROpj6Q/opfu
+Dr/XwSjtTkz0SKdbNJN7/TZHNULjplXIGkbFVFQOs4hbVuVwWhu0ag4j9KZ5l/A
aEQnGjRy3I2qQOhvbcupqq056a8MDv4DaFoC0dtwxUDgqw90SgmwtUfm7G46hZsk
SspK1mZoXgkvLph0KD/Ee+BVEmNnIieyWVEIm4rjjQBs/6Ti+J4z8bcLaaK/OOD/
lQ7b/FlR5ok97P4gHfCj0a+gGC/23GpcFFb8+V3YDNRYpWm58nnt11Q1OQIxULyL
+9jDJGwSbhOEZdfSh6tuHrh9CwNKMHpyEMm08QpqL0to9utXb2jfK8rNORQ7zJHw
BgBNi3NkFduP+XpzVHm2psPaiMpN9HPUyok7r2Sod9qhTlhaA2iAmHzD1gM2jY4e
bFZVzU+px2SqICWYyZKB4BhiK9qZtUdh4jlbg7zyJGY6tCATkhsrXME4JHKb9Wb3
Prba28WzLno5q689k88mED9cvOzNUYxuIaeOqiCDfVT8ZHz/CDqSCVASzvgaegON
Jl0vRboGPGVQgHPP1xE3dBmY7NXw5YbGrgcpY/5tM56eVgsDHRbrWSWsk7TIiw6X
5T3LkdcaO/yBI7hVPqsjDcP67Ej6Rppb3bUeC5cmIapu4srXYP+fVTBhxm0MHRNT
H3KzwRJXnDLioSu0EDFi8Im99COUXNUiiQbjdfmVlhosgHNaVFQr6Kg9Cp2hH4Kf
iu3Xk/H9QBnn7l1TAX1/GcHYAcDz1pJVYOuCVOTq8j4V8bPRgPw+5K04JDOsZeh5
V0P9uYJvhfFEhyXSd8b3T4vAbKkVElPNO9dmFSbV/AKY0u9dykkQGmuxkRDJI1gC
yHY9mqEBFnEhQcr1s1RBjDcVpoIy/XbF6DjOdOJSGfZzoTUJsngphBgBv657IZLP
dVaKMH7Aqw0rpqOL5o8H2n4aJQDDzdtOKIKh+Nm4P1Rh060XSGOMbCgxcp3CKPiO
o2zKj7aLB0yPX+s/YuoeqUVsw9m24QVQRpiiSL2CWUXRgqJHY2if2/BmM5hsuLAU
tk8ss4qkc6rELtaR+lOPPJXvPXioIjwmZEr5APIIDyLjeG7FhD7SyYPEhu41/aHi
peEuZZl4gtbTyhtK7ZL7C/+jMGiHv2k/0mSH3wxjbLi/YRZbzWmcwwy295YJulDc
CM2Ut7TOPunbPwD/viWTMNuNAGbLrCLMyz/vTQBU9+lFliMnXO2iQsqaRYSGQ0fZ
eYE1bIYahGv9
-----END TYPE III KEYRING-----
"""
        os.mkdir(os.path.join(self.mmc.basedir, "keys"), 0700)
        k = open(os.path.join(self.mmc.basedir, "keys", "keyring.txt"), "w")
        k.write(keyring_txt)
        k.close()
        # this should give us the same environment that created the SURB

        # A message was sent through the SURB with the following command:
        # mixminion send -R surb.txt -i msg.txt --subject "PETMAIL MESSAGE"

        # the following encrypted message is what came back from the
        # remailer network about an hour later

        encrypted = \
"""This is a Type III anonymous message, sent to you by the Winston Smith
Project mixminion server at firenze.linux.it. If you do not want to
receive anonymous messages, please contact antani-
admin@firenze.linux.it. For more information about anonymity, see
https://remailer.firenze.linux.it or
https://e-privacy.firenze.linux.it.

This message is not in plaintext.  It's either 1) a reply; 2) a forward
message encrypted to you; or 3) junk.

-----BEGIN TYPE III ANONYMOUS MESSAGE-----
Message-type: encrypted
Decoding-handle: bZXd/CtutLGyXfJpVy5ZHJQSD9o=
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-----END TYPE III ANONYMOUS MESSAGE-----
"""
        # it decrypts (by hand at least) to a --BEGIN PETMAIL MSG-- block,
        # preceded by a "SUBJECT:PETMAIL MESSAGE" header. The original
        # message has a subject line added by the exit remailer, in this
        # case "Type III Anonymous Message from Antani anonymous remailer",
        # so using --subject in the mixminion invocation is not useful, and
        # messages that arrive to our SURBs will not have distinctive
        # subject lines.

        i = mail.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return True
        i.canDecryptType3 = canDecryptType3
        def decryptType3(encoded, self=self):
            return self.mmc.decrypt(encoded)
        i.decryptType3 = decryptType3

        self.messages = []
        dr(i.gotText(encrypted))
        self.failUnless(len(self.messages), 1)
        out, (inlet, idstring, surbid) = self.messages[0]

        self.failUnlessEqual(out, message)
        self.failUnlessEqual(inlet, i)
        self.failUnlessEqual(idstring, "foo")
        self.failIf(surbid) # TODO: assert this gives something useful
        # but surbids must wait until mixminion supports them
    testDecodeEncrypted.todo = "test fails after a month, keys expired"

    def testCountSURBs(self):
        raise unittest.SkipTest("mixminion does not yet support count-surbs")
        msg = "test message\n"
        count = dr(self.mmc.countSURBs(msg))
        self.failUnlessEqual(count, 1)
        # TODO: try a larger message, probably some randomly-generated
        # (incompressible) data


def MM_usable(mm_path):
    d = mm_util.getVersion(mm_path)
    version, error = dr(d, 5)
    if not version:
        return "couldn't locate mixminion. Install it on your PATH " + \
               "or define PETMAIL_TEST_MIXMINION=/path/to/mixminion"
    mmc = mm_util.MixminionContext("test_mm")
    mmc.set_mm_path(mm_path)
    r = dr(mmc.checkDirectoryDownload())
    if not r:
        return "mixminion cannot contact its directory servers"
    return None

def setup():
    mm_path = os.environ.get("PETMAIL_TEST_MIXMINION", "mixminion")
    why = MM_usable(mm_path)
    if why:
        TestMixminion.skip = why
    else:
        TestMixminion.mm_path = mm_path

setup()
