#! /usr/bin/python

from twisted.trial import unittest
dr = unittest.deferredResult
from twisted.internet import defer
from twisted.python import failure, log
import sys

#log.startLogging(sys.stderr)

from petmail import agent, owner, addressbook, interfaces, plugin
from petmail.message import OutboundMessage, ReceivedMessage
from petmail.idrecord import IDRecord

class MyError(Exception):
    pass

class Plugin(plugin.Plugin):
    __implements__ = (interfaces.IEntryPlugin, interfaces.IOwnerPlugin,
                      interfaces.IAgentPlugin, plugin.Plugin.__implements__)
    def __init__(self, name, test):
        self.name = name
        self.test = test
    def outboundMessage(self, msg):
        self.test.log.append((self.name, "outboundMessage", msg))
    def inboundMessage(self, msg):
        self.test.log.append((self.name, "inboundMessage", msg))
    def entryChanged(self, action, entry, lognode):
        self.test.log.append((self.name, "entryChanged", action, entry))
    def ownerChanged(self, action, owner, lognode):
        self.test.log.append((self.name, "ownerChanged", action, owner))

class SlowPlugin(Plugin):
    def outboundMessage(self, msg):
        self.d = defer.Deferred()
        return self.d
    def fire(self, msg):
        Plugin.outboundMessage(self, msg)
        self.d.callback(msg)

class FailingPlugin(SlowPlugin):
    def fail(self, msg):
        self.d.errback(failure.Failure(MyError()))

class TestDispatch(unittest.TestCase):
    def setUp(self):
        self.log = []

    def addP(self, parent, name, pluginclass=Plugin):
        p = pluginclass(name, self)
        parent.addPlugin(p)
        return p

    def check(self, events):
        self.failUnlessEqual(self.log, events)
        self.log = []


    def testOwner(self):
        o = owner.OwnerEntry()
        o.deleteAllPlugins()

        self.addP(o, "a")
        self.check([])

        m1 = OutboundMessage(None)
        dr(o.runPlugins("outboundMessage", m1))
        self.check([("a", "outboundMessage",m1)])

        m2 = OutboundMessage(None)
        dr(o.runPlugins("inboundMessage", m2))
        self.check([("a", "inboundMessage", m2)])

    def testEntry(self):
        o = addressbook.AddressBookEntry()
        o.deleteAllPlugins()

        self.addP(o, "a")

        self.check([])

        m1 = OutboundMessage(None)
        dr(o.runPlugins("outboundMessage", m1))
        self.check([("a", "outboundMessage", m1)])

        m2 = OutboundMessage(None)
        dr(o.runPlugins("inboundMessage", m2))
        self.check([("a", "inboundMessage", m2)])

    def testBoth(self):
        # o is the sender, and has a single plugin "o"
        # it has an addressbook
        o = owner.OwnerEntry()
        o.deleteAllPlugins()

        self.addP(o, "o")

        book = addressbook.AddressBook()
        book.setServiceParent(o)

        # entry1 is the recipient, and has two plugins: "e1" and "e2" (slow)
        entry1 = addressbook.AddressBookEntry()
        entry1.idRecord = IDRecord()
        entry1.deleteAllPlugins()
        entry1.setServiceParent(book)
        self.addP(entry1, "e1")
        p2 = self.addP(entry1, "e2", SlowPlugin)

        # entry2 is some other recipient, with a plugin "e3"
        entry2 = addressbook.AddressBookEntry()
        entry2.deleteAllPlugins()
        entry2.setServiceParent(book)
        self.addP(entry2, "e3")

        m = OutboundMessage(None)
        m.setSenderEntry(o)
        m.setRecipientEntry(entry1)
        d = m.handlePlugins(None)
        self.check([("e1", "outboundMessage", m)])
        # e2 hasn't fired yet, and OwnerPlugins run after all EntryPlugins

        p2.fire(m)
        dr(d)
        self.check([("e2", "outboundMessage", m),
                    ("o", "outboundMessage", m)])

    def testFailing(self):
        # o is the sender, and has a single plugin "o"
        # it has an addressbook
        o = owner.OwnerEntry()
        o.deleteAllPlugins()

        self.addP(o, "o")
        book = addressbook.AddressBook()
        book.setServiceParent(o)

        # entry1 is the recipient, and has two plugins: "e1" and "e2" (slow)
        entry1 = addressbook.AddressBookEntry()
        entry1.idRecord = IDRecord()
        entry1.deleteAllPlugins()
        entry1.setServiceParent(book)
        self.addP(entry1, "e1")
        p2 = self.addP(entry1, "e2", FailingPlugin)

        m = OutboundMessage(None)
        m.setSenderEntry(o)
        m.setRecipientEntry(entry1)
        d = m.handlePlugins(None)
        self.check([("e1", "outboundMessage", m)])

        p2.fail(m)
        f = unittest.deferredError(d)
        self.failUnless(f.check(MyError))
        # o2 is not run, because an EntryPlugin failed
        self.check([])

    def test1(self):
        # build a tree:
        #  Agent1: a1
        #   Owner1: o1, o2
        #    Entry1: e1, e2
        #    Entry2: e3
        #   Owner2:
        #    Entry3: e4

        agent1 = agent.Agent("basedir")
        agent1.deleteAllPlugins()
        self.addP(agent1, "a1")
        self.check([])

        owner1 = owner.OwnerEntry()
        owner1.deleteAllPlugins()
        dr(agent1.addOwner(owner1))
        self.check([("a1", "ownerChanged", "add", owner1)])

        # add owner plugins
        self.addP(owner1, "o1")
        self.check([])
        o2 = self.addP(owner1, "o2")
        self.check([])

        book1 = agent1.getAddressBook(owner1.ID)
        entry1 = addressbook.AddressBookEntry()
        entry1.idRecord = IDRecord()
        entry1.deleteAllPlugins()
        dr(book1.addEntry(entry1))
        # owner plugins sees it first, then the Agent
        self.check([("o1", "entryChanged", "add", entry1),
                    ("o2", "entryChanged", "add", entry1),
                    ("a1", "entryChanged", "add", entry1),
                    ])

        # add entry plugins
        self.addP(entry1, "e1")
        self.addP(entry1, "e2")
        self.check([])

        entry2 = addressbook.AddressBookEntry()
        entry2.deleteAllPlugins()
        dr(book1.addEntry(entry2))
        self.check([("o1", "entryChanged", "add", entry2),
                    ("o2", "entryChanged", "add", entry2),
                    ("a1", "entryChanged", "add", entry2),
                    ])

        self.addP(entry2, "e3")
        self.check([])

        # add owner2
        owner2 = owner.OwnerEntry()
        owner2.deleteAllPlugins()
        dr(agent1.addOwner(owner2))
        self.check([("a1", "ownerChanged", "add", owner2)])

        book2 = agent1.getAddressBook(owner2.ID)
        entry3 = addressbook.AddressBookEntry()
        entry3.deleteAllPlugins()
        dr(book2.addEntry(entry3))
        self.check([("a1", "entryChanged", "add", entry3),
                    ])

        self.addP(entry3, "e4")
        self.check([])


        # update entry1
        dr(entry1.modified(None))
        self.check([("e1", "entryChanged", "change", entry1),
                    ("e2", "entryChanged", "change", entry1),
                    ("o1", "entryChanged", "change", entry1),
                    ("o2", "entryChanged", "change", entry1),
                    ("a1", "entryChanged", "change", entry1),
                    ])

        # update entry3
        dr(entry3.modified(None))
        self.check([("e4", "entryChanged", "change", entry3),
                    ("a1", "entryChanged", "change", entry3),
                    ])

        owner1.removePlugin(o2)
        # update entry1
        dr(entry1.modified(None))
        self.check([("e1", "entryChanged", "change", entry1),
                    ("e2", "entryChanged", "change", entry1),
                    ("o1", "entryChanged", "change", entry1),
                    ("a1", "entryChanged", "change", entry1),
                    ])
        
        # check outboundMessage
        m = OutboundMessage(agent1)
        m.setSenderEntry(owner1)
        m.setRecipientEntry(entry1)
        dr(m.handlePlugins(None))
        # note: e1/e2 could be called in either order. It happens that they
        # are started in order, but in theory they are run in parallel.
        self.check([("e1", "outboundMessage", m),
                    ("e2", "outboundMessage", m),
                    ("o1", "outboundMessage", m),
                    ("a1", "outboundMessage", m),
                    ])

        # and inboundMessage
        m = ReceivedMessage(agent1, "encrypted")
        m.senderEntry = entry2
        m.recipientEntry = owner1
        dr(m.handlePlugins(None))
        self.check([("e3", "inboundMessage", m),
                    ("o1", "inboundMessage", m),
                    ("a1", "inboundMessage", m),
                    ])

