
from twisted.trial import unittest
from twisted.internet import defer
from twisted.python.components import implements
def dr(d, timeout=10):
    if isinstance(d, defer.Deferred):
        return unittest.deferredResult(d, timeout)
    return d
def de(d, timeout=10):
    return unittest.deferredError(d, timeout)

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import agent, processors, message, addressbook, permission
from petmail import ticket, crypto, interfaces, util, owner
from petmail.server import ticketserver
from petmail.idrecord import IDRecord
Unsendable = interfaces.Unsendable
NotPermitted = interfaces.NotPermitted

class FakeAddressBook(addressbook.AddressBook):
    def __init__(self):
        addressbook.AddressBook.__init__(self)

    def addEntry(self, entry, message):
        entry.addressbook = self
        self.entries.append((entry, message))
    def lookupID(self, ID):
        return None

class FakeAgent(agent.Agent):
    debugProcessing = False

    def __init__(self):
        agent.Agent.__init__(self, "basedir")
        self.ticketAgents = [FakeTicketAgent()]
        self.me = FakeOwnerEntry(1, None)

    def log(self, *args):
        pass

    def acceptRFC822(self, message, why):
        pass

    def rejectRFC822(self, message, why):
        assert isinstance(why, Exception) # or maybe a Failure?

    def getTicketAgents(self):
        return self.ticketAgents
    def getAddressBook(self, senderID):
        return self.addressbook
    def getSender(self, senderID):
        return self.me

class FakeTicketAgent:
    def __init__(self):
        self.requests = []
    def handleRequirements(self, requirements, msg):
        dl = []
        for r in requirements:
            if r[0] == "ticket":
                d = defer.Deferred()
                self.requests.append(d)
                dl.append(d)
        return dl
    def fire(self, ticket):
        r = self.requests.pop(0)
        r.callback(ticket)
    def error(self, why):
        r = self.requests.pop(0)
        r.errback(why)

class FakeProcessor(processors.OutboundProcessor):
    dependencies = []
    provides = []
    def __init__(self, name, test):
        processors.OutboundProcessor.__init__(self)
        self.name = name
        self.test = test
    def checkMessage(self, message, lognode=None):
        self.test.calls.append(self.name)
        for p in self.provides:
            message.provide(p)
    def __repr__(self):
        return "<FakeProcessor %s>" % self.name

class SlowProcessor(FakeProcessor):
    def checkMessage(self, message, lognode=None):
        self.test.called.append(self.name)
        self.d = defer.Deferred()
        return self.d
    def callback(self, res, msg):
        self.test.completed.append(self.name)
        for p in self.provides:
            msg.provide(p)
        self.d.callback(res)
    def errback(self, why, msg):
        self.test.completed.append(self.name)
        for p in self.provides:
            msg.provide(p)
        self.d.errback(why)

class FakeLookupAgent(object):
    def __init__(self, test):
        self.test = test
    def lookup(self, agent, address):
        self.test.lookups.append((self, agent, address))
        self.d = defer.Deferred()
        return self.d
    def callback(self, record):
        self.d.callback(record)
    def errback(self, why):
        self.d.errback(why)

class FailingLookupAgent(FakeLookupAgent):
    def lookup(self, agent, address):
        raise KeyError

class FakeID(object):
    __implements__ = interfaces.IIDRecord
    def __init__(self, ID, test):
        self.ID = ID
        self.test = test
        self.transport = None
        self.error = False
    def getTransport(self, sender):
        self.test.gettransports.append(sender)
        if self.error:
            raise Unsendable
        if self.transport:
            return defer.succeed(self.transport)
        self.d = defer.Deferred()
        return self.d
    def callback(self, transport):
        self.d.callback(transport)
    def errback(self, why):
        self.d.errback(why)

class FakeOwnerEntry(owner.OwnerEntry):
    def __init__(self, ID, test):
        owner.OwnerEntry.__init__(self)
        self.ID = ID
        self.test = test

class FakeMessage(message.OutboundMessage):
    def logOutboundMessage(self):
        pass

class AgentTest(unittest.TestCase):
    def setup1(self):
        self.agent = agent.Agent("testagent")
        p1 = FakeProcessor("p1", self)
        p1.dependencies = []; p1.provides = ["1"]
        p2 = FakeProcessor("p2", self)
        p2.dependencies = ["4"]; p2.provides = ["2"]
        p3 = FakeProcessor("p3", self)
        p3.dependencies = ["1", "2"]; p3.provides = ["3"]
        p4 = FakeProcessor("p4", self)
        p4.dependencies = []; p4.provides = ["4"]
        self.agent.outboundProcessors = [p1, p2, p3, p4]

    def testNoProcessors1(self):
        self.setup1()
        self.agent.outboundProcessors = []
        m = FakeMessage(self.agent)
        m.processors = []
        self.failUnlessEqual(m.findRunnableProcesses(), [])

    def testNoProcessors2(self):
        self.setup1()
        self.agent.outboundProcessors = []
        m = FakeMessage(self.agent)
        d = m.startProcessing()
        dr = unittest.deferredResult(d)
        self.failUnlessEqual(dr, m)


    def runnable(self, reachable, p):
        for f in p.dependencies:
            if not f in reachable:
                return False
        return True

    def checkDependencies(self, processors, axioms):
        allDeps = []
        for p in processors:
            for f in p.dependencies:
                if f not in allDeps:
                    allDeps.append(f)
        reachable = []
        remaining = allDeps[:]
        for f in axioms:
            reachable.append(f)
            remaining.remove(f)
        while remaining:
            #print "processors", processors
            progress = False
            for p in processors[:]:
                if self.runnable(reachable, p):
                    #print " runnable", p
                    progress = True
                    for f in p.provides:
                        if f not in reachable:
                            reachable.append(f)
                        if f in remaining:
                            remaining.remove(f)
                    processors.remove(p)
            if not progress:
                why = "unrunnable Processors: %s\n" % \
                      [p.__class__.__name__ for p in processors]
                why += "unfulfilled dependencies: %s\n" % remaining
                self.fail(why)

    def testInboundDependencies(self):
        a = agent.Agent("testagent")
        a.populateProcessors()
        processors = a.getInboundProcessors()
        axioms = []
        # interestingly, we must either start with acceptedRFC822 or
        # senderEntry (CreateSenderEntry will provide senderEntry if we
        # accepted an rfc822 message). If neither is possible, most of the
        # inbound processors are unrunnable.
        self.checkDependencies(processors, axioms)

    def testOutboundDependencies(self):
        a = agent.Agent("testagent")
        a.populateProcessors()
        processors = a.getOutboundProcessors()
        axioms = ['recipient']
        self.checkDependencies(processors, axioms)


    def testAgent1(self):
        self.setup1()
        m = FakeMessage(self.agent)
        self.calls = []
        #self.agent.debugProcessing = True

        # internally the loop should do p1+p4, then p2, then p3
        # but it is no longer easy to test that

        d = m.startProcessing()
        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)
        self.failUnlessEqual(self.calls, ["p1", "p4", "p2", "p3"])

    def testAgent2(self):
        self.setup1()
        m = FakeMessage(self.agent)
        self.calls = []

        # internally the loop should do p4, then p2, then p3
        # but it is no longer easy to test that
        m.provides = ["1"]
        d = m.startProcessing() # should skip p1, do p4, then p2, then p3
        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)
        self.failUnlessEqual(self.calls, ["p4", "p2", "p3"])

    def setup2(self):
        self.agent = agent.Agent("testagent")
        p1 = SlowProcessor("p1", self)
        p1.dependencies = []; p1.provides = ["1"]
        p2 = SlowProcessor("p2", self)
        p2.dependencies = ["4"]; p2.provides = ["2"]
        p3 = SlowProcessor("p3", self)
        p3.dependencies = ["1", "2"]; p3.provides = ["3"]
        p4 = SlowProcessor("p4", self)
        p4.dependencies = []; p4.provides = ["4"]
        self.agent.outboundProcessors = [p1, p2, p3, p4]
        return p1, p2, p3, p4

    def testAgent3(self):
        p1, p2, p3, p4 = self.setup2()
        m = FakeMessage(self.agent)
        self.called = []; self.completed = []

        d = m.startProcessing()

        # on the first pass, p1 and p4 should start
        self.failUnlessEqual(self.called, ["p1", "p4"])
        self.failUnlessEqual(self.completed, [])

        p1.callback(None, m)
        self.failUnlessEqual(self.called, ["p1", "p4"])
        self.failUnlessEqual(self.completed, ["p1"])

        p4.callback(None, m)
        # calling p4 should trigger p2
        self.failUnlessEqual(self.called, ["p1", "p4", "p2"])
        self.failUnlessEqual(self.completed, ["p1", "p4"])

        p2.callback(None, m)
        # calling p2 should trigger p3
        self.failUnlessEqual(self.called, ["p1", "p4", "p2", "p3"])
        self.failUnlessEqual(self.completed, ["p1", "p4", "p2"])

        p3.callback(None, m)
        # calling p3 should complete the message
        self.failUnlessEqual(self.called, ["p1", "p4", "p2", "p3"])
        self.failUnlessEqual(self.completed, ["p1", "p4", "p2", "p3"])

        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)

    def testAgentGetsAllProcessors(self):
        a = agent.Agent("testagent")
        a.populateProcessors()
        if a.inboundProcessors:
            self.failUnlessEqual(len(a.inboundProcessors),
                                 len(processors.allInbound))
            self.failUnless(isinstance(a.inboundProcessors[0],
                                       processors.allInbound[0]))
        if a.outboundProcessors:
            self.failUnlessEqual(len(a.outboundProcessors),
                                 len(processors.allOutbound))
            self.failUnless(isinstance(a.outboundProcessors[0],
                                       processors.allOutbound[0]))
        

class Outbound(unittest.TestCase):
    def setUp(self):
        self.agent = FakeAgent()
        self.agent.crypto = crypto.EncryptionAgent()
        self.agent.addressbook = FakeAddressBook()
        self.agent.addressbook.setServiceParent(self.agent.me)


    def testRFC822_1(self):
        raise unittest.SkipTest("was moved to test_agent")
        # simple test of a no-permission case
        m = message.OutboundMessage(self.agent)
        m.addRFC822("This is the message body.")
        rr = m.recipientRecord = IDRecord()
        # rr.genericPermission is None: no new senders are accepted
        m.provide('recipient')
        # recipientEntry is None: no relationship, so no specific permission

        p = processors.RFC822Permission()
        #p.checkMessage(m)
        self.failUnlessRaises(Unsendable, p.checkMessage, m)

    def testRFC822_2(self):
        # single ticket retrieved successfully
        ta = self.agent.ticketAgents[0]
        m = message.OutboundMessage(self.agent)
        m.addRFC822("This is the message body.")
        rr = IDRecord()
        gp = rr.genericPermission = permission.PermissionSlip()
        gp.ticketURLs = ["http://ticketserver.com"]
        # new senders require a ticket
        m.recipientRecord = rr
        m.provide('recipient')
        # recipientEntry is None: no relationship, so no specific permission

        p = processors.RFC822Permission()
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(len(ta.requests), 1)
        self.failIf(m.getNode('rfc822').get("tickets"))

        ta.fire(123)
        self.failUnlessEqual(m.getNode('rfc822').get("tickets"), [123])
        self.failUnless(dr)

    def testRFC822_3(self):
        # ticket failed
        ta = self.agent.ticketAgents[0]
        m = message.OutboundMessage(self.agent)
        m.addRFC822("This is the message body.")
        rr = IDRecord()
        gp = rr.genericPermission = permission.PermissionSlip()
        gp.ticketURLs = ["http://ticketserver.com"]
        # new senders require a ticket
        m.recipientRecord = rr
        m.provide('recipient')
        # recipientEntry is None: no relationship, so no specific permission

        p = processors.RFC822Permission()
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(len(ta.requests), 1)
        self.failIf(m.getNode('rfc822').get("tickets"))

        ta.error(Unsendable("ticket request failed"))
        self.failIf(m.getNode('rfc822').get("tickets"))
        f = de(r)
        self.failUnless(f.check(type(())))

class AddressLookup(unittest.TestCase):
    def setUp(self):
        rx_agent = self.rx_agent = FakeAgent()
        rx_agent.crypto = crypto.EncryptionAgent()
        rx_agent.addressbook = FakeAddressBook()
        rx_agent.addressbook.setServiceParent(rx_agent.me)
        me = dr(rx_agent.createIdentity())
        me.setName("alice")
        rx_agent.addOwner(me)
        self.signed = dr(me.sign())

    def testAddressLookup1(self):
        # this tests Agent.retrieveAddress with both successful and failing
        # LookupAgents
        a = agent.Agent("homedir")
        la1 = FakeLookupAgent(self)
        la2 = FailingLookupAgent(self)
        a.addressLookupAgents = [la1, la2]

        self.lookups = []
        d = a.retrieveAddress("alice@example.com")
        self.failUnlessEqual(self.lookups, [(la1, a, "alice@example.com")])
        la1.callback(self.signed)
        r = dr(d)
        self.failUnlessEqual(r, self.signed)
        self.failUnlessEqual(self.lookups, [(la1, a, "alice@example.com")])

    def testAddressLookup2(self):
        a = agent.Agent("homedir")
        la1 = FailingLookupAgent(self)
        la2 = FakeLookupAgent(self)
        a.addressLookupAgents = [la1, la2]

        self.lookups = []
        d = a.retrieveAddress("alice@example.com")
        self.failUnlessEqual(self.lookups, [(la2, a, "alice@example.com")])
        la2.callback(self.signed)
        r = dr(d)
        self.failUnlessEqual(r, self.signed)

    def testAddressLookup3(self):
        a = agent.Agent("homedir")
        la1 = FakeLookupAgent(self)
        la2 = FakeLookupAgent(self)
        a.addressLookupAgents = [la1, la2]

        self.lookups = []
        d = a.retrieveAddress("alice@example.com")
        self.failUnlessEqual(self.lookups, [(la1, a, "alice@example.com")])
        la1.errback(KeyError())
        self.failUnlessEqual(self.lookups, [(la1, a, "alice@example.com"),
                                            (la2, a, "alice@example.com")])
        la2.callback(self.signed)
        r = dr(d)
        self.failUnlessEqual(r, self.signed)

    def testAddressLookup4(self):
        a = agent.Agent("homedir")
        la1 = FakeLookupAgent(self)
        la2 = FakeLookupAgent(self)
        a.addressLookupAgents = [la1, la2]

        self.lookups = []
        d = a.retrieveAddress("alice@example.com")
        la1.errback(KeyError())
        self.failUnlessEqual(self.lookups, [(la1, a, "alice@example.com"),
                                            (la2, a, "alice@example.com")])
        la2.errback(KeyError())
        e = de(d)
        self.failUnless(e.check(Unsendable))

class Inbound(unittest.TestCase):
    skip = "must rewrite in terms of plugins"
    def setUp(self):
        self.agent = FakeAgent()
        self.agent.crypto = self.crypto = crypto.EncryptionAgent()
        self.agent.me = owner.OwnerEntry()
        self.addressbook = addressbook.AddressBook()
        self.addressbook.setServiceParent(self.agent.me)

    def makeReceivedMessage(self):
        rm = message.ReceivedMessage(self.agent, None)
        rm.recipientEntry = self.agent.me
        rm.provide('senderEntry')
        rm.senderEntry = addressbook.AddressBookEntry()
        rm.senderEntry.idRecord = IDRecord()
        self.addressbook.addEntry(rm.senderEntry)
        rm.provide("updatedIDRecord")
        rm.headers = {'senderID': rm.senderEntry.idRecord.ID,
                      'messageID': "2",
                      'rfc822': {'body': "message body", 'tickets': []},
                      }
        return rm

    def testRFC822_1(self):
        # no permission
        rm = self.makeReceivedMessage()
        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822, False)

    def testRFC822_2(self):
        # generic permission accepts anybody
        self.agent.me.setGenericPermission(permission.OwnerPermission())
        rm = self.makeReceivedMessage()
        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822, "(generic) no ticket required")

    def testRFC822_3(self):
        # require ticket, but don't provide one
        gp = permission.OwnerPermission()
        self.agent.me.setGenericPermission(gp)
        url = "http://ticketserver.com"
        vendkey = dr(self.crypto.generateKey("ticketserver.com"), 60)
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        gp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        
        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822, False)

    def testRFC822_4(self):
        # require ticket
        gp = permission.OwnerPermission()
        self.agent.me.setGenericPermission(gp)
        url = "http://ticketserver.com"
        vendkey = dr(self.crypto.generateKey("ticketserver.com"), 60)
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        gp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        t1 = dr(ticketserver.Ticket(url, util.now(), "1").signed(vendkey), 60)
        rm.headers['rfc822']['tickets'].append(t1)
        
        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822,
                             "(generic) valid ticket from http://ticketserver.com")

    def testRFC822_5(self):
        # specific permission which allows everything

        rm = self.makeReceivedMessage()
        rm.senderEntry.doUpdate = False
        # otherwise this change triggers a message send
        rm.senderEntry.setTheirPermission(permission.OwnerPermission())

        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822, "(specific) no ticket required")

    def testRFC822_6(self):
        # specific permission requires ticket, but none provided
        sp = permission.OwnerPermission()
        url = "http://ticketserver.com"
        vendkey = dr(self.crypto.generateKey("ticketserver.com"), 60)
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        sp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        rm.senderEntry.theirPermission = sp
        
        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822, False)

    def testRFC822_7(self):
        # specific permission requires ticket
        sp = permission.OwnerPermission()
        url = "http://ticketserver.com"
        vendkey = dr(self.crypto.generateKey("ticketserver.com"), 60)
        tt = ticket.TicketTaker(url, self.agent, vendkey)
        sp.addTicketTaker(tt)
        
        rm = self.makeReceivedMessage()
        rm.senderEntry.theirPermission = sp
        t1 = dr(ticketserver.Ticket(url, util.now(), "1").signed(vendkey), 60)
        rm.headers['rfc822']['tickets'].append(t1)
        
        p = processors.CheckRFC822Permission()
        dr(p.checkMessage(rm))
        self.failUnless("checkedRFC822" in rm.provides)
        self.failUnlessEqual(rm.allowRFC822,
                             "(specific) valid ticket from http://ticketserver.com")

    # new tests needed:

    # testUpdateMyIDRecord
    # testUpdateOurPermission
    # testAcceptRFC822
    # testGrantThirdPartyPermission
