#! /usr/bin/python

from twisted.trial import unittest
from twisted.python import components, log
import sys

#log.startLogging(sys.stderr)

from petmail import agent, names, interfaces, crypto
crypto.EncryptionAgent = crypto.FakeEncryptionAgent

def dr(d, timeout=5):
    return unittest.deferredResult(d, timeout)

class Remote(unittest.TestCase):
    def buildAgent(self, basedir, nameserver):
        a = agent.Agent(basedir)
        a.crypto = crypto.EncryptionAgent()
        a.registerAddressLookupAgent(nameserver)
        a.populateProcessors()
        a.doNotScheduleEvents = True
        return a

    def setUp(self):
        self.agent = self.buildAgent("remote", names.FakeNameServer())

    def failUnlessImplements(self, obj, iface):
        self.failUnless(components.implements(obj, iface))

    def test1(self):
        a = interfaces.IAgent(self.agent)
        self.failUnlessEqual(a.getAllIdentities(), [])

        i1 = dr(a.newIdentity())
        self.failUnlessImplements(i1, interfaces.IIdentity)
        i1_id = i1.getProperty("id")
        self.failUnlessEqual(type(i1_id), str)
        i1.setProperty("myName", "alice")
        name = "Mrs. A"
        i1.setProperty("name", name)
        addresses = ["alice@example.com"]
        i1.setProperty("addresses", addresses)
        i1.setProperty("expirationTime", 600)
        dr(a.addIdentity(i1))

        i2 = a.getIdentity(i1_id)
        self.failUnlessEqual(i2.getProperty("myName"), "alice")
        self.failUnlessEqual(i2.getProperty("name"), name)
        self.failUnlessEqual(i2.getProperty("addresses"), addresses)
        self.failUnlessEqual(i2.getProperty("expirationTime"), 600)

        self.failUnlessEqual(a.getAllIdentities(), [i2])

        # Identities are created without inbound permission
        self.failUnlessRaises(IndexError, i2.getGenericPermission, "bogus")
        self.failUnless(i2.getGenericPermission("rfc822") is None)

        # create the Generic Permission
        p0 = a.newPermission("rfc822")
        self.failUnlessImplements(p0, interfaces.IInboundPermission_rfc822)
        p0.setProperty("maxsize", 10000)
        i2.setGenericPermission("rfc822", p0)

        # check Inlets
        in1 = a.newInlet("dummy", name="foo")
        i2.setInlets([in1])

        p = i2.getGenericPermission("rfc822")
        self.failUnlessImplements(p, interfaces.IInboundPermission_rfc822)
        self.failUnlessEqual(p.getProperty("maxsize"), 10000)
        self.failUnlessEqual(p.listTicketTakerURLs(), [])

        # TODO: maybe move to auto-publish-on-change, maybe not
        dr(i2.publish())

        # test address book

        b = i2.getAddressBook()
        self.failUnlessImplements(b, interfaces.IAddressBook)
        self.failUnlessEqual(b.getOwner(), i2)
        self.failUnlessEqual(b.countEntries(), 0)
        self.failUnlessEqual(b.lookupID("123"), None)
        self.failUnlessEqual(b.lookupName("foo"), None)
        self.failUnlessEqual(b.lookupAddress("foo@foo"), None)

        dr(b.addAddress("alice@example.com"))

        self.failUnlessEqual(b.countEntries(), 1)
        r = b.lookupID(i1_id)
        self.failUnlessImplements(r, interfaces.ICorrespondent)
        self.failUnlessEqual(r.getProperty("id"), i1_id)
        self.failUnlessEqual(r.getProperty("name"), name)
        self.failUnlessEqual(r.getProperty("addresses"), addresses)
        self.failUnlessEqual(r.getProperty("transports"), ["Dummy:foo"])

        self.failUnless(r.getTheirPermission("rfc822") is None)

        p2 = r.getOurPermission("rfc822")
        self.failUnlessImplements(p2, interfaces.IOutboundPermission_rfc822)
        self.failUnlessEqual(p2.getProperty("maxsize"), 10000)
        self.failUnlessEqual(r.getOurPermissionType("rfc822"), "generic")

    def buildIdentity(self, name, address):
        i1 = dr(self.agent.newIdentity())
        i1.setProperty("myName", name)
        i1.setProperty("name", name)
        i1.setProperty("addresses", [address])
        i1.setGenericPermission("rfc822", self.agent.newPermission("rfc822"))
        i1.setInlets([self.agent.newInlet("dummy", name=name)])
        dr(self.agent.addIdentity(i1))
        return i1

    def testMessage(self):
        a = interfaces.IAgent(self.agent)
        alice = self.buildIdentity("alice", "alice@example.com")
        bob = self.buildIdentity("bob", "bob@example.com")
        # test address book

        b = alice.getAddressBook()
        to_bob = dr(b.addAddress("bob@example.com"))

        inbox = []
        bob.subscribeMailbox(lambda t,m: inbox.append(m), "rfc822")
        m = self.agent.newMessage("rfc822")
        m.setSender(alice)
        m.addRecipient_correspondent(to_bob)
        body = "rfc822 body goes here"
        m.addRFC822(body)
        dr(m.send())

        self.failUnlessEqual(len(inbox), 1)
        m1 = inbox.pop()
        self.failUnlessImplements(m1, interfaces.IInboundMessage)
        self.failUnlessImplements(m1, interfaces.IInboundMessage_rfc822)
        self.failUnlessEqual(m1.getMessage(), body)
        self.failUnlessEqual(dr(m1.getFromName()), '"alice"')
        self.failUnlessEqual(m1.getRecipient(), bob)
        self.failUnlessEqual(m1.getSender(), None)
        self.failUnlessEqual(m1.getSenderID(), alice.getProperty("id"))
        self.failUnlessEqual(m1.getSenderIDRecord(),
                             dr(alice.getArmored()))
        self.failUnlessEqual(m1.canAddSender(), True)
        self.failUnlessEqual(m1.canAddReferences(), [])
        # TODO: canReply needs improvement: it ought to return True for
        # unknown senders as long as there is an idRecord in the message
        self.failUnlessEqual(m1.canReply(), False)

        # set a petname on the previously-added correspondent, and test lookup
        bobby = b.lookupID(bob.getProperty("id"))
        bobby.setProperty("petname", "bobby")

        self.failUnlessEqual(b.lookupID(bob.getProperty("id")), bobby)
        self.failUnlessEqual(b.lookupName("bobby"), bobby)
        self.failUnlessEqual(b.lookupAddress("bob@example.com"), bobby)

        # test sending by petname
        m = self.agent.newMessage("rfc822")
        m.setSender(alice)
        m.addRecipient_petname("bobby")
        body = "rfc822 body goes here"
        m.addRFC822(body)
        dr(m.send())
        self.failUnlessEqual(len(inbox), 1)
        inbox.pop()

        # now send by ID number
        m = self.agent.newMessage("rfc822")
        m.setSender(alice)
        m.addRecipient_id(bob.getProperty("id"))
        body = "rfc822 body goes here"
        m.addRFC822(body)
        dr(m.send())
        self.failUnlessEqual(len(inbox), 1)
        m2 = inbox.pop()

        # test the response side
        inbox = []
        alice.subscribeMailbox(lambda t,m: inbox.append(m), "rfc822")

        if 0: # TODO: turnenable this once .reply works correctly
            self.failUnless(m.canReply())
            m1 = m.reply()
            m1.addRFC822(body)
            dr(m.send())
            self.failUnlessEqual(len(inbox), 1)
            m1 = inbox.pop()
            # TODO: test m1

        self.failUnless(m2.canAddSender())
        ally = dr(m2.addSender())
        m1 = self.agent.newMessage("rfc822")
        m1.setSender(bob)
        m1.addRecipient_correspondent(ally)
        m1.addRFC822(body)
        dr(m1.send())

        self.failUnlessEqual(len(inbox), 1)
        m2 = inbox.pop()
        self.failUnlessEqual(m2.getMessage(), body)
        self.failUnlessEqual(m2.getSender(), bobby)
        bob_armored = m2.getSenderIDRecord()

        dr(b.deleteCorrespondent(bobby))
        self.failIf(b.lookupName("bobby"))

        bobby2 = dr(b.addArmoredIDRecord(bob_armored))
        self.failUnlessEqual(bobby2.getProperty("id"),
                             bobby.getProperty("id"))




# TODO: coverage for:
#  *.getPlugins
#  IAddressBook
#   .subscribe
#  ICorrespondent.getOurPermission
#  IOutboundMessage_rfc822:
#   addReference
#  IInboundMessage_rfc822: more
