
import sys

from twisted.trial import unittest
from twisted.python import log
from twisted.python.components import implements
from twisted.internet import defer
dr = unittest.deferredResult

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import addressbook, message, permission, util, ticket, crypto
from petmail import interfaces, agent, owner
from petmail.server import ticketserver
from petmail.idrecord import IDRecord

NotPermitted = interfaces.NotPermitted
Unsendable = interfaces.Unsendable

#log.startLogging(sys.stdout)

alicetext = \
"""From: alice@foo.bar
To: bob@boo.far
Subject: petmail test

Howdy bob. Aren't you glad this isn't spam?
 -Alice
"""

bobtext = \
"""From: bob@boo.far
To: alice@foo.bar
Subject: Re: petmail test

> Howdy bob. Aren't you glad this isn't spam?
>  -Alice

Yessiree bob. Er, alice. Good thing we switched to petmail!
 -Bob
"""

class FakeAgent(agent.Agent):
    validateSource = False
    doNotScheduleEvents = True

    def __init__(self, *args):
        agent.Agent.__init__(self, *args)
        self.inbox = []
        self.failures = []
        self.ticketAgents = []
    def watch(self, owner):
        owner.subscribeMailbox(self.acceptRFC822, "rfc822")
    def acceptRFC822(self, mtype, rm):
        assert mtype == "rfc822"
        assert isinstance(rm, message.ReceivedMessage)
        self.inbox.append((True, rm, rm.allowRFC822_reason))
    def rejectRFC822(self, rm, why):
        assert isinstance(rm, message.ReceivedMessage)
        self.inbox.append((False, rm, why))
    def _receiveMessageFailed(self, why, msg):
        self.failures.append(why)

    def getTicketAgents(self):
        return self.ticketAgents

# in the SendTests, alice sends to bob
# in the ReceiveTests, bob sends to alice

class SetupMixin:
    def makeCrypto(self):
        self.crypto = crypto.EncryptionAgent()

    def buildAlice(self, aliceAcceptsTickets, bobHasPermission):
        alice_agent = FakeAgent("alice_agent")
        alice_agent.populateProcessors()
        alice_agent.crypto = self.crypto

        us = dr(alice_agent.newIdentity())
        # TODO: we used to do self.crypto.generateKey("alice"), now
        # newIdentity just uses "name"
        us.doPublish = False
        us.grantReturnPermission = False
        us.grantInboundPermission = False
        dr(alice_agent.addIdentity(us))
        alice_agent.watch(us)

        if aliceAcceptsTickets:
            # we (alice) require tickets from strangers
            gp = permission.OwnerPermission()
            us.setGenericPermission("rfc822", gp)
            vendkey = dr(self.crypto.generateKey("ticketserver.com"), 60)
            tt = ticket.TicketTaker("http://ticketserver.com",
                                    alice_agent, vendkey)
            gp.addTicketTaker(tt)
            self.alice_ticketkey = vendkey

        bob = addressbook.AddressBookEntry()
        bob.setProperty("petname", "bob")
        bob.ID = "2"

        bob.idRecord = IDRecord()
        bob.idRecord.ID = bob.ID
        bob.idRecord.pubkey = "bobkey" # this should point at bob
        if bobHasPermission:
            # tell Alice to let bob's messages in
            p = permission.OwnerPermission()
            p.minIMG = 10
            bob.setTheirPermission("rfc822", p)

        book = alice_agent.addressbooks[us] # TODO: use getAddressBook
        book.addEntry(bob)

        self.alice_agent = alice_agent
        self.alice_owner = us
        self.alice_book = book

    def buildBob(self, bobThinksHeHasPermission,
                 bobThinksAliceAcceptsTickets):

        bob_agent = FakeAgent("bob_agent")
        bob_agent.populateProcessors()
        bob_agent.crypto = self.crypto

        us = dr(bob_agent.newIdentity())
        us.doPublish = False
        us.grantReturnPermission = False
        us.grantInboundPermission = False
        dr(bob_agent.addIdentity(us))
        bob_agent.watch(us)

        alice = addressbook.AddressBookEntry()
        alice.setProperty("petname", "alice")
        alice.ID = "1"
        if bobThinksHeHasPermission:
            p = permission.PermissionSlip()
            p.minIMG = 10
            alice._setOurPermission(p)
        alice.idRecord = IDRecord()
        alice.idRecord.ID = alice.ID
        alice.idRecord.pubkey = "alicekey"
        if bobThinksAliceAcceptsTickets:
            # alice says she requires tickets from strangers
            alice.idRecord.genericPermission = gp = permission.PermissionSlip()
            gp.ticketURLs.append("http://ticketserver.com")

        book = bob_agent.addressbooks[us]
        book.addEntry(alice)

        self.bob_agent = bob_agent
        self.bob_owner = us
        self.bob_book = book

    def build(self,
              bobThinksHeHasPermission = 0,
              bobThinksAliceAcceptsTickets = 0,
              aliceAcceptsTickets = 0,
              bobHasPermission = 0,
              ):
        self.makeCrypto()
        self.buildAlice(aliceAcceptsTickets, bobHasPermission)
        self.buildBob(bobThinksHeHasPermission, bobThinksAliceAcceptsTickets)

        bob = self.bob_owner
        b = self.alice_book.lookupPetName("bob")
        b.idRecord.pubkey = bob.privkey.pubkey()
        b.idRecord.ID = bob.ID
        b.ID = bob.ID

        alice = self.alice_owner
        a = self.bob_book.lookupPetName("alice")
        a.idRecord.pubkey = alice.privkey.pubkey()
        a.idRecord.ID = alice.ID
        a.ID = alice.ID

    def buildMessage(self, sender, recipient, text=alicetext):
        a = sender.parent
        m = message.OutboundMessage(a)
        m.setSenderEntry(sender)
        m.setRecipientEntry(recipient)
        m.addRFC822(text)
        return m


class FakeTicketAgent(object):
    asked = False
    def handleRequirements(self, requirements, msg):
        self.asked = True
        return [defer.succeed(self.ticket)]

class SendTest(unittest.TestCase, SetupMixin):
    # alice sends to bob, and we verify things about the created message

    def testHeader1(self):
        self.build()

        us = self.alice_owner

        bob = self.alice_book.lookupPetName("bob")
        assert(bob)
        m = self.buildMessage(us, bob)

        # bob does not yet accept any mail from strangers, so this will fail
        d = m.startProcessing()

        # make sure it is not sendable
        de = unittest.deferredError(d)
        self.failUnless(de.check(Unsendable))

        # make bob accept all mail from strangers
        bob.idRecord.genericPermission = permission.PermissionSlip()

        # send another message, make sure it is sendable
        m = self.buildMessage(us, bob)
        d = m.startProcessing()

        r = unittest.deferredResult(d)
        self.failUnlessEqual(r, m)
        self.failUnless(m.getNode('messageID'))
        
    def testHeader2(self):
        self.build()

        us = self.alice_owner
        
        bob = self.alice_book.lookupPetName("bob")
        assert(bob)
        m = self.buildMessage(us, bob)

        # make bob require tickets from strangers
        gp = bob.idRecord.genericPermission = permission.PermissionSlip()
        url = "http://ticketserver.com"
        gp.ticketURLs.append(url)

        # add the ticket
        ta = FakeTicketAgent()
        m.agent.ticketAgents = [ta]
        vendkey = dr(self.crypto.generateKey("vendor"), 60)
        t1 = dr(ticketserver.Ticket(url, util.now(), "1").signed(vendkey),
                60)
        ta.ticket = t1

        # this should ask a ticketAgent for a ticket
        d = m.startProcessing()
        r = unittest.deferredResult(d)
        self.failUnlessIdentical(r, m)
        self.failUnless(ta.asked)
        self.failUnless(m.getNode('rfc822')['tickets'])
        self.failUnlessEqual(m.getNode('rfc822')['tickets'][0], t1)

    def testSend(self):
        self.build()

        us = self.alice_owner

        bob = self.alice_book.lookupPetName("bob")
        assert(bob)
        m = self.buildMessage(us, bob)

        # make bob accept everything
        bob.idRecord.genericPermission = permission.PermissionSlip()

        d = m.startProcessing()
        r = unittest.deferredResult(d)

        encrypted = dr(m.encrypt(m.senderEntry.privkey,
                                 m.recipientRecord.pubkey), 60)
        self.failUnless(encrypted)
        #print "\n--- encrypted ---\n%s\n--- end ---" % encrypted
        

class ReceiveTest(unittest.TestCase, SetupMixin):
    # bob sends to alice, and we verify things about both sending and
    # receiving

    def doMessage(self, from_, to):
        return self.buildMessage(from_, to, bobtext)

    def testReceive1(self):
        # bob tries to send a message to alice, but his agent knows he
        # doesn't have permission so it doesn't let him.
        self.build(bobThinksHeHasPermission=0)

        alice = self.bob_book.lookupPetName("alice")

        # adding the body will fail, because we know we have no way to get it
        # accepted

        m = self.doMessage(self.bob_owner, alice)

        d = m.startProcessing()
        de = unittest.deferredError(d)
        self.failUnless(de.check(Unsendable))

    def makeAgent(self, book, us):
        aa = FakeAgent("testagent")
        aa.populateProcessors()
        aa.addressbooks[us] = book
        aa.crypto = self.crypto
        aa.setServiceParent(aa)
        return aa

    def send(self, m):
        d = m.startProcessing()
        self.failUnless(unittest.deferredResult(d))
        encrypted = dr(m.encrypt(m.senderEntry.privkey,
                                 m.recipientRecord.pubkey), 60)
        return encrypted

    def sendFails(self, m):
        d = m.startProcessing()
        de = unittest.deferredError(d)
        return de

    def received(self, aa):
        if aa.failures:
            for f in aa.failures:
                print f
            self.fail("failure occurred during processing")
        self.failUnless(len(aa.inbox) == 1)
        accepted, rm, why = aa.inbox.pop()
        self.failUnless(accepted)
        self.failUnless(rm.allowRFC822)
        self.failUnless(isinstance(rm, message.ReceivedMessage))
        return rm, why

    def rejected(self, aa):
        self.failUnless(len(aa.inbox) == 1)
        accepted, rm, why = aa.inbox.pop()
        self.failUnless(not accepted)
        if not isinstance(rm, message.ReceivedMessage):
            print rm
        self.failUnless(isinstance(rm, message.ReceivedMessage))
        self.failUnless(not rm.allowRFC822)
        #self.failUnless(isinstance(why, failure.Failure)) #??
        if aa.failures:
            print aa.failures
            self.fail("failure occurred during processing")
        return rm, why
        
    def testReceive2(self):
        # bob sends a message to alice, thinking it will get through, but it
        # is dropped because he doesn't have permission
        self.build(bobThinksHeHasPermission=1)

        # we start as bob, sending a message to alice
        alice = self.bob_book.lookupPetName("alice")
        m = self.doMessage(self.bob_owner, alice)
        encrypted = self.send(m)

        # now we play alice, receiving bob's message
        aa = self.alice_agent
        aa.receiveMessage(encrypted)

        rm, why = self.rejected(aa)
        self.failUnless(isinstance(why, NotPermitted))
        self.failUnlessEqual(why.args, ("no permission",))

    def testReceive3(self):
        # give bob permission to send everything to alice
        self.build(bobThinksHeHasPermission=1, bobHasPermission=1,
                   bobThinksAliceAcceptsTickets=1, aliceAcceptsTickets=1)

        alice = self.bob_book.lookupPetName("alice")

        m = self.doMessage(self.bob_owner, alice)
    
        # bob has specific permission, so there should be no additional
        # requirements
        #self.failUnlessEqual(m.requirements, [])
        encrypted = self.send(m)

        aa = self.alice_agent
        aa.receiveMessage(encrypted)

        rm, why = self.received(aa)

        self.failUnlessEqual(rm.getNode('rfc822')['body'], bobtext)
        self.failUnlessEqual(rm.senderEntry,
                             self.alice_book.lookupPetName("bob"))
        self.failUnlessEqual(why, "(specific) no ticket required")

    def testReceive4(self):
        # in which bob gets a ticket so he can send to alice
        self.build(bobThinksAliceAcceptsTickets=1, aliceAcceptsTickets=1)

        alice = self.bob_book.lookupPetName("alice")

        m = self.doMessage(self.bob_owner, alice)

        # add the ticket
        url = "http://ticketserver.com"
        ta = FakeTicketAgent()
        m.agent.ticketAgents = [ta]

        tt = self.alice_owner.genericPermission.ticketTakers[url]
        t1 = dr(ticketserver.Ticket(url, util.now(),
                                    "1").signed(self.alice_ticketkey), 60)
        ta.ticket = t1

        # sending the message should invoke the ticketAgent
        encrypted = self.send(m)
        self.failUnless(ta.asked)
        self.failUnless(m.getNode('rfc822')['tickets'])
        self.failUnless(m.getNode('rfc822')['tickets'][0] is t1)

        aa = self.alice_agent
        aa.receiveMessage(encrypted)

        # and alice should accept the message
        rm, why = self.received(aa)

        self.failUnlessEqual(rm.getNode('rfc822')['body'], bobtext)
        self.failUnlessEqual(rm.senderEntry,
                             self.alice_book.lookupPetName("bob"))
        self.failUnlessEqual(why, "(generic) valid ticket from http://ticketserver.com")

    def testReceive5(self):
        # in which bob has permission but sends too many messages and is
        # throttled by his own agent
        self.build(bobThinksHeHasPermission=1, bobHasPermission=1)

        alice = self.bob_book.lookupPetName("alice")
        aa = self.alice_agent

        # first message, sendable
        m = self.doMessage(self.bob_owner, alice)
        encrypted = self.send(m)
        aa.receiveMessage(encrypted)
        rm, why = self.received(aa)

        # now send another right away. This will fail because the
        # Inter-Message-Gap has been violated.
        m = self.doMessage(self.bob_owner, alice)
        err = self.sendFails(m)
        self.failUnless(err.check(NotPermitted))
        # TODO: make a new kind of exception rather than checking .args
        self.failUnlessEqual(err.value.args,
                             ("last message was sent too recently",))

        # now send another after "15" seconds. This one will succeed.
        util.timeOffset = 15
        m = self.doMessage(self.bob_owner, alice)
        encrypted = self.send(m)
        aa.receiveMessage(encrypted)
        rm, why = self.received(aa)

    def tearDown(self):
        util.timeOffset = 0

    def testReceive6(self):
        # in which bob has permission but sends too many messages and is
        # throttled by alice
        self.build(bobThinksHeHasPermission=1, bobHasPermission=1)

        alice = self.bob_book.lookupPetName("alice")
        alice.ourPermission.minIMG = 0
        aa = self.alice_agent

        # first message, accepted
        m = self.doMessage(self.bob_owner, alice)
        encrypted = self.send(m)
        aa.receiveMessage(encrypted)
        rm, why = self.received(aa)

        # now send another, rejected by alice
        m = self.doMessage(self.bob_owner, alice)
        encrypted = self.send(m)
        aa.receiveMessage(encrypted)
        rm, why = self.rejected(aa)
        self.failUnless(isinstance(why, NotPermitted))
        self.failUnlessEqual(why.args, ("last message was sent too recently",))

        # now send another after "15" seconds
        m = self.doMessage(self.bob_owner, alice)
        encrypted = self.send(m)
        util.warpForward(15)
        aa.receiveMessage(encrypted)
        rm, why = self.received(aa)


# TODO: another test would be to make sure that attempts which fail
# (because of size limits) don't count against the rate limits
