
from twisted.trial import unittest
dr = unittest.deferredResult

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import util, ticket, crypto
from petmail.server import ticketserver

class EncodeTest(unittest.TestCase):
    def testEncode1(self):
        url = "http://ticketserver.com"
        when = util.now()
        ticketID = "1"
        t = ticketserver.Ticket(url, when, ticketID)
        e = t.encode()
        e2 = util.decode(e)
        self.failUnlessEqual(e2['URL'], url)
        self.failUnlessEqual(e2['Timestamp'], when)
        self.failUnlessEqual(e2['TicketID'], ticketID)
    
class TakerTest(unittest.TestCase):
    def setUp(self):
        self.crypto = crypto.EncryptionAgent()
        self.key1 = dr(self.crypto.generateKey("ticketserver.com"), 60)
        self.taker1 = ticket.TicketTaker("http://ticketserver.com", None,
                                         self.key1.pubkey())
        self.taker1.keyID = self.key1.getID()

        self.key2 = dr(self.crypto.generateKey("ticketserver2.com"), 60)
        self.taker2 = ticket.TicketTaker("http://ticketserver2.com", None,
                                         self.key2.pubkey())
        self.taker2.keyID = self.key2.getID()

    def build(self, url, when, ticketID):
        t = ticketserver.Ticket(url, when, ticketID)
        s = dr(t.signed(self.key1), 60)
        return s

    def build2(self, url, when, ticketID):
        signed = self.build(url, when, ticketID)
        return ticket.joinTickets([signed])

    def testDecode1(self):
        when = util.now()
        s = self.build("http://ticketserver.com", when, "1")
        s2 = ticket.joinTickets([s])

        tickets = dr(ticket.extractTickets(s2, self.crypto), 60)
        self.failUnlessEqual(len(tickets), 1)
        url, sigID, t = tickets[0]
        self.failUnlessEqual(url, "http://ticketserver.com")
        self.failUnlessEqual(sigID, self.key1.getID())
        self.failUnlessEqual(t['URL'], url)
        self.failUnlessEqual(t['TicketID'], "1")
        self.failUnlessEqual(t['Timestamp'], when)

    def checkTicketValid(self, taker, signer, t):
        try:
            taker.checkTicket(signer, t)
        except ticket.InvalidTicket, why:
            self.fail("got InvalidTicket: %s" % why.args)

    def checkTicketInvalid(self, expectedReason, taker, sigID, t):
        try:
            taker.checkTicket(sigID, t)
        except ticket.InvalidTicket, why:
            self.failUnless(why.args[0].startswith(expectedReason))
        else:
            self.fail("Didn't raise InvalidTicket")
        
    def testValidate(self):
        when = util.now()
        oldticket = self.build2(self.taker1.url, when - 10*24*60*60, "1")
        goodticket1 = self.build2(self.taker1.url, when, "2")
        goodticket2 = self.build2(self.taker1.url, when, "3")

        ((url, sigID, t),) = dr(ticket.extractTickets(oldticket,
                                                      self.crypto), 60)
        self.checkTicketInvalid("The ticket was too old",
                                self.taker1, sigID, t)

        ((url, sigID, t),) = dr(ticket.extractTickets(goodticket1,
                                                      self.crypto), 60)
        self.checkTicketValid(self.taker1, sigID, t)

        ((url, sigID, t),) = dr(ticket.extractTickets(goodticket1,
                                                      self.crypto), 60)
        self.checkTicketInvalid("The ticket was signed by the wrong key",
                                self.taker2, sigID, t)

        ((url, sigID, t),) = dr(ticket.extractTickets(goodticket1,
                                                      self.crypto), 60)
        self.checkTicketInvalid("The ticket was issued by the wrong vendor",
                                self.taker2, self.key2.getID(), t)

        ((url, sigID, t),) = dr(ticket.extractTickets(goodticket2,
                                                      self.crypto), 60)
        self.checkTicketValid(self.taker1, sigID, t)

        # replay
        ((url, sigID, t),) = dr(ticket.extractTickets(goodticket2,
                                                      self.crypto), 60)
        self.checkTicketInvalid("Ticket is a duplicate",
                                self.taker1, sigID, t)
