
from twisted.trial import unittest
from twisted.internet import reactor, protocol, defer
from twisted.protocols import basic
dr = unittest.deferredResult

import os, sys
from StringIO import StringIO
from twisted.python import log
#log.startLogging(sys.stderr)

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import util, mm_util
from petmail.owner import OwnerEntry
from petmail.message import OutboundMessage
from petmail.interfaces import UndecryptableError
from petmail.transport import base, dummy, mail, pb, imap, pop

try:
   from petmail.transport import jabber
except ImportError:
   # happens the first time the interpreter tries to import it
   jabber = None
if jabber and not jabber.supported:
   # happens second and later times
   jabber = None

class FakeAgent:
    pass
class FakeOwner:
    pass

class TestInlet(unittest.TestCase):

    def testDummy(self):
        i = dummy.DummyInlet("foo")
        desc = i.encode()
        self.failUnlessEqual(desc, ("Dummy", "foo"))

    def testPB(self):
        i = pb.PBInlet("localhost", 0)
        desc = i.encode()
        self.failUnlessEqual(desc, ("PB", "localhost", 0))

    def testMailSocket(self):
        i = mail.MailSocket(0, "alice@foo.com")
        desc = i.encode()
        self.failUnlessEqual(desc, ("SMTP", "alice@foo.com"))

    def testMaildir(self):
        i = mail.MaildirInlet("~/mdir-petmail", "alice@foo.com")
        desc = i.encode()
        self.failUnlessEqual(desc, ("SMTP", "alice@foo.com"))
        

class TestTransport(unittest.TestCase):
    def setUp(self):
        self.messages = []

    def receiveMessage(self, msg, source):
        self.messages.append((msg, source))

    def testDummy(self):
        i = dummy.DummyInlet("foo")
        i.parent = self
        t = base.makeTransport(i.encode(), None)
        t.send("message")
        self.failUnlessEqual(self.messages, [("message", (i, None, None))])

    def testEquivalence(self):
        # inlets are used as keys in the Agent, so equivalent ones must
        # compare to equal
        i1 = dummy.DummyInlet("foo")
        i1.parent = 9
        i2 = dummy.DummyInlet("foo")
        i3 = dummy.DummyInlet("bar")
        self.failUnlessEqual(i1, i2)
        self.failIfEqual(i1, i3)
        d = {}
        d[i1] = 1
        self.failUnless(i2 in d)
        self.failUnlessEqual(pb.PBInlet("localhost", 0),
                             pb.PBInlet("localhost", 0))
        self.failUnlessEqual(mail.MaildirInlet("maildir", "foo"),
                             mail.MaildirInlet("maildir", "foo"))
        self.failUnlessEqual(mail.MailSocket(0, "alice@foo.com"),
                             mail.MailSocket(0, "alice@foo.com"))
        self.failUnlessEqual(imap.IMAPPoller("localhost", 143,
                                             "user", "pass", 60,
                                             None, "alice@foo.com"),
                             imap.IMAPPoller("localhost", 143,
                                             "user", "pass", 60,
                                             None, "alice@foo.com"))
        if jabber:
            j1 = jabber.JabberInlet("petmail.lothar.com",
                                    "alice", "alice",
                                    "petmail.lothar.com", 5222)
            j2 = jabber.JabberInlet("petmail.lothar.com",
                                    "alice", "alice",
                                    "petmail.lothar.com", 5222)
            self.failUnlessEqual(j1, j2)

    def testPB(self):
        i = pb.PBInlet("localhost", 0)
        i.parent = self
        i.startService()
        portnum = i.services[0]._port.getHost().port
        target = ("PB", "localhost", portnum)
        t = base.makeTransport(target, None)
        t.send("message")
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, [("message", (i, None, None))])


    def testEncapsulate(self):
        message = "message\x00\x01\x02" * 40
        encoded = mail.encapsulateInText(message)
        # the encoded form should survive weird newlines and arbitrary
        # header/footer additions

        encoded.replace("\n", "\n\r", 5)
        encoded = "here to fool you\n" + encoded + "footer\r" + "feet\n"
        encoded = encoded + "little bwitty toeses\n\n"
        
        i = mail.SMTPEncapsulatedInlet()
        i.parent = self

        # test .gotText
        i.gotText(encoded)
        self.failUnlessEqual(self.messages, [(message, (i, None, None))])

        # test .gotFile
        self.messages = []
        s = StringIO(encoded)
        i.gotFile(s)
        self.failUnlessEqual(self.messages, [(message, (i, None, None))])
        
    def testMixminionPlaintext1(self):
        message = "message\x00\x01\x02" * 40
        encoded = mail.encapsulateInText(message)

        # when an anonymous sender uses Mixminion to deliver a message (when
        # *not* using a SURB), it will be delivered in a mixminion
        # "plaintext" envelope. We must unwrap that envelope to accept the
        # petmail messages hiding within. This can be done with 'mixminion
        # decode', but we can also do it ourselves, which means that
        # recipients of anonymous messages don't need to have mixminion
        # installed.
        encoded = "\n" + encoded
        encoded = encoded.replace("\n--", "\n- --")
        encoded = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----\n" + \
                  "Message-type: plaintext\n" + \
                  encoded + \
                  "-----END TYPE III ANONYMOUS MESSAGE-----\n"

        i = mail.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return False
        i.canDecryptType3 = canDecryptType3

        i.gotText(encoded)
        self.failUnlessEqual(self.messages, [(message, (i, None, None))])

    def testMixminionEncrypted1(self):
        # on the other hand, when a message *is* sent through a SURB, it
        # will be delivered in an encrypted envelope, and we require the
        # mixminion executable to decrypt it. We also need to mark the SURB
        # they used so we know it's no longer outstanding.

        # this test just verifies that we reject the message when we don't
        # have mixminion support

        message = "message\x00\x01\x02" * 40
        encoded = mail.encapsulateInText(message)

        encoded = "\n" + encoded
        encoded = encoded.replace("\n--", "\n- --")
        encoded = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----\n" + \
                  "Message-type: encrypted\n" + \
                  "Decoding-handle: blahblahblah\n" + \
                  encoded + \
                  "-----END TYPE III ANONYMOUS MESSAGE-----\n"

        i = mail.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return False
        i.canDecryptType3 = canDecryptType3

        self.failUnlessRaises(UndecryptableError, i.gotText, encoded)

    def testMixminionEncrypted2(self):
        # this test verifies that we pass the message to the owner when we
        # need to. We have the owner pretend to decrypt the message, to
        # verify the other half.

        message = "message\x00\x01\x02" * 40
        encoded = mail.encapsulateInText(message)

        mmencoded = "\n" + encoded
        mmencoded = mmencoded.replace("\n--", "\n- --")
        mmencoded = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----\n" + \
                    "Message-type: encrypted\n" + \
                    "Decoding-handle: blahblahblah\n" + \
                    mmencoded + \
                    "-----END TYPE III ANONYMOUS MESSAGE-----\n"

        i = mail.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return True
        i.canDecryptType3 = canDecryptType3
        def decryptType3(encrypted, self=self):
            self.encrypted = encrypted
            return defer.succeed((encoded, "idstring", "surbid"))
        i.decryptType3 = decryptType3

        dr(i.gotText(mmencoded))
        self.failUnlessEqual(self.encrypted, mmencoded)
        self.failUnlessEqual(self.messages,
                             [(message, (i, "idstring", "surbid"))])


    def gotFile(self, fp):
        m = fp.read()
        self.receiveMessage(m, ("i", None, None))

    def testMaildir(self):
        os.mkdir("maildir")
        newdir = os.path.join("maildir", "new")
        olddir = os.path.join("maildir", "cur")
        tmpdir = os.path.join("maildir", "tmp")
        os.mkdir(newdir)
        os.mkdir(olddir)
        os.mkdir(tmpdir)
        i = mail.MaildirInlet("maildir", "foo")
        i.gotFile = self.gotFile
        i.pollinterval = 0.5
        i.startService()
        self.failUnlessEqual(self.messages, [])
        open(os.path.join(tmpdir, "bar"), "w").write("message")
        os.rename(os.path.join(tmpdir, "bar"), os.path.join(newdir, "bar"))
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, [("message", ("i", None, None))])
        self.failIf(os.path.exists(os.path.join(newdir, "bar")))
        self.failUnless(os.path.exists(os.path.join(olddir, "bar")))

    def testMail(self):
        if not os.environ.get("PETMAIL_TEST_SEND"):
            # this actually sends mail, only do it manually
            raise unittest.SkipTest("define PETMAIL_TEST_SEND=1 to run this")
        i = mail.MaildirInlet("~/Maildir-petmailtest", "foo")
        i.parent = self
        i.pollinterval = 0.5
        i.startService()
        a = FakeAgent()
        a.smarthost = "localhost"
        owner = OwnerEntry()
        owner.fromEmail = "me@localhost"
        owner.parent = a
        msg = OutboundMessage(a)
        msg.senderEntry = owner
        t = base.makeTransport(("SMTP", "warner-petmailtest@localhost"), msg)
        message = "message\x00\x01\x02" * 40
        d = t.send(message)
        r = unittest.deferredResult(d)
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, [(message, (i, None, None))])

    def testMailSocket(self):
        i = mail.MailSocket(0, "alice@foo.com")
        i.parent = self
        i.startService()
        portnum = i.services[0]._port.getHost().port
        # this is nicked from twisted.test.test_tpfile.FileSendingTestCase

        class FileSendingClient(protocol.Protocol):
            def __init__(self, f):
                self.f = f

            def connectionMade(self):
                s = basic.FileSender()
                d = s.beginFileTransfer(self.f, self.transport, lambda x: x)
                d.addCallback(lambda r: self.transport.loseConnection())

        class SendingFactory(protocol.ClientFactory):
            def buildProtocol(self, addr):
                return self.p

        message = "message\x00\x01\x02" * 40
        e = mail.encapsulateInText(message)
        f = SendingFactory()
        f.p = FileSendingClient(StringIO(e))
        reactor.connectTCP("localhost", portnum, f)
        now = util.now()
        while util.now() < now+5:
            if self.messages:
                break
            reactor.iterate(0.1)
        i.stopService()
        self.failUnlessEqual(self.messages, [(message, (i, None, None))])
        
