
from StringIO import StringIO

from twisted.python import log
from twisted.internet import defer

from petmail import crypto, util
from petmail.server import ticketserver

"""
A 'ticket' is a small piece of data which proves that you've paid something.
In this system, ticket servers give you a ticket to show that you've spent
human time (a CAPTCHA challenge) or significant CPU time (hashcash). Access
to the mailbox by unknown correspondents is rate-limited by requiring
tickets for each message.

Recipients will accept tickets from a set of 'ticket vendors'. These are
described with a URL. This URL will always contain the recipient's ID (the
SHA-1 hash of their public key, as an ASCII hex string). The would-be sender
can point their browser at the URL.. somehow. [TODO]

The ticket itself is a signed encoded dictionary with the following keys:
 URL: the URL of the ticket
 Timestamp: a time() string (seconds since 01Jan1970)
 TicketID: a string, generated like a MessageID.

Receivers will accept tickets with timestamps up to 5 days in the past, or 1
day in the future (the '5 days' should match the delivery timeout). The
combination of URL+Timestamp+TicketID should be globally unique: duplicates
will be rejected to prevent replay attacks. TicketID can be empty if the
vendor never issues two tickets during the same second, or it could be
globally unique all by itself if the vendor wanted to use it to reference
historical logs of issued tickets.

Tickets are checked by a TicketTaker. In general, only one of the recipient's
TicketTakers has to be satisfied to let the message through.
"""

class InvalidTicket(Exception):
    pass

class TicketTaker(object):
    debug = 0

    def __init__(self, url, agent, pubkey=None):
        self.url = url
        self.agent = agent
        if not pubkey:
            d = ticketserver.PBTicketKeyClient(url).getKey()
            d.addCallback(self._gotPubkey1)
            d.addErrback(log.err)
        else:
            self.pubkey = pubkey
        # anti-replay history
        self.historyLength = 5*24*60*60 # 5 days
        self.history = [] # list of (when, TicketID) tuples

    def _gotPubkey1(self, data):
        d = self.agent.crypto.addPubkey(data)
        d.addCallback(self._gotPubkey2)
        return d

    def _gotPubkey2(self, pubkey):
        self.pubkey = pubkey

    def __cmp__(self, other):
        return cmp(self.__dict__, self.__dict__)

    def __repr__(self):
        return "<TicketTaker %s>" % self.url

    def checkTicket(self, sigID, t):
        """Check a decoded ticket. Returns 1 if the ticket is valid, raises
        InvalidTicket if not.
        """
        self.when = None
        if sigID != self.pubkey.getID():
            why = "The ticket was signed by the wrong key: %s not %s" % \
                  (sigID, self.pubkey.getID())
            raise InvalidTicket, why
        if t.get('URL') != self.url:
            why = "The ticket was issued by the wrong vendor: %s not %s" % \
                  (t.get('URL'), self.url)
            raise InvalidTicket, why
        timestamp = t.get('Timestamp')
        forgetfulness = util.now() - self.historyLength
        if timestamp < forgetfulness:
            raise InvalidTicket, "The ticket was too old"
        ticketID = t.get('TicketID')
        newhistory = []
        for h_timestamp, h_ticketID in self.history:
            if (h_timestamp, h_ticketID) == (timestamp, ticketID):
                #age = util.now() - h_timestamp
                raise InvalidTicket, "Ticket is a duplicate"
            if h_timestamp >= forgetfulness:
                newhistory.append((h_timestamp, h_ticketID))
        newhistory.append((timestamp, ticketID))
        self.history = newhistory

        # ticket is valid
        return 1

# the input to extractTickets, and the output of joinTickets, is:
#  jelly([sign(jelly(ticket1)), sign(jelly(ticket2))])
# because sign(jelly(ticket1)) is what the ticket server gives back to us

class TicketExtractor(object):
    def __init__(self):
        self.tickets = []
    def run(self, signedTickets, ca):
        dl = []
        for signedTicket in signedTickets:
            d = defer.maybeDeferred(ca.decrypt, signedTicket)
            d.addCallback(self.gotTicket)
            dl.append(d)
        dlist = defer.DeferredList(dl)
        dlist.addCallback(self.gotList)
        return dlist

    def gotTicket(self, results):
        (recip, sigID, plaintext) = results
        ticket = util.decode(plaintext)
        url = ticket.get('URL')
        self.tickets.append((url, sigID, ticket))
        return None

    def gotList(self, results):
        return self.tickets

def extractTickets(signedTickets, ca):
    """I return a Deferred which fires with a list of tuples. Each tuple is
    (url, sigID, t)."""
    t = TicketExtractor()
    return t.run(signedTickets, ca)

def joinTickets(tickets):
    for t in tickets:
        assert(isinstance(t, type("")))
    return tickets
