#! /usr/bin/python

import os, os.path, re
import email
from email import Encoders, Parser, Message
import base64

from twisted.internet import protocol, error, defer, reactor
from twisted.application import internet
from twisted.protocols.smtp import sendmail

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from petmail import interfaces, pbutil, util
from petmail.transport import base
from petmail.interfaces import Unsendable


SMTP_SUBJECT = "PETMAIL MESSAGE"
SMTP_HEADER = "-----BEGIN PETMAIL MESSAGE-----"
SMTP_FOOTER = "-----END PETMAIL MESSAGE-----"
TYPE3_HEADER = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----"
TYPE3_FOOTER = "-----END TYPE III ANONYMOUS MESSAGE-----"

class SMTPEncapsulatedInlet(base.BaseInlet):

    # the next three methods may raise an exception, in which case the error
    # should be logged and the message dropped (or perhaps put in a log
    # directory for a limited period of time)

    def gotText(self, text):
        """This method should be called when the inlet has received an
        incoming message ready to be scanned for the ---BEGIN PETMAIL.. line
        This particular function should be given the message as a string."""
        f = StringIO(text)
        return self.gotFile(f)

    def gotFile(self, fp):
        """This function receives the message as an open file. The object
        must support the readline() and seek() methods.
        """
        # the message is either a --PETMAIL MESSAGE-- or a --TYPE III--
        d = self.unwrapPadding(fp, acceptType3=True)
        d.addCallback(self._gotFile2)
        return d

    def _gotFile2(self, results):
        encoded, identity, surbid = results
        return self.gotMessage(encoded, (identity, surbid))

    def unwrapPadding(self, fp, acceptType3):
        """This method receives the message as an open file. The object must
        support the readline() and seek() methods. It returns a Deferred
        that fires with a tuple of encrypted/idstring/surbid.
        """
        mm = "Malformed message "
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError(mm + "(no BEGIN banner)")
            if line.rstrip() == SMTP_HEADER:
                line = fp.readline()
                if not line:
                    raise ValueError(mm + "(missing separator)")
                if line.strip():
                    raise ValueError(mm + "(unexpected headers)")
                break
            if line.rstrip() == TYPE3_HEADER:
                if acceptType3:
                    return self.unwrapType3(fp)
                else:
                    raise ValueError(mm + "(recursive type3)")
        body = ""
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError(mm + "(no END banner)")
            if line.rstrip() == SMTP_FOOTER:
                break
            body += line

        encoded = base64.decodestring(body)
        results = (encoded, None, None)
        return defer.succeed(results)


    def unwrapType3(self, fp):
        # first off, do we have the mixminion client? If so, use it to
        # unwrap the type3-encoding.
        if self.canDecryptType3():
            fp.seek(0)
            text = fp.read()
            d = self.decryptType3(text)
            d.addCallback(self._decryptedType3)
            return d
        # we've just read the BEGIN line. Next are the header lines
        plaintext = False
        while 1:
            line = fp.readline()
            if not line:
                raise ValueError("Malformed message")
            line = line.rstrip()
            if re.search(r'^Message-type:\s*plaintext$', line, re.I):
                plaintext = True
            elif re.search(r'^Message-type:', line, re.I):
                plaintext = False
            elif line == '':
                # end of headers
                break
        if not plaintext:
            msg = "Type3 message requires mixminion executable"
            raise interfaces.UndecryptableError(msg)
        body = ""
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError("Malformed message")
            if line.rstrip() == TYPE3_FOOTER:
                break
            if line.startswith("- "):
                line = line[2:]
            body += line
        return self.unwrapPadding(StringIO(body), acceptType3=False)

    def _decryptedType3(self, results):
        text, idstring, surbid = results
        # unwrap the plaintext, but tack on the idstring/surbid
        d = self.unwrapPadding(StringIO(text), acceptType3=False)
        d.addCallback(self._unwrappedDecryptedType3, idstring, surbid)
        return d

    def _unwrappedDecryptedType3(self, results, idstring, surbid):
        text, dummy1, dummy2 = results
        return (text, idstring, surbid)


class PollingSMTPInlet(SMTPEncapsulatedInlet):
    pollinterval = 600 # seconds

    def startService(self):
        # TimerService is not restartable in twisted-1.3.0rc1: they serialize
        # ok but cannot be reactivated after .stopService . I fixed the
        # problem in CVS, but the fix has not yet propagated into a release.
        # As a result, we must create a new TimerService each time we're
        # started. This is particularly irritating because the service is one
        # of our children.
        for child in list(self):
            child.disownServiceParent()
        timer = internet.TimerService(self.pollinterval, self.check)
        timer.setServiceParent(self)

        SMTPEncapsulatedInlet.startService(self)

    def check(self):
        """I check the mailbox and return a Deferred which will fire when the
        check has completed.
        """
        raise Unimplemented

class MaildirInlet(PollingSMTPInlet):
    """I watch a maildir for inbound messages."""
    __implements__ = (PollingSMTPInlet.__implements__,
                      interfaces.IInlet2_Maildir)
    itype = "maildir"
    pollinterval = 10 # seconds

    def __init__(self, basedir, address):
        PollingSMTPInlet.__init__(self)
        self.basedir = basedir
        self.address = address
        self.newdir = os.path.join(os.path.expanduser(basedir), "new")
        self.olddir = os.path.join(os.path.expanduser(basedir), "cur")
        self.description = "SMTP:%s" % address

    def encode(self):
        return ("SMTP", self.address)

    def __cmp__(self, other):
        return (cmp(type(self), type(other)) or
                cmp(self.__class__, other.__class__) or
                cmp((self.basedir, self.address),
                    (other.basedir, other.address)))
    def __hash__(self):
        return hash((self.__class__, self.basedir, self.address))

    def check(self):
        for f in os.listdir(self.newdir):
            newfile = os.path.join(self.newdir, f)
            oldfile = os.path.join(self.olddir, f)
            self.gotFile(open(newfile, "r"))
            os.rename(newfile, oldfile)

    def getProperty(self, pname):
        if pname in ("basedir", "address"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))
                      

class MailSocketProtocol(protocol.Protocol):
    def __init__(self):
        self.data = ""
    def dataReceived(self, data):
        self.data += data
    def connectionLost(self, reason):
        if reason.check(error.ConnectionDone):
            self.factory.parent.gotText(self.data)

class MailSocket(SMTPEncapsulatedInlet):
    # [MailSocket:]9990:alice@foo.com
    # socketnum : inletaddress
    __implements__ = (SMTPEncapsulatedInlet.__implements__,
                      interfaces.IInlet2_MailSocket)
    itype = "mailsocket"

    def __init__(self, portnum, address):
        SMTPEncapsulatedInlet.__init__(self)
        assert type(portnum) == int
        self.portnum = portnum
        self.address = address
        f = protocol.ServerFactory()
        f.protocol = MailSocketProtocol
        f.parent = self
        server = internet.TCPServer(portnum, f)
        server.setServiceParent(self)
        self.description = "SMTP:%s" % address

    def encode(self):
        return ("SMTP", self.address)

    def __cmp__(self, other):
        return (cmp(type(self), type(other)) or
                cmp(self.__class__, other.__class__) or
                cmp((self.portnum, self.address),
                    (other.portnum, other.address)))
    def __hash__(self):
        return hash((self.__class__, self.portnum, self.address))

    def getProperty(self, pname):
        if pname in ("portnum", "address"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))



#scheme, smarthost, fromEmail = sender.parent.smtpScheme.split(":")

def encapsulateInText(encrypted):
    """This puts the PETmail message body into a ---PETMAIL MESSAGE---
    payload block. It returns a string ready to be inserted into the body
    of a Message object. The subject line of the resulting message must
    be 'PETMAIL MESSAGE'."""

    return util.armor(encrypted, "MESSAGE")

class JustMailIt(object):
    """This Transport class uses twisted.mail to deliver the message via SMTP
    to a designated relay host. For normally-abled unix boxes that run an MTA
    like sendmail, this defaults to 'localhost'. Others will need to set it
    to a host which *is* capable of sending mail.
    """
    __implements__ = interfaces.ITransport,

    def __init__(self, trecord, message):
        t, self.dest = trecord
        owner = message.senderEntry
        self.fromEmail = owner.fromEmail
        agent = owner.parent
        self.smarthost = agent.smarthost

    def send(self, encrypted):
        """This returns a Deferred which will fire when the message has been
        sent successfully."""
        m = Message.Message()
        m.set_payload(encapsulateInText(encrypted))
        m['Subject'] = SMTP_SUBJECT
        m['X-PETmail-URL'] = "http://petmail.lothar.com/"
        m['From'] = self.fromEmail
        m['To'] = self.dest
        s = m.as_string()
        return sendmail(self.smarthost, self.fromEmail, self.dest, s)
base.registerTransport("SMTP", False, JustMailIt)


class MailAnonymously(object):
    """This subclass sends the message out through a remailer network to a
    known recipient, making it nigh-impossible to figure out where it came
    from. The only information available to the recipient (or an
    eavesdropper) is the contents of the IDRecord.

    This is the anonymous-sender, non-anonymous-recipient case. For
    bidirectional anonymity or anonymous-recipient, use MailToSURB."""
    __implements__ = interfaces.IUntraceableTransport,

    def __init__(self, trecord, message):
        t, self.dest = trecord
        self.mmc = message.agent.mmc

    def send(self, encrypted):
        body = encapsulateInText(encrypted)
        return self.mmc.sendToDest(body, self.dest, SMTP_SUBJECT)
base.registerTransport("SMTP", True, MailAnonymously)


class MailToSURB(object):
    """This subclass sends the message out through a remailer network to an
    anonymous recipient through a SURB. This transport type is only good for
    a single small message (it must compress to smaller than 28k).

    This is the bidirectional anonymity / anonymous-recipient case."""

    __implements__ = interfaces.IUntraceableTransport,

    def __init__(self, trecord, message):
        t, self.destSURB = trecord
        assert type(self.destSURB) == str
        self.mmc = message.agent.mmc

    def send(self, encrypted):
        body = encapsulateInText(encrypted)
        # how many surbs will we need?
        d = self.mmc.countSURBs(body)
        d.addCallback(self._counted, encrypted)
        return d

    def _counted(self, count, encrypted):
        if count > 1:
            raise Unsendable("Message is too big to fit in one block")
        d = mmc.sendToSURBs(body, self.destSURB, SMTP_SUBJECT)
        return d
base.registerTransport("SURB", True, MailToSURB)

class SURBServer(object):
    """This subclass retrieves some SURBs from a SURBServer and uses them to
    send the message.
    """

    __implements__ = interfaces.ITransport,
    # This is not untraceable because retrieving SURBs will reveal your
    # location to the SURBServer, even if actually using the resulting SURBs
    # is secure.

    def __init__(self, trecord, message):
        t, server = trecord
        sid = idrecord.IDRecord()
        d = sid.createFromSigned(server, message.agent.crypto)
        d.addCallback(self._gotSID, message)
        return d

    def _gotSID(self, sid, message):
        self.sid = sid
        # now request a SURB from them
        # TODO
        raise Unimplemented("TODO")
#base.registerTransport("SURBServer", False, SURBServer)

class MailWithSURBs(object):
    """This Transport sends messages with SURBs that come from a local
    SURBSupply, which is periodically replenished by the recipient. It is
    not created from a Transport Record like other transports.
    """

    def __init__(self, surbsupply, message):
        self.supply = surbsupply
        self.mmc = message.agent.mmc
        # TODO: if we don't have mixminion, we may be able to use a gateway
        # (at least for single-SURB messages, since we won't have any way to
        # do count-surbs)

    def send(self, encrypted):
        body = encapsulateInText(encrypted)
        # how many surbs will we need?
        d = self.mmc.countSURBs(body)
        d.addCallback(self._counted, encrypted)
        return d

    def _counted(self, count, encrypted):
        surbs = self.supply.get(count) # may raise Unsendable
        surbdata = "\n".join(surbs)
        d = mmc.sendToSURBs(body, surbdata, SMTP_SUBJECT)
        return d

